@@ -2,7 +2,8 @@
  * compress.c - Compressed attribute handling code.  Part of the Linux-NTFS
  *		project.
  *
- * Copyright (c) 2004 Anton Altaparmakov
+ * Copyright (c) 2004-2005 Anton Altaparmakov
+ * Copyright (c)      2005 Yura Pakhuchiy
  *
  * This program/include file is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as published
@@ -22,10 +23,18 @@
 
 #include "config.h"
 
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_ERRNO_H
 #include <errno.h>
+#endif
 
 #include "attrib.h"
 #include "debug.h"
@@ -84,7 +93,7 @@ static int ntfs_decompress(u8 *dest, const u32 dest_size,
 	u8 tag;			/* Current tag. */
 	int token;		/* Loop counter for the eight tokens in tag. */
 
-	Dprintf("Entering, cb_size = 0x%x.\n", cb_size);
+	Dprintf("Entering, cb_size = 0x%x.\n", (unsigned)cb_size);
 do_next_sb:
 	Dprintf("Beginning sub-block at offset = 0x%x in the cb.\n",
 			cb - cb_start);
@@ -253,21 +262,27 @@ static __inline__ BOOL ntfs_is_cb_compressed(ntfs_attr *na,
 		runlist_element *rl, VCN cb_start_vcn, int cb_clusters)
 {
 	/*
-	 * The simplest case:  the run starting at @cb_start_vcn contains
+	 * The simplest case: the run starting at @cb_start_vcn contains
 	 * @cb_clusters clusters which are all not sparse, thus the cb is not
 	 * compressed.
 	 */
-	if (rl->length - (cb_start_vcn - rl->vcn) >= cb_clusters)
-		return FALSE;
+restart:
 	cb_clusters -= rl->length - (cb_start_vcn - rl->vcn);
-	do {
+	while (cb_clusters > 0) {
 		/* Go to the next run. */
 		rl++;
 		/* Map the next runlist fragment if it is not mapped. */
 		if (rl->lcn < LCN_HOLE || !rl->length) {
+			cb_start_vcn = rl->vcn;
 			rl = ntfs_attr_find_vcn(na, rl->vcn);
 			if (!rl || rl->lcn < LCN_HOLE || !rl->length)
 				return TRUE;
+			/*
+			 * If the runs were merged need to deal with the
+			 * resulting partial run so simply restart.
+			 */
+			if (rl->vcn < cb_start_vcn)
+				goto restart;
 		}
 		/* If the current run is sparse, the cb is compressed. */
 		if (rl->lcn == LCN_HOLE)
@@ -276,7 +291,7 @@ static __inline__ BOOL ntfs_is_cb_compressed(ntfs_attr *na,
 		if (rl->length >= cb_clusters)
 			return FALSE;
 		cb_clusters -= rl->length;
-	} while (cb_clusters > 0);
+	};
 	/* All cb_clusters were not sparse thus the cb is not compressed. */
 	return FALSE;
 }
