@@ -21,15 +21,33 @@
 
 #include "config.h"
 
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
+#endif
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
+#ifdef HAVE_ERRNO_H
 #include <errno.h>
+#endif
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
+#endif
+#ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
+#endif
+#ifdef HAVE_FCNTL_H
 #include <fcntl.h>
-#include <sys/ioctl.h>
+#endif
+#ifdef HAVE_SYS_IOCTL_H
+#	include <sys/ioctl.h>
+#endif
 #ifdef HAVE_LINUX_FD_H
 #	include <linux/fd.h>
 #endif
@@ -272,7 +290,7 @@ s64 ntfs_mst_pread(struct ntfs_device *dev, const s64 pos, s64 count,
 {
 	s64 br, i;
 
-	if (bksize & (bksize - 1) || bksize % NTFS_SECTOR_SIZE) {
+	if (bksize & (bksize - 1) || bksize % NTFS_BLOCK_SIZE) {
 		errno = EINVAL;
 		return -1;
 	}
@@ -307,7 +325,7 @@ s64 ntfs_mst_pread(struct ntfs_device *dev, const s64 pos, s64 count,
  * @dev at position @pos.
  *
  * On success, return the number of successfully written blocks. If this number
- * is lower than @count this means that the write has been interrutped or that
+ * is lower than @count this means that the write has been interrupted or that
  * an error was encountered during the write so that the write is partial. 0
  * means nothing was written (also return 0 when @count or @bksize are 0).
  *
@@ -329,7 +347,7 @@ s64 ntfs_mst_pwrite(struct ntfs_device *dev, const s64 pos, s64 count,
 {
 	s64 written, i;
 
-	if (count < 0 || bksize % NTFS_SECTOR_SIZE) {
+	if (count < 0 || bksize % NTFS_BLOCK_SIZE) {
 		errno = EINVAL;
 		return -1;
 	}
@@ -529,7 +547,7 @@ s64 ntfs_device_size_get(struct ntfs_device *dev, int block_size)
  *
  * The following error codes are defined:
  *	EINVAL		Input parameter error
- *	ENOTSUP		System does not support HDIO_GETGEO ioctl
+ *	EOPNOTSUPP	System does not support HDIO_GETGEO ioctl
  *	ENOTTY		@dev is a file or a device not supporting HDIO_GETGEO
  */
 s64 ntfs_device_partition_start_sector_get(struct ntfs_device *dev)
@@ -548,7 +566,7 @@ s64 ntfs_device_partition_start_sector_get(struct ntfs_device *dev)
 		}
 	}
 #else
-	errno = ENOTSUP;
+	errno = EOPNOTSUPP;
 #endif
 	return -1;
 }
@@ -562,7 +580,7 @@ s64 ntfs_device_partition_start_sector_get(struct ntfs_device *dev)
  *
  * The following error codes are defined:
  *	EINVAL		Input parameter error
- *	ENOTSUP		System does not support HDIO_GETGEO ioctl
+ *	EOPNOTSUPP	System does not support HDIO_GETGEO ioctl
  *	ENOTTY		@dev is a file or a device not supporting HDIO_GETGEO
  */
 int ntfs_device_heads_get(struct ntfs_device *dev)
@@ -582,7 +600,7 @@ int ntfs_device_heads_get(struct ntfs_device *dev)
 		}
 	}
 #else
-	errno = ENOTSUP;
+	errno = EOPNOTSUPP;
 #endif
 	return -1;
 }
@@ -596,7 +614,7 @@ int ntfs_device_heads_get(struct ntfs_device *dev)
  *
  * The following error codes are defined:
  *	EINVAL		Input parameter error
- *	ENOTSUP		System does not support HDIO_GETGEO ioctl
+ *	EOPNOTSUPP	System does not support HDIO_GETGEO ioctl
  *	ENOTTY		@dev is a file or a device not supporting HDIO_GETGEO
  */
 int ntfs_device_sectors_per_track_get(struct ntfs_device *dev)
@@ -616,7 +634,7 @@ int ntfs_device_sectors_per_track_get(struct ntfs_device *dev)
 		}
 	}
 #else
-	errno = ENOTSUP;
+	errno = EOPNOTSUPP;
 #endif
 	return -1;
 }
