@@ -1,7 +1,8 @@
 /*
  * dir.c - Directory handling code. Part of the Linux-NTFS project.
  *
- * Copyright (c) 2002-2004 Anton Altaparmakov
+ * Copyright (c) 2002-2005 Anton Altaparmakov
+ * Copyright (c)      2005 Yura Pakhuchiy
  *
  * This program/include file is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as published
@@ -37,6 +38,10 @@
 #include "inode.h"
 #include "dir.h"
 #include "volume.h"
+#include "mft.h"
+#include "index.h"
+#include "ntfstime.h"
+#include "lcnalloc.h"
 
 /*
  * The little endian Unicode string "$I30" as a global constant.
@@ -109,9 +114,10 @@ u64 ntfs_inode_lookup_by_name(ntfs_inode *dir_ni, const ntfschar *uname,
 	ir = (INDEX_ROOT*)((u8*)ctx->attr +
 			le16_to_cpu(ctx->attr->value_offset));
 	index_block_size = le32_to_cpu(ir->index_block_size);
-	if (index_block_size < NTFS_SECTOR_SIZE ||
+	if (index_block_size < NTFS_BLOCK_SIZE ||
 			index_block_size & (index_block_size - 1)) {
-		Dprintf("Index block size %u is invalid.\n", index_block_size);
+		Dprintf("Index block size %u is invalid.\n",
+				(unsigned)index_block_size);
 		goto put_err_out;
 	}
 	index_end = (u8*)&ir->index + le32_to_cpu(ir->index.index_length);
@@ -286,9 +292,9 @@ descend_into_child_node:
 		Dprintf("Index buffer (VCN 0x%llx) of directory inode 0x%llx "
 				"has a size (%u) differing from the directory "
 				"specified size (%u).\n", (long long)vcn,
-				(unsigned long long)dir_ni->mft_no,
+				(unsigned long long)dir_ni->mft_no, (unsigned)
 				le32_to_cpu(ia->index.allocated_size) + 0x18,
-				index_block_size);
+				(unsigned)index_block_size);
 		errno = EIO;
 		goto close_err_out;
 	}
@@ -459,6 +465,111 @@ close_err_out:
 	goto eo_put_err_out;
 }
 
+/**
+ * ntfs_pathname_to_inode - Find the inode which represents the given pathname
+ * @vol:       An ntfs volume obtained from ntfs_mount
+ * @parent:    A directory inode to begin the search (may be NULL)
+ * @pathname:  Pathname to be located
+ *
+ * Take an ASCII pathname and find the inode that represents it.  The function
+ * splits the path and then descends the directory tree.  If @parent is NULL,
+ * then the root directory '.' will be used as the base for the search.
+ *
+ * Return:  inode  Success, the pathname was valid
+ *	    NULL   Error, the pathname was invalid, or some other error occurred
+ */
+ntfs_inode *ntfs_pathname_to_inode(ntfs_volume *vol, ntfs_inode *parent,
+		const char *pathname)
+{
+	u64 inum;
+	int len, err = 0;
+	char *p, *q;
+	ntfs_inode *ni;
+	ntfs_inode *result = NULL;
+	ntfschar *unicode = NULL;
+	char *ascii = NULL;
+
+	if (!vol || !pathname) {
+		errno = EINVAL;
+		return NULL;
+	}
+
+	if (parent) {
+		ni = parent;
+	} else {
+		ni = ntfs_inode_open(vol, FILE_root);
+		if (!ni) {
+			Dprintf("Couldn't open the inode of the root "
+					"directory.\n");
+			err = EIO;
+			goto close;
+		}
+	}
+
+	unicode = calloc(1, MAX_PATH);
+	ascii = strdup(pathname);
+	if (!unicode || !ascii) {
+		Dprintf("Out of memory.\n");
+		err = ENOMEM;
+		goto close;
+	}
+
+	p = ascii;
+	/* Remove leading /'s. */
+	while (p && *p && *p == PATH_SEP)
+		p++;
+	while (p && *p) {
+		/* Find the end of the first token. */
+		q = strchr(p, PATH_SEP);
+		if (q != NULL) {
+			*q = '\0';
+			q++;
+		}
+
+		len = ntfs_mbstoucs(p, &unicode, MAX_PATH);
+		if (len < 0) {
+			Dprintf("Couldn't convert name to Unicode: %s.\n", p);
+			err = EILSEQ;
+			goto close;
+		}
+
+		inum = ntfs_inode_lookup_by_name(ni, unicode, len);
+		if (inum == (u64) -1) {
+			Dprintf("Couldn't find name '%s' in "
+					"pathname '%s'.\n", p, pathname);
+			err = ENOENT;
+			goto close;
+		}
+
+		if (ni != parent)
+			ntfs_inode_close(ni);
+
+		inum = MREF(inum);
+		ni = ntfs_inode_open(vol, inum);
+		if (!ni) {
+			Dprintf("Cannot open inode %llu: %s.\n",
+					(unsigned long long)inum, p);
+			err = EIO;
+			goto close;
+		}
+
+		p = q;
+		while (p && *p && *p == PATH_SEP)
+			p++;
+	}
+
+	result = ni;
+	ni = NULL;
+close:
+	if (ni && (ni != parent))
+		ntfs_inode_close(ni);
+	free(ascii);
+	free(unicode);
+	if (err)
+		errno = err;
+	return result;
+}
+
 /*
  * The little endian Unicode string ".." for ntfs_readdir().
  */
@@ -699,9 +810,10 @@ int ntfs_readdir(ntfs_inode *dir_ni, s64 *pos,
 
 	/* Determine the size of a vcn in the directory index. */
 	index_block_size = le32_to_cpu(ir->index_block_size);
-	if (index_block_size < NTFS_SECTOR_SIZE ||
+	if (index_block_size < NTFS_BLOCK_SIZE ||
 			index_block_size & (index_block_size - 1)) {
-		Dprintf("Index block size %u is invalid.\n", index_block_size);
+		Dprintf("Index block size %u is invalid.\n",
+				(unsigned)index_block_size);
 		goto dir_err_out;
 	}
 	index_block_size_bits = ffs(index_block_size) - 1;
@@ -803,7 +915,7 @@ skip_index_root:
 	if (br != bmp_buf_size) {
 		if (br != -1)
 			errno = EIO;
-		Dperror("Failed to read from inde bitmap attribute");
+		Dperror("Failed to read from index bitmap attribute");
 		goto err_out;
 	}
 
@@ -826,12 +938,12 @@ find_next_index_buffer:
 		if (br != bmp_buf_size) {
 			if (br != -1)
 				errno = EIO;
-			Dperror("Failed to read from inde bitmap attribute");
+			Dperror("Failed to read from index bitmap attribute");
 			goto err_out;
 		}
 	}
 
-	Dprintf("Handling index block 0x%llx.", (long long)bmp_pos);
+	Dprintf("Handling index block 0x%llx.\n", (long long)bmp_pos);
 
 	/* Read the index block starting at bmp_pos. */
 	br = ntfs_attr_mst_pread(ia_na, bmp_pos << index_block_size_bits, 1,
@@ -858,9 +970,9 @@ find_next_index_buffer:
 				"has a size (%u) differing from the directory "
 				"specified size (%u).\n",
 				(long long)ia_start >> index_vcn_size_bits,
-				(unsigned long long)dir_ni->mft_no,
+				(unsigned long long)dir_ni->mft_no, (unsigned)
 				le32_to_cpu(ia->index.allocated_size) + 0x18,
-				index_block_size);
+				(unsigned)index_block_size);
 		goto dir_err_out;
 	}
 	index_end = (u8*)&ia->index + le32_to_cpu(ia->index.index_length);
@@ -947,3 +1059,450 @@ err_out:
 	return -1;
 }
 
+/**
+ * ntfs_create - create file or directory on ntfs volume
+ * @dir_ni:	ntfs inode for directory in which create new object
+ * @name:	unicode name of new object
+ * @name_len:	length of the name in unicode characters
+ * @type:	type of the object to create
+ *
+ * @type can be either NTFS_DT_REG to create regular file or NTFS_DT_DIR to
+ * create directory, other valuer are invalid.
+ *
+ * Return opened ntfs inode that describes created file on success or NULL
+ * on error with errno set to the error code.
+ */
+ntfs_inode *ntfs_create(ntfs_inode *dir_ni, ntfschar *name, u8 name_len,
+		const unsigned type)
+{
+	ntfs_inode *ni;
+	FILE_NAME_ATTR *fn = NULL;
+	STANDARD_INFORMATION *si = NULL;
+	int err, fn_len, si_len;
+
+	ntfs_debug("Entering.");
+	/* Sanity checks. */
+	if (!dir_ni || !name || !name_len ||
+			(type != NTFS_DT_REG && type != NTFS_DT_DIR)) {
+		ntfs_error(, "Invalid arguments.");
+		return NULL;
+	}
+	/* Allocate MFT record for new file. */
+	ni = ntfs_mft_record_alloc(dir_ni->vol, NULL);
+	if (!ni) {
+		ntfs_error(, "Failed to allocate new MFT record.");
+		return NULL;
+	}
+	/*
+	 * Create STANDARD_INFORMATION attribute. Write STANDARD_INFORMATION
+	 * version 1.2, windows will upgrade it to version 3 if needed.
+	 */
+	si_len = offsetof(STANDARD_INFORMATION, v1_end);
+	si = calloc(1, si_len);
+	if (!si) {
+		err = errno;
+		ntfs_error(, "Not enough memory.");
+		goto err_out;
+	}
+	si->creation_time = utc2ntfs(ni->creation_time);
+	si->last_data_change_time = utc2ntfs(ni->last_data_change_time);
+	si->last_mft_change_time = utc2ntfs(ni->last_mft_change_time);
+	si->last_access_time = utc2ntfs(ni->last_access_time);
+	/* Add STANDARD_INFORMATION to inode. */
+	if (ntfs_attr_add(ni, AT_STANDARD_INFORMATION, AT_UNNAMED, 0,
+			(u8*)si, si_len)) {
+		err = errno;
+		ntfs_error(, "Failed to add STANDARD_INFORMATION attribute.");
+		goto err_out;
+	}
+	if (type == NTFS_DT_DIR) {
+		INDEX_ROOT *ir = NULL;
+		INDEX_ENTRY *ie;
+		int ir_len, index_len;
+
+		/* Create INDEX_ROOT attribute. */
+		index_len = sizeof(INDEX_HEADER) + sizeof(INDEX_ENTRY_HEADER);
+		ir_len = offsetof(INDEX_ROOT, index) + index_len;
+		ir = calloc(1, ir_len);
+		if (!ir) {
+			err = errno;
+			ntfs_error(, "Not enough memory.");
+			goto err_out;
+		}
+		ir->type = AT_FILE_NAME;
+		ir->collation_rule = COLLATION_FILE_NAME;
+		ir->index_block_size = cpu_to_le32(ni->vol->indx_record_size);
+		if (ni->vol->cluster_size <= ni->vol->indx_record_size)
+			ir->clusters_per_index_block =
+					ni->vol->indx_record_size >>
+					ni->vol->cluster_size_bits;
+		else
+			ir->clusters_per_index_block =
+					-ni->vol->indx_record_size_bits;
+		ir->index.entries_offset = cpu_to_le32(sizeof(INDEX_HEADER));
+		ir->index.index_length = cpu_to_le32(index_len);
+		ir->index.allocated_size = cpu_to_le32(index_len);
+		ie = (INDEX_ENTRY*)((u8*)ir + sizeof(INDEX_ROOT));
+		ie->length = cpu_to_le16(sizeof(INDEX_ENTRY_HEADER));
+		ie->key_length = 0;
+		ie->flags = INDEX_ENTRY_END;
+		/* Add INDEX_ROOT attribute to inode. */
+		if (ntfs_attr_add(ni, AT_INDEX_ROOT, I30, 4, (u8*)ir, ir_len)) {
+			err = errno;
+			free(ir);
+			ntfs_error(, "Failed to add INDEX_ROOT attribute.");
+			goto err_out;
+		}
+	} else {
+		/* Add DATA attribute to inode. */
+		if (ntfs_attr_add(ni, AT_DATA, AT_UNNAMED, 0, NULL, 0)) {
+			err = errno;
+			ntfs_error(, "Failed to add DATA attribute.");
+			goto err_out;
+		}
+	}
+	/* Create FILE_NAME attribute. */
+	fn_len = sizeof(FILE_NAME_ATTR) + name_len * sizeof(ntfschar);
+	fn = calloc(1, fn_len);
+	if (!fn) {
+		err = errno;
+		ntfs_error(, "Not enough memory.");
+		goto err_out;
+	}
+	fn->parent_directory = MK_LE_MREF(dir_ni->mft_no,
+			le16_to_cpu(dir_ni->mrec->sequence_number));
+	fn->file_name_length = name_len;
+	fn->file_name_type = FILE_NAME_POSIX;
+	if (type == NTFS_DT_DIR)
+		fn->file_attributes = FILE_ATTR_DUP_FILE_NAME_INDEX_PRESENT;
+	fn->creation_time = utc2ntfs(ni->creation_time);
+	fn->last_data_change_time = utc2ntfs(ni->last_data_change_time);
+	fn->last_mft_change_time = utc2ntfs(ni->last_mft_change_time);
+	fn->last_access_time = utc2ntfs(ni->last_access_time);
+	memcpy(fn->file_name, name, name_len * sizeof(ntfschar));
+	/* Add FILE_NAME attribute to inode. */
+	if (ntfs_attr_add(ni, AT_FILE_NAME, AT_UNNAMED, 0, (u8*)fn, fn_len)) {
+		err = errno;
+		ntfs_error(, "Failed to add FILE_NAME attribute.");
+		goto err_out;
+	}
+	/* Add FILE_NAME attribute to index. */
+	if (ntfs_index_add_filename(dir_ni, fn, MK_MREF(ni->mft_no,
+			le16_to_cpu(ni->mrec->sequence_number)))) {
+		err = errno;
+		ntfs_error(, "Failed to add entry to the index.");
+		goto err_out;
+	}
+	/* Set hard links count and directory flag. */
+	ni->mrec->link_count = cpu_to_le16(1);
+	if (type == NTFS_DT_DIR)
+		ni->mrec->flags |= MFT_RECORD_IS_DIRECTORY;
+	ntfs_inode_mark_dirty(ni);
+	/* Done! */
+	free(fn);
+	free(si);
+	ntfs_debug("Done.");
+	return ni;
+err_out:
+	ntfs_debug("Failed.");
+	if (ntfs_mft_record_free(ni->vol, ni))
+		ntfs_error(, "Failed to free MFT record.  "
+				"Leaving inconsist metadata. Run chkdsk.");
+	if (fn)
+		free(fn);
+	if (si)
+		free(si);
+	errno = err;
+	return NULL;
+}
+
+/**
+ * ntfs_delete - delete file or directory from ntfs volume
+ * @ni:		ntfs inode for object to delte
+ * @dir_ni:	ntfs inode for directory in which delete object
+ * @name:	unicode name of the object to delete
+ * @name_len:	length of the name in unicode characters
+ *
+ * @ni is always closed after using of this function (even if it failed),
+ * user do not need to call ntfs_inode_close himself.
+ *
+ * Return 0 on success or -1 on error with errno set to the error code.
+ */
+int ntfs_delete(ntfs_inode *ni, ntfs_inode *dir_ni, ntfschar *name, u8 name_len)
+{
+	ntfs_attr_search_ctx *actx = NULL;
+	ntfs_index_context *ictx = NULL;
+	FILE_NAME_ATTR *fn = NULL;
+	BOOL looking_for_dos_name = FALSE, looking_for_win32_name = FALSE;
+	int err = 0;
+
+	ntfs_debug("Entering.");
+	if (!ni || !dir_ni || !name || !name_len) {
+		ntfs_error(, "Invalid arguments.");
+		errno = EINVAL;
+		goto err_out;
+	}
+	if (ni->nr_extents == -1)
+		ni = ni->base_ni;
+	if (dir_ni->nr_extents == -1)
+		dir_ni = dir_ni->base_ni;
+	/* If deleting directory check it to be empty. */
+	if (ni->mrec->flags & MFT_RECORD_IS_DIRECTORY) {
+		ntfs_attr *na;
+
+		na = ntfs_attr_open(ni, AT_INDEX_ROOT, I30, 4);
+		if (!na) {
+			ntfs_error(, "Corrupt directory or library bug.");
+			errno = EIO;
+			goto err_out;
+		}
+		if (na->data_size != sizeof(INDEX_ROOT) +
+				sizeof(INDEX_ENTRY_HEADER)) {
+			ntfs_attr_close(na);
+			ntfs_error(, "Directory is not empty.");
+			errno = ENOTEMPTY;
+			goto err_out;
+		}
+		ntfs_attr_close(na);
+	}
+	/*
+	 * Search for FILE_NAME attribute with such name. If it's in POSIX or
+	 * WIN32_AND_DOS namespace, then simply remove it from index and inode.
+	 * If filename in DOS or in WIN32 namespace, then remove DOS name first,
+	 * only then remove WIN32 name. Mark WIN32 name as POSIX name to prevent
+	 * chkdsk to complain about DOS name absentation, in case if DOS name
+	 * had been successfully deleted, but WIN32 name removing failed.
+	 */
+	actx = ntfs_attr_get_search_ctx(ni, NULL);
+	if (!actx)
+		goto err_out;
+search:
+	while (!ntfs_attr_lookup(AT_FILE_NAME, AT_UNNAMED, 0, CASE_SENSITIVE,
+			0, NULL, 0, actx)) {
+		errno = 0;			
+		fn = (FILE_NAME_ATTR*)((u8*)actx->attr +
+				le16_to_cpu(actx->attr->value_offset));
+		if (looking_for_dos_name) {
+			if (fn->file_name_type == FILE_NAME_DOS)
+				break;
+			else
+				continue;
+		}
+		if (looking_for_win32_name) {
+			if  (fn->file_name_type == FILE_NAME_WIN32)
+				break;
+			else
+				continue;
+		}
+		if (dir_ni->mft_no == MREF_LE(fn->parent_directory) &&
+				ntfs_names_are_equal(fn->file_name,
+				fn->file_name_length, name,
+				name_len, (fn->file_name_type ==
+				FILE_NAME_POSIX) ? CASE_SENSITIVE : IGNORE_CASE,
+				ni->vol->upcase, ni->vol->upcase_len)) {
+			if (fn->file_name_type == FILE_NAME_WIN32) {
+				looking_for_dos_name = TRUE;
+				ntfs_attr_reinit_search_ctx(actx);
+				continue;
+			}
+			if (fn->file_name_type == FILE_NAME_DOS)
+				looking_for_dos_name = TRUE;
+			break;
+		}
+	}
+	if (errno)
+		goto err_out;
+	/* Search for such FILE_NAME in index. */
+	ictx = ntfs_index_ctx_get(dir_ni, I30, 4);
+	if (!ictx)
+		goto err_out;
+	if (ntfs_index_lookup(fn, le32_to_cpu(actx->attr->value_length), ictx))
+		goto err_out;
+	/* Set namespace to POSIX for WIN32 name. */
+	if (fn->file_name_type == FILE_NAME_WIN32) {
+		fn->file_name_type = FILE_NAME_POSIX;
+		ntfs_inode_mark_dirty(actx->ntfs_ino);
+		((FILE_NAME_ATTR*)ictx->data)->file_name_type = FILE_NAME_POSIX;
+		ntfs_index_entry_mark_dirty(ictx);
+	}
+	/* Do not support reparse point deletion yet. */
+	if (((FILE_NAME_ATTR*)ictx->data)->file_attributes &
+			FILE_ATTR_REPARSE_POINT) {
+		errno = EOPNOTSUPP;
+		goto err_out;
+	}
+	/* Remove FILE_NAME from index. */
+	if (ntfs_index_rm(ictx))
+		goto err_out;
+	ictx = NULL;
+	/* Remove FILE_NAME from inode. */
+	if (ntfs_attr_record_rm(actx))
+		goto err_out;
+	/* Decerement hard link count. */
+	ni->mrec->link_count = cpu_to_le16(le16_to_cpu(
+			ni->mrec->link_count) - 1);
+	ntfs_inode_mark_dirty(ni);
+	if (looking_for_dos_name) {
+		looking_for_dos_name = FALSE;
+		looking_for_win32_name = TRUE;
+		ntfs_attr_reinit_search_ctx(actx);
+		goto search;
+	}
+	/*
+	 * If hard link count is not equal to zero then we are done. In other
+	 * case there are no reference to this inode left, so we should free all
+	 * non-resident atributes and mark inode as not in use. 
+	 */
+	if (ni->mrec->link_count)
+		goto out;
+	ntfs_attr_reinit_search_ctx(actx);
+	while (!ntfs_attrs_walk(actx)) {
+		if (actx->attr->non_resident) {
+			runlist *rl;
+
+			rl = ntfs_mapping_pairs_decompress(ni->vol, actx->attr,
+					NULL);
+			if (!rl) {
+				err = errno;
+				ntfs_error(, "Failed to decompress runlist.  "
+						"Leaving inconsist metadata.");
+				continue;
+			}
+			if (ntfs_cluster_free_from_rl(ni->vol, rl)) {
+				err = errno;
+				ntfs_error(, "Failed to free clusters.  "
+						"Leaving inconsist metadata.");
+				continue;
+			}
+			free(rl);
+		}
+	}
+	if (errno != ENOENT) {
+		err = errno;
+		ntfs_error(, "Atribute enumeration failed.  "
+				"Probably leaving inconsist metadata.");
+	}
+	/* All extents should be attached after attribute walk. */
+	while (ni->nr_extents)
+		if (ntfs_mft_record_free(ni->vol, *(ni->extent_nis))) {
+			err = errno;
+			ntfs_error(, "Failed to free extent MFT record.  "
+					"Leaving inconsist metadata.");
+		}
+	if (ntfs_mft_record_free(ni->vol, ni)) {
+		err = errno;
+		ntfs_error(, "Failed to free base MFT record.  "
+				"Leaving inconsist metadata.");
+	}
+	ni = NULL;
+out:
+	if (actx)
+		ntfs_attr_put_search_ctx(actx);
+	if (ictx)
+		ntfs_index_ctx_put(ictx);
+	if (ni)
+		ntfs_inode_close(ni);
+	if (err) {
+		ntfs_error(, "Failed.");
+		errno = err;
+		return -1;
+	}
+	ntfs_debug("Done.");
+	return 0;
+err_out:
+	err = errno;
+	goto out;
+}
+
+/**
+ * ntfs_link - create hard link for file or directory
+ * @ni:		ntfs inode for object to create hard link
+ * @dir_ni:	ntfs inode for directory in which new link should be placed
+ * @name:	unicode name of the new link
+ * @name_len:	length of the name in unicode characters
+ *
+ * Return 0 on success or -1 on error with errno set to the error code.
+ */
+int ntfs_link(ntfs_inode *ni, ntfs_inode *dir_ni, ntfschar *name, u8 name_len)
+{
+	FILE_NAME_ATTR *fn = NULL;
+	int fn_len, err;
+
+	ntfs_debug("Entering.");
+	if (!ni || !dir_ni || !name || !name_len) {
+		err = errno;
+		ntfs_error(, "Invalid arguments.");
+		goto err_out;
+	}
+	/* Create FILE_NAME attribute. */
+	fn_len = sizeof(FILE_NAME_ATTR) + name_len * sizeof(ntfschar);
+	fn = calloc(1, fn_len);
+	if (!fn) {
+		err = errno;
+		ntfs_error(, "Not enough memory.");
+		goto err_out;
+	}
+	fn->parent_directory = MK_LE_MREF(dir_ni->mft_no,
+			le16_to_cpu(dir_ni->mrec->sequence_number));
+	fn->file_name_length = name_len;
+	fn->file_name_type = FILE_NAME_POSIX;
+	if (ni->mrec->flags & MFT_RECORD_IS_DIRECTORY)
+		fn->file_attributes = FILE_ATTR_DUP_FILE_NAME_INDEX_PRESENT;
+	fn->creation_time = utc2ntfs(ni->creation_time);
+	fn->last_data_change_time = utc2ntfs(ni->last_data_change_time);
+	fn->last_mft_change_time = utc2ntfs(ni->last_mft_change_time);
+	fn->last_access_time = utc2ntfs(ni->last_access_time);
+	memcpy(fn->file_name, name, name_len * sizeof(ntfschar));
+	/* Add FILE_NAME attribute to index. */
+	if (ntfs_index_add_filename(dir_ni, fn, MK_MREF(ni->mft_no,
+			le16_to_cpu(ni->mrec->sequence_number)))) {
+		err = errno;
+		ntfs_error(, "Failed to add entry to the index.");
+		goto err_out;
+	}
+	/* Add FILE_NAME attribute to inode. */
+	if (ntfs_attr_add(ni, AT_FILE_NAME, AT_UNNAMED, 0, (u8*)fn, fn_len)) {
+		ntfs_index_context *ictx;
+
+		err = errno;
+		ntfs_error(, "Failed to add FILE_NAME attribute.");
+		/* Try to remove just added attribute from index. */
+		ictx = ntfs_index_ctx_get(dir_ni, I30, 4);
+		if (!ictx)
+			goto rollback_failed;
+		if (ntfs_index_lookup(fn, fn_len, ictx)) {
+			ntfs_index_ctx_put(ictx);
+			goto rollback_failed;
+		}
+		if (ntfs_index_rm(ictx)) {
+			ntfs_index_ctx_put(ictx);
+			goto rollback_failed;
+		}
+		goto err_out;
+	}
+	/* Increment hard links count. */
+	ni->mrec->link_count = cpu_to_le16(le16_to_cpu(
+			ni->mrec->link_count) + 1);
+	/*
+	 * Do not set attributes and file size, instead of this mark filenames
+	 * dirty to force attribute and size update during sync. 
+	 * NOTE: File size may will be not updated and not all attributes will
+	 * be set, but it is acceptable since windows driver does not update
+	 * all file names when one of the hard links changed.
+	 * FIXME: It will be nice to update them all.
+	 */
+	NInoFileNameSetDirty(ni);
+	/* Done! */
+	ntfs_inode_mark_dirty(ni);
+	free(fn);
+	ntfs_debug("Done.");
+	return 0;
+rollback_failed:
+	ntfs_error(, "Rollback failed. Leaving inconsist metadata.");
+err_out:
+	ntfs_error(, "Failed.");
+	if (fn)
+		free(fn);
+	errno = err;
+	return -1;
+}
