@@ -3,8 +3,8 @@
  *			libntfs. Part of the Linux-NTFS project.
  *
  * Copyright (c) 2003 Jan Kratochvil <project-captive@jankratochvil.net>
- * Copyright (c) 2003 Anton Altaparmakov
- * 
+ * Copyright (c) 2003-2005 Anton Altaparmakov
+ *
  * This program/include file is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as published
  * by the Free Software Foundation; either version 2 of the License, or
@@ -34,7 +34,9 @@
 #include <glib/gmessages.h>
 #include "gnome-vfs-module.h"
 #include <glib/ghash.h>
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
 #include <libgnomevfs/gnome-vfs-utils.h>
 
 #include "volume.h"
@@ -194,7 +196,7 @@ static GnomeVFSResult inode_open_by_pathname(ntfs_inode **inode_return,
 	strcpy(pathname_parse, pathname);
 	mref = FILE_root;
 	for (;;) {
-		uchar_t *pathname_parse_ucs2;
+		ntfschar *pathname_parse_ucs2;
 		gchar *pathname_parse_unescaped;
 		int i;
 
@@ -247,7 +249,8 @@ struct libntfs_directory {
 
 static GnomeVFSResult libntfs_gnomevfs_open_directory(GnomeVFSMethod *method,
 		GnomeVFSMethodHandle **method_handle, GnomeVFSURI *uri,
-		GnomeVFSFileInfoOptions options, GnomeVFSContext *context)
+		GnomeVFSFileInfoOptions options __attribute__((unused)),
+		GnomeVFSContext *context __attribute__((unused)))
 {
 	GnomeVFSResult errvfsresult;
 	ntfs_volume *volume;
@@ -276,7 +279,8 @@ static GnomeVFSResult libntfs_gnomevfs_open_directory(GnomeVFSMethod *method,
 }
 
 static GnomeVFSResult libntfs_gnomevfs_close_directory(GnomeVFSMethod *method,
-		GnomeVFSMethodHandle *method_handle, GnomeVFSContext *context)
+		GnomeVFSMethodHandle *method_handle,
+		GnomeVFSContext *context __attribute__((unused)))
 {
 	struct libntfs_directory *libntfs_directory;
 	int errint;
@@ -314,7 +318,7 @@ static GnomeVFSResult libntfs_gnomevfs_close_directory(GnomeVFSMethod *method,
 	return GNOME_VFS_OK;
 }
 
-static gchar *libntfs_uchar_to_utf8(const uchar_t *name, const int name_len)
+static gchar *libntfs_ntfscharo_utf8(const ntfschar *name, const int name_len)
 {
 	GString *gstring;
 	int i;
@@ -331,7 +335,8 @@ static gchar *libntfs_uchar_to_utf8(const uchar_t *name, const int name_len)
  */
 static int libntfs_gnomevfs_read_directory_filldir(
 		struct libntfs_directory *libntfs_directory /* dirent */,
-		const uchar_t *name, const int name_len, const int name_type,
+		const ntfschar *name, const int name_len,
+		const int name_type __attribute__((unused)),
 		const s64 pos, const MFT_REF mref, const unsigned dt_type)
 {
 	GnomeVFSFileInfo *file_info;
@@ -346,7 +351,7 @@ static int libntfs_gnomevfs_read_directory_filldir(
 		return 0;	/* continue traversal */
 
 	file_info = gnome_vfs_file_info_new();
-	file_info->name = libntfs_uchar_to_utf8(name, name_len);
+	file_info->name = libntfs_ntfscharo_utf8(name, name_len);
 	file_info->valid_fields = 0;
 
 	switch (dt_type) {
@@ -390,8 +395,8 @@ static int libntfs_gnomevfs_read_directory_filldir(
 
 			attr = ntfs_attr_open(inode,	/* ni */
 					AT_DATA,	/* type */
-					NULL,	/* name */
-					0);	/* name_len */
+					AT_UNNAMED,	/* name */
+					0);		/* name_len */
 			/* FIXME: Check failed 'attr' open. */
 			if (attr) {
 				/* FIXME: Is 'data_size' the right field? */
@@ -413,7 +418,8 @@ static int libntfs_gnomevfs_read_directory_filldir(
 
 static GnomeVFSResult libntfs_gnomevfs_read_directory(GnomeVFSMethod *method,
 		GnomeVFSMethodHandle *method_handle,
-		GnomeVFSFileInfo *file_info, GnomeVFSContext *context)
+		GnomeVFSFileInfo *file_info,
+		GnomeVFSContext *context __attribute__((unused)))
 {
 	GnomeVFSResult errvfsresult;
 	struct libntfs_directory *libntfs_directory;
@@ -484,7 +490,7 @@ static GnomeVFSResult libntfs_open_attr(struct libntfs_file *libntfs_file)
 		libntfs_file->attr = ntfs_attr_open(
 				libntfs_file->inode,	/* ni */
 				AT_DATA,	/* type */
-				NULL,	/* name */
+				AT_UNNAMED,	/* name */
 				0);	/* name_len */
 		G_UNLOCK(libntfs);
 		if (!libntfs_file->attr)
@@ -497,7 +503,8 @@ static GnomeVFSResult libntfs_open_attr(struct libntfs_file *libntfs_file)
 
 static GnomeVFSResult libntfs_gnomevfs_open(GnomeVFSMethod *method,
 		GnomeVFSMethodHandle **method_handle_return, GnomeVFSURI *uri,
-		GnomeVFSOpenMode mode, GnomeVFSContext *context)
+		GnomeVFSOpenMode mode,
+		GnomeVFSContext *context __attribute__((unused)))
 {
 	GnomeVFSResult errvfsresult;
 	ntfs_volume *volume;
@@ -530,8 +537,10 @@ static GnomeVFSResult libntfs_gnomevfs_open(GnomeVFSMethod *method,
 
 static GnomeVFSResult libntfs_gnomevfs_create(GnomeVFSMethod *method,
 		GnomeVFSMethodHandle **method_handle_return, GnomeVFSURI *uri,
-		GnomeVFSOpenMode mode, gboolean exclusive, guint perm,
-		GnomeVFSContext *context)
+		GnomeVFSOpenMode mode __attribute__((unused)),
+		gboolean exclusive __attribute__((unused)),
+		guint perm __attribute__((unused)),
+		GnomeVFSContext *context __attribute__((unused)))
 {
 	GnomeVFSResult errvfsresult;
 	ntfs_volume *volume;
@@ -549,7 +558,8 @@ static GnomeVFSResult libntfs_gnomevfs_create(GnomeVFSMethod *method,
 }
 
 static GnomeVFSResult libntfs_gnomevfs_close(GnomeVFSMethod *method,
-		GnomeVFSMethodHandle *method_handle, GnomeVFSContext *context)
+		GnomeVFSMethodHandle *method_handle,
+		GnomeVFSContext *context __attribute__((unused)))
 {
 	struct libntfs_file *libntfs_file;
 	int errint;
@@ -579,7 +589,7 @@ static GnomeVFSResult libntfs_gnomevfs_close(GnomeVFSMethod *method,
 static GnomeVFSResult libntfs_gnomevfs_read(GnomeVFSMethod *method,
 		GnomeVFSMethodHandle *method_handle, gpointer buffer,
 		GnomeVFSFileSize num_bytes, GnomeVFSFileSize *bytes_read_return,
-		GnomeVFSContext *context)
+		GnomeVFSContext *context __attribute__((unused)))
 {
 	GnomeVFSResult errvfsresult;
 	struct libntfs_file *libntfs_file;
@@ -616,7 +626,7 @@ static GnomeVFSResult libntfs_gnomevfs_read(GnomeVFSMethod *method,
 static GnomeVFSResult libntfs_gnomevfs_seek(GnomeVFSMethod *method,
 		GnomeVFSMethodHandle *method_handle,
 		GnomeVFSSeekPosition whence, GnomeVFSFileOffset offset,
-		GnomeVFSContext *context)
+		GnomeVFSContext *context __attribute__((unused)))
 {
 	GnomeVFSResult errvfsresult;
 	struct libntfs_file *libntfs_file;
@@ -681,10 +691,11 @@ static gboolean libntfs_gnomevfs_is_local(GnomeVFSMethod *method,
 	return gnome_vfs_uri_is_local(uri->parent);
 }
 
-GnomeVFSResult libntfs_gnomevfs_get_file_info_from_handle(
+static GnomeVFSResult libntfs_gnomevfs_get_file_info_from_handle(
 		GnomeVFSMethod *method, GnomeVFSMethodHandle *method_handle,
-		GnomeVFSFileInfo *file_info, GnomeVFSFileInfoOptions options,
-		GnomeVFSContext *context)
+		GnomeVFSFileInfo *file_info,
+		GnomeVFSFileInfoOptions options __attribute__((unused)),
+		GnomeVFSContext *context __attribute__((unused)))
 {
 	GnomeVFSResult errvfsresult;
 	struct libntfs_file *libntfs_file;
@@ -761,9 +772,9 @@ static GnomeVFSResult libntfs_gnomevfs_get_file_info(GnomeVFSMethod *method,
 	return GNOME_VFS_OK;
 }
 
-GnomeVFSResult libntfs_gnomevfs_check_same_fs(GnomeVFSMethod *method,
+static GnomeVFSResult libntfs_gnomevfs_check_same_fs(GnomeVFSMethod *method,
 		GnomeVFSURI *a, GnomeVFSURI *b, gboolean *same_fs_return,
-		GnomeVFSContext *context)
+		GnomeVFSContext *context __attribute__((unused)))
 {
 	ntfs_volume *volume_a;
 	ntfs_volume *volume_b;
