@@ -22,9 +22,15 @@
 
 #include "config.h"
 
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_ERRNO_H
 #include <errno.h>
+#endif
 
 #include "types.h"
 #include "attrib.h"
@@ -37,10 +43,10 @@
 /**
  * ntfs_cluster_alloc - allocate clusters on an ntfs volume
  * @vol:	mounted ntfs volume on which to allocate the clusters
+ * @start_vcn:	vcn to use for the first allocated cluster
  * @count:	number of clusters to allocate
  * @start_lcn:	starting lcn at which to allocate the clusters (or -1 if none)
  * @zone:	zone from which to allocate the clusters
- * @start_vcn:	
  *
  * Allocate @count clusters preferably starting at cluster @start_lcn or at the
  * current allocator position if @start_lcn is -1, on the mounted ntfs volume
@@ -90,8 +96,8 @@
  * possible code paths. So at least for now, I am leaving the double logic -
  * better safe than sorry... (AIA)
  */
-runlist *ntfs_cluster_alloc(ntfs_volume *vol, s64 count, LCN start_lcn,
-		const NTFS_CLUSTER_ALLOCATION_ZONES zone, VCN start_vcn)
+runlist *ntfs_cluster_alloc(ntfs_volume *vol, VCN start_vcn, s64 count,
+		LCN start_lcn, const NTFS_CLUSTER_ALLOCATION_ZONES zone)
 {
 	LCN zone_start, zone_end, bmp_pos, bmp_initial_pos, last_read_pos, lcn;
 	LCN prev_lcn = 0, prev_run_len = 0, mft_zone_size;
@@ -111,7 +117,7 @@ runlist *ntfs_cluster_alloc(ntfs_volume *vol, s64 count, LCN start_lcn,
 		errno = EINVAL;
 		return NULL;
 	}
-	
+
 	/* Return empty runlist if @count == 0 */
 	if (!count) {
 		rl = malloc(0x1000);
@@ -429,7 +435,7 @@ runlist *ntfs_cluster_alloc(ntfs_volume *vol, s64 count, LCN start_lcn,
 					if (rl)
 						free(rl);
 					free(buf);
-					NTFS_BUG("switch(search_zone)");
+					NTFS_BUG("switch(search_zone) 1");
 					return NULL;
 				}
 				Dprintf("%s(): Going to done_ret.\n",
@@ -490,7 +496,7 @@ zone_pass_done:	/* Finished with the current zone pass. */
 				zone_start = 0;
 				break;
 			default:
-				NTFS_BUG("switch(search_zone), 2");
+				NTFS_BUG("switch(search_zone) 2");
 			}
 			/* Sanity check. */
 			if (zone_end < zone_start)
