@@ -1,8 +1,8 @@
-/*
+/**
  * runlist.c - Run list handling code. Part of the Linux-NTFS project.
  *
- * Copyright (c) 2002-2004 Anton Altaparmakov
- * Copyright (c) 2002 Richard Russon
+ * Copyright (c) 2002-2005 Anton Altaparmakov
+ * Copyright (c) 2002-2005 Richard Russon
  * Copyright (c) 2004 Yura Pakhuchiy
  *
  * This program/include file is free software; you can redistribute it and/or
@@ -23,10 +23,18 @@
 
 #include "config.h"
 
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_ERRNO_H
 #include <errno.h>
+#endif
 
 #include "compat.h"
 
@@ -52,7 +60,7 @@ static __inline__ void ntfs_rl_mm(runlist_element *base, int dst, int src,
 /**
  * Internal:
  *
- * rl_mc - runlist memory copy
+ * ntfs_rl_mc - runlist memory copy
  */
 static __inline__ void ntfs_rl_mc(runlist_element *dstbase, int dst,
 		runlist_element *srcbase, int src, int size)
@@ -111,14 +119,21 @@ static __inline__ BOOL ntfs_rl_are_mergeable(runlist_element *dst,
 		return FALSE;
 	}
 
-	if ((dst->lcn < 0) || (src->lcn < 0))     /* Are we merging holes? */
+	/* We can merge unmapped regions even if they are misaligned. */
+	if ((dst->lcn == LCN_RL_NOT_MAPPED) && (src->lcn == LCN_RL_NOT_MAPPED))
+		return TRUE;
+	/* If the runs are misaligned, we cannot merge them. */
+	if ((dst->vcn + dst->length) != src->vcn)
 		return FALSE;
-	if ((dst->lcn + dst->length) != src->lcn) /* Are the runs contiguous? */
-		return FALSE;
-	if ((dst->vcn + dst->length) != src->vcn) /* Are the runs misaligned? */
-		return FALSE;
-
-	return TRUE;
+	/* If both runs are non-sparse and contiguous, we can merge them. */
+	if ((dst->lcn >= 0) && (src->lcn >= 0) &&
+		((dst->lcn + dst->length) == src->lcn))
+		return TRUE;
+	/* If we are merging two holes, we can merge them. */
+	if ((dst->lcn == LCN_HOLE) && (src->lcn == LCN_HOLE))
+		return TRUE;
+	/* Cannot merge. */
+	return FALSE;
 }
 
 /**
@@ -138,31 +153,6 @@ static __inline__ void __ntfs_rl_merge(runlist_element *dst,
 	dst->length += src->length;
 }
 
-/**
- * Internal:
- *
- * ntfs_rl_merge - test if two runlists can be joined together and merge them
- * @dst:	original, destination runlist
- * @src:	new runlist to merge with @dst
- *
- * Test if two runlists can be joined together. For this, their VCNs and LCNs
- * must be adjacent. If they can be merged, perform the merge, writing into
- * the destination runlist @dst.
- *
- * Return: TRUE   Success, the runlists have been merged.
- *	   FALSE  Failure, the runlists cannot be merged and have not been
- *		  modified.
- */
-static __inline__ BOOL ntfs_rl_merge(runlist_element *dst,
-		runlist_element *src)
-{
-	BOOL merge = ntfs_rl_are_mergeable(dst, src);
-
-	if (merge)
-		__ntfs_rl_merge(dst, src);
-	return merge;
-}
-
 /**
  * Internal:
  *
@@ -188,8 +178,8 @@ static __inline__ BOOL ntfs_rl_merge(runlist_element *dst,
 static __inline__ runlist_element *ntfs_rl_append(runlist_element *dst,
 		int dsize, runlist_element *src, int ssize, int loc)
 {
-	BOOL right;
-	int magic;
+	BOOL right = FALSE;	/* Right end of @src needs merging */
+	int marker;		/* End of the inserted runs */
 
 	if (!dst || !src) {
 		Dputs("Eeek. ntfs_rl_append() invoked with NULL pointer!");
@@ -198,12 +188,13 @@ static __inline__ runlist_element *ntfs_rl_append(runlist_element *dst,
 	}
 
 	/* First, check if the right hand end needs merging. */
-	right = ntfs_rl_are_mergeable(src + ssize - 1, dst + loc + 1);
+	if ((loc + 1) < dsize)
+		right = ntfs_rl_are_mergeable(src + ssize - 1, dst + loc + 1);
 
 	/* Space required: @dst size + @src size, less one if we merged. */
 	dst = ntfs_rl_realloc(dst, dsize, dsize + ssize - right);
 	if (!dst)
-		return dst;
+		return NULL;
 	/*
 	 * We are guaranteed to succeed from here so can start modifying the
 	 * original runlists.
@@ -213,19 +204,19 @@ static __inline__ runlist_element *ntfs_rl_append(runlist_element *dst,
 	if (right)
 		__ntfs_rl_merge(src + ssize - 1, dst + loc + 1);
 
-	/* FIXME: What does this mean? (AIA) */
-	magic = loc + ssize;
+	/* marker - First run after the @src runs that have been inserted */
+	marker = loc + ssize + 1;
 
 	/* Move the tail of @dst out of the way, then copy in @src. */
-	ntfs_rl_mm(dst, magic + 1, loc + 1 + right, dsize - loc - 1 - right);
+	ntfs_rl_mm(dst, marker, loc + 1 + right, dsize - loc - 1 - right);
 	ntfs_rl_mc(dst, loc + 1, src, 0, ssize);
 
 	/* Adjust the size of the preceding hole. */
 	dst[loc].length = dst[loc + 1].vcn - dst[loc].vcn;
 
 	/* We may have changed the length of the file, so fix the end marker */
-	if (dst[magic + 1].lcn == LCN_ENOENT)
-		dst[magic + 1].vcn = dst[magic].vcn + dst[magic].length;
+	if (dst[marker].lcn == LCN_ENOENT)
+		dst[marker].vcn = dst[marker-1].vcn + dst[marker-1].length;
 
 	return dst;
 }
@@ -255,10 +246,9 @@ static __inline__ runlist_element *ntfs_rl_append(runlist_element *dst,
 static __inline__ runlist_element *ntfs_rl_insert(runlist_element *dst,
 		int dsize, runlist_element *src, int ssize, int loc)
 {
-	BOOL left = FALSE;
-	BOOL disc = FALSE;	/* Discontinuity */
-	BOOL hole = FALSE;	/* Following a hole */
-	int magic;
+	BOOL left = FALSE;	/* Left end of @src needs merging */
+	BOOL disc = FALSE;	/* Discontinuity between @dst and @src */
+	int marker;		/* End of the inserted runs */
 
 	if (!dst || !src) {
 		Dputs("Eeek. ntfs_rl_insert() invoked with NULL pointer!");
@@ -267,9 +257,7 @@ static __inline__ runlist_element *ntfs_rl_insert(runlist_element *dst,
 	}
 
 	/* disc => Discontinuity between the end of @dst and the start of @src.
-	 *	   This means we might need to insert a hole.
-	 * hole => @dst ends with a hole or an unmapped region which we can
-	 *	   extend to match the discontinuity.
+	 *	   This means we might need to insert a "notmapped" run.
 	 */
 	if (loc == 0)
 		disc = (src[0].vcn > 0);
@@ -283,16 +271,14 @@ static __inline__ runlist_element *ntfs_rl_insert(runlist_element *dst,
 			merged_length += src->length;
 
 		disc = (src[0].vcn > dst[loc - 1].vcn + merged_length);
-		if (disc)
-			hole = (dst[loc - 1].lcn == LCN_HOLE);
 	}
 
 	/* Space required: @dst size + @src size, less one if we merged, plus
-	 * one if there was a discontinuity, less one for a trailing hole.
+	 * one if there was a discontinuity.
 	 */
-	dst = ntfs_rl_realloc(dst, dsize, dsize + ssize - left + disc - hole);
+	dst = ntfs_rl_realloc(dst, dsize, dsize + ssize - left + disc);
 	if (!dst)
-		return dst;
+		return NULL;
 	/*
 	 * We are guaranteed to succeed from here so can start modifying the
 	 * original runlist.
@@ -301,42 +287,34 @@ static __inline__ runlist_element *ntfs_rl_insert(runlist_element *dst,
 	if (left)
 		__ntfs_rl_merge(dst + loc - 1, src);
 
-	/* FIXME: What does this mean? (AIA) */
-	magic = loc + ssize - left + disc - hole;
+	/*
+	 * marker - First run after the @src runs that have been inserted
+	 * Nominally: marker = @loc + @ssize (location + number of runs in @src)
+	 * If "left", then the first run in @src has been merged with one in @dst.
+	 * If "disc", then @dst and @src don't meet and we need an extra run to fill the gap.
+	 */
+	marker = loc + ssize - left + disc;
 
 	/* Move the tail of @dst out of the way, then copy in @src. */
-	ntfs_rl_mm(dst, magic, loc, dsize - loc);
-	ntfs_rl_mc(dst, loc + disc - hole, src, left, ssize - left);
+	ntfs_rl_mm(dst, marker, loc, dsize - loc);
+	ntfs_rl_mc(dst, loc + disc, src, left, ssize - left);
 
-	/* Adjust the VCN of the last run ... */
-	if (dst[magic].lcn <= LCN_HOLE)
-		dst[magic].vcn = dst[magic - 1].vcn + dst[magic - 1].length;
+	/* Adjust the VCN of the first run after the insertion ... */
+	dst[marker].vcn = dst[marker - 1].vcn + dst[marker - 1].length;
 	/* ... and the length. */
-	if (dst[magic].lcn == LCN_HOLE || dst[magic].lcn == LCN_RL_NOT_MAPPED)
-		dst[magic].length = dst[magic + 1].vcn - dst[magic].vcn;
+	if (dst[marker].lcn == LCN_HOLE || dst[marker].lcn == LCN_RL_NOT_MAPPED)
+		dst[marker].length = dst[marker + 1].vcn - dst[marker].vcn;
 
 	/* Writing beyond the end of the file and there's a discontinuity. */
 	if (disc) {
-		if (hole)
-			dst[loc - 1].length = dst[loc].vcn - dst[loc - 1].vcn;
-		else {
-			if (loc > 0) {
-				dst[loc].vcn = dst[loc - 1].vcn +
-						dst[loc - 1].length;
-				dst[loc].length = dst[loc + 1].vcn -
-						dst[loc].vcn;
-			} else {
-				dst[loc].vcn = 0;
-				dst[loc].length = dst[loc + 1].vcn;
-			}
-			dst[loc].lcn = LCN_RL_NOT_MAPPED;
+		if (loc > 0) {
+			dst[loc].vcn = dst[loc - 1].vcn + dst[loc - 1].length;
+			dst[loc].length = dst[loc + 1].vcn - dst[loc].vcn;
+		} else {
+			dst[loc].vcn = 0;
+			dst[loc].length = dst[loc + 1].vcn;
 		}
-
-		magic += hole;
-
-		if (dst[magic].lcn == LCN_ENOENT)
-			dst[magic].vcn = dst[magic - 1].vcn +
-					dst[magic - 1].length;
+		dst[loc].lcn = LCN_RL_NOT_MAPPED;
 	}
 	return dst;
 }
@@ -365,9 +343,10 @@ static __inline__ runlist_element *ntfs_rl_insert(runlist_element *dst,
 static __inline__ runlist_element *ntfs_rl_replace(runlist_element *dst,
 		int dsize, runlist_element *src, int ssize, int loc)
 {
-	BOOL left = FALSE;
-	BOOL right;
-	int magic;
+	BOOL left  = FALSE;	/* Left end of @src needs merging */
+	BOOL right = FALSE;	/* Right end of @src needs merging */
+	int tail;		/* Start of tail of @dst */
+	int marker;		/* End of the inserted runs */
 
 	if (!dst || !src) {
 		Dputs("Eeek. ntfs_rl_replace() invoked with NULL pointer!");
@@ -375,36 +354,51 @@ static __inline__ runlist_element *ntfs_rl_replace(runlist_element *dst,
 		return NULL;
 	}
 
-	/* First, merge the left and right ends, if necessary. */
-	right = ntfs_rl_are_mergeable(src + ssize - 1, dst + loc + 1);
+	/* First, see if the left and right ends need merging. */
+	if ((loc + 1) < dsize)
+		right = ntfs_rl_are_mergeable(src + ssize - 1, dst + loc + 1);
 	if (loc > 0)
 		left = ntfs_rl_are_mergeable(dst + loc - 1, src);
 
 	/* Allocate some space. We'll need less if the left, right, or both
-	 * ends were merged.
+	 * ends get merged.
 	 */
 	dst = ntfs_rl_realloc(dst, dsize, dsize + ssize - left - right);
 	if (!dst)
-		return dst;
+		return NULL;
 	/*
 	 * We are guaranteed to succeed from here so can start modifying the
 	 * original runlists.
 	 */
+
+	/* First, merge the left and right ends, if necessary. */
 	if (right)
 		__ntfs_rl_merge(src + ssize - 1, dst + loc + 1);
 	if (left)
 		__ntfs_rl_merge(dst + loc - 1, src);
 
-	/* FIXME: What does this mean? (AIA) */
-	magic = loc + ssize - left;
+	/*
+	 * tail - Offset of the tail of @dst
+	 * Nominally: @tail = @loc + 1 (location, skipping the replaced run)
+	 * If "right", then one of @dst's runs is already merged into @src.
+	 */
+	tail = loc + right + 1;
+
+	/*
+	 * marker - First run after the @src runs that have been inserted
+	 * Nominally: @marker = @loc + @ssize (location + number of runs in @src)
+	 * If "left", then the first run in @src has been merged with one in @dst.
+	 */
+	marker = loc + ssize - left;
 
 	/* Move the tail of @dst out of the way, then copy in @src. */
-	ntfs_rl_mm(dst, magic, loc + right + 1, dsize - loc - right - 1);
+	ntfs_rl_mm(dst, marker, tail, dsize - tail);
 	ntfs_rl_mc(dst, loc, src, left, ssize - left);
 
 	/* We may have changed the length of the file, so fix the end marker */
-	if (dst[magic].lcn == LCN_ENOENT)
-		dst[magic].vcn = dst[magic - 1].vcn + dst[magic - 1].length;
+	if (((dsize - tail) > 0) && (dst[marker].lcn == LCN_ENOENT))
+		dst[marker].vcn = dst[marker - 1].vcn + dst[marker - 1].length;
+
 	return dst;
 }
 
@@ -522,6 +516,7 @@ runlist_element *ntfs_runlists_merge(runlist_element *drl,
 			/* Scan to the end of the source runlist. */
 			for (dend = 0; drl[dend].length; dend++)
 				;
+			dend++;
 			drl = ntfs_rl_realloc(drl, dend, dend + 1);
 			if (!drl)
 				return drl;
@@ -598,7 +593,7 @@ runlist_element *ntfs_runlists_merge(runlist_element *drl,
 		  (srl[send - 1].vcn + srl[send - 1].length)));
 
 	/* Or we'll lose an end marker */
-	if (start && finish && (drl[dins].length == 0))
+	if (finish && !drl[dins].length)
 		ss++;
 	if (marker && (drl[dins].vcn + drl[dins].length > srl[send - 1].vcn))
 		finish = FALSE;
@@ -743,7 +738,7 @@ runlist_element *ntfs_mapping_pairs_decompress(const ntfs_volume *vol,
 	u8 b;			/* Current byte offset in buf. */
 
 	Dprintf("%s(): Entering for attr 0x%x.\n", __FUNCTION__,
-			le32_to_cpu(attr->type));
+			(unsigned)le32_to_cpu(attr->type));
 	/* Make sure attr exists and is non-resident. */
 	if (!attr || !attr->non_resident ||
 			sle64_to_cpu(attr->lowest_vcn) < (VCN)0) {
@@ -1238,10 +1233,12 @@ __inline__ int ntfs_get_nr_significant_bytes(const s64 n)
  * ntfs_get_size_for_mapping_pairs - get bytes needed for mapping pairs array
  * @vol:	ntfs volume (needed for the ntfs version)
  * @rl:		runlist for which to determine the size of the mapping pairs
+ * @start_vcn:	vcn at which to start the mapping pairs array
  *
  * Walk the runlist @rl and calculate the size in bytes of the mapping pairs
- * array corresponding to the runlist @rl. This for example allows us to
- * allocate a buffer of the right size when building the mapping pairs array.
+ * array corresponding to the runlist @rl, starting at vcn @start_vcn.  This
+ * for example allows us to allocate a buffer of the right size when building
+ * the mapping pairs array.
  *
  * If @rl is NULL, just return 1 (for the single terminator byte).
  *
@@ -1249,41 +1246,90 @@ __inline__ int ntfs_get_nr_significant_bytes(const s64 n)
  * errno set to the error code.  The following error codes are defined:
  *	EINVAL	- Run list contains unmapped elements. Make sure to only pass
  *		  fully mapped runlists to this function.
+ *		- @start_vcn is invalid.
  *	EIO	- The runlist is corrupt.
  */
 int ntfs_get_size_for_mapping_pairs(const ntfs_volume *vol,
-		const runlist_element *rl)
+		const runlist_element *rl, const VCN start_vcn)
 {
 	LCN prev_lcn;
-	int i, rls;
+	int rls;
 
-	if (!rl)
+	if (start_vcn < 0) {
+		Dprintf("%s(): start_vcn %lld (should be >= 0)",
+			__FUNCTION__, (long long) start_vcn);
+		errno = EINVAL;
+		return -1;
+	}
+	if (!rl) {
+		if (start_vcn) {
+			Dprintf("%s(): rl NULL, start_vcn %lld (should be > 0)",
+				__FUNCTION__, (long long) start_vcn);
+			errno = EINVAL;
+			return -1;
+		}
 		return 1;
-	/* Always need the termining zero byte. */
+	}
+	/* Skip to runlist element containing @start_vcn. */
+	while (rl->length && start_vcn >= rl[1].vcn)
+		rl++;
+	if ((!rl->length && start_vcn > rl->vcn) || start_vcn < rl->vcn) {
+		errno = EINVAL;
+		return -1;
+	}
+	prev_lcn = 0;
+	/* Always need the terminating zero byte. */
 	rls = 1;
-	for (prev_lcn = i = 0; rl[i].length; i++) {
-		if (rl[i].length < 0 || rl[i].lcn < LCN_HOLE)
+	/* Do the first partial run if present. */
+	if (start_vcn > rl->vcn) {
+		s64 delta;
+
+		/* We know rl->length != 0 already. */
+		if (rl->length < 0 || rl->lcn < LCN_HOLE)
 			goto err_out;
+		delta = start_vcn - rl->vcn;
 		/* Header byte + length. */
-		rls += 1 + ntfs_get_nr_significant_bytes(rl[i].length);
+		rls += 1 + ntfs_get_nr_significant_bytes(rl->length - delta);
 		/*
 		 * If the logical cluster number (lcn) denotes a hole and we
 		 * are on NTFS 3.0+, we don't store it at all, i.e. we need
 		 * zero space. On earlier NTFS versions we just store the lcn.
+		 * Note: this assumes that on NTFS 1.2-, holes are stored with
+		 * an lcn of -1 and not a delta_lcn of -1 (unless both are -1).
 		 */
-		if (rl[i].lcn == LCN_HOLE && vol->major_ver >= 3)
-			continue;
+		if (rl->lcn >= 0 || vol->major_ver < 3) {
+			prev_lcn = rl->lcn;
+			if (rl->lcn >= 0)
+				prev_lcn += delta;
+			/* Change in lcn. */
+			rls += ntfs_get_nr_significant_bytes(prev_lcn);
+		}
+		/* Go to next runlist element. */
+		rl++;
+	}
+	/* Do the full runs. */
+	for (; rl->length; rl++) {
+		if (rl->length < 0 || rl->lcn < LCN_HOLE)
+			goto err_out;
+		/* Header byte + length. */
+		rls += 1 + ntfs_get_nr_significant_bytes(rl->length);
 		/*
-		 * Change in lcn.  Note: this assumes that on NTFS 1.2-, holes
-		 * are stored with an lcn of -1 and _not_ a delta_lcn of -1
-		 * (unless both are -1).
+		 * If the logical cluster number (lcn) denotes a hole and we
+		 * are on NTFS 3.0+, we don't store it at all, i.e. we need
+		 * zero space. On earlier NTFS versions we just store the lcn.
+		 * Note: this assumes that on NTFS 1.2-, holes are stored with
+		 * an lcn of -1 and not a delta_lcn of -1 (unless both are -1).
 		 */
-		rls += ntfs_get_nr_significant_bytes(rl[i].lcn - prev_lcn);
-		prev_lcn = rl[i].lcn;
+		if (rl->lcn >= 0 || vol->major_ver < 3) {
+			/* Change in lcn. */
+			rls += ntfs_get_nr_significant_bytes(rl->lcn -
+					prev_lcn);
+			prev_lcn = rl->lcn;
+		}
 	}
 	return rls;
 err_out:
-	if (rl[i].lcn == LCN_RL_NOT_MAPPED)
+	if (rl->lcn == LCN_RL_NOT_MAPPED)
 		errno = EINVAL;
 	else
 		errno = EIO;
@@ -1302,13 +1348,13 @@ err_out:
  * write.
  *
  * This is used when building the mapping pairs array of a runlist to compress
- * a given logical cluster number (lcn) or a specific run length to the minumum
+ * a given logical cluster number (lcn) or a specific run length to the minimum
  * size possible.
  *
  * Return the number of bytes written on success. On error, i.e. the
  * destination buffer @dst is too small, return -1 with errno set ENOSPC.
  */
-__inline__ int ntfs_write_significant_bytes(s8 *dst, const s8 *dst_max,
+__inline__ int ntfs_write_significant_bytes(u8 *dst, const u8 *dst_max,
 		const s64 n)
 {
 	s64 l = n;
@@ -1329,12 +1375,12 @@ __inline__ int ntfs_write_significant_bytes(s8 *dst, const s8 *dst_max,
 		if (dst > dst_max)
 			goto err_out;
 		i++;
-		*dst = (s8)-1;
+		*dst = (u8)-1;
 	} else if (n > 0LL && j < 0) {
 		if (dst > dst_max)
 			goto err_out;
 		i++;
-		*dst = (s8)0;
+		*dst = 0;
 	}
 	return i;
 err_out:
@@ -1348,46 +1394,115 @@ err_out:
  * @dst:	destination buffer to which to write the mapping pairs array
  * @dst_len:	size of destination buffer @dst in bytes
  * @rl:		runlist for which to build the mapping pairs array
+ * @start_vcn:	vcn at which to start the mapping pairs array
+ * @stop_vcn:	first vcn outside destination buffer on success or ENOSPC error
  *
- * Create the mapping pairs array from the runlist @rl and save the array in
- * @dst. @dst_len is the size of @dst in bytes and it should be at least equal
- * to the value obtained by calling ntfs_get_size_for_mapping_pairs().
+ * Create the mapping pairs array from the runlist @rl, starting at vcn
+ * @start_vcn and save the array in @dst.  @dst_len is the size of @dst in
+ * bytes and it should be at least equal to the value obtained by calling
+ * ntfs_get_size_for_mapping_pairs().
  *
  * If @rl is NULL, just write a single terminator byte to @dst.
  *
+ * On success or ENOSPC error, if @stop_vcn is not NULL, *@stop_vcn is set to
+ * the first vcn outside the destination buffer. Note that on error @dst has
+ * been filled with all the mapping pairs that will fit, thus it can be treated
+ * as partial success, in that a new attribute extent needs to be created or the
+ * next extent has to be used and the mapping pairs build has to be continued
+ * with @start_vcn set to *@stop_vcn.
+ *
  * Return 0 on success.  On error, return -1 with errno set to the error code.
  * The following error codes are defined:
  *	EINVAL	- Run list contains unmapped elements. Make sure to only pass
  *		  fully mapped runlists to this function.
+ *		- @start_vcn is invalid.
  *	EIO	- The runlist is corrupt.
  *	ENOSPC	- The destination buffer is too small.
  */
-int ntfs_mapping_pairs_build(const ntfs_volume *vol, s8 *dst,
-		const int dst_len, const runlist_element *rl)
+int ntfs_mapping_pairs_build(const ntfs_volume *vol, u8 *dst,
+		const int dst_len, const runlist_element *rl,
+		const VCN start_vcn, VCN *const stop_vcn)
 {
 	LCN prev_lcn;
-	s8 *dst_max;
-	int i;
+	u8 *dst_max, *dst_next;
 	s8 len_len, lcn_len;
 
+	if (start_vcn < 0)
+		goto val_err;
 	if (!rl) {
-		if (dst_len < 1)
-			goto size_err;
+		if (start_vcn)
+			goto val_err;
+		if (stop_vcn)
+			*stop_vcn = 0;
+		if (dst_len < 1) {
+			errno = ENOSPC;
+			return -1;
+		}
 		/* Terminator byte. */
 		*dst = 0;
 		return 0;
 	}
+	/* Skip to runlist element containing @start_vcn. */
+	while (rl->length && start_vcn >= rl[1].vcn)
+		rl++;
+	if ((!rl->length && start_vcn > rl->vcn) || start_vcn < rl->vcn)
+		goto val_err;
 	/*
 	 * @dst_max is used for bounds checking in
 	 * ntfs_write_significant_bytes().
 	 */
 	dst_max = dst + dst_len - 1;
-	for (prev_lcn = i = 0; rl[i].length; i++) {
-		if (rl[i].length < 0 || rl[i].lcn < LCN_HOLE)
+	prev_lcn = 0;
+	/* Do the first partial run if present. */
+	if (start_vcn > rl->vcn) {
+		s64 delta;
+
+		/* We know rl->length != 0 already. */
+		if (rl->length < 0 || rl->lcn < LCN_HOLE)
+			goto err_out;
+		delta = start_vcn - rl->vcn;
+		/* Write length. */
+		len_len = ntfs_write_significant_bytes(dst + 1, dst_max,
+				rl->length - delta);
+		if (len_len < 0)
+			goto size_err;
+		/*
+		 * If the logical cluster number (lcn) denotes a hole and we
+		 * are on NTFS 3.0+, we don't store it at all, i.e. we need
+		 * zero space. On earlier NTFS versions we just write the lcn
+		 * change.  FIXME: Do we need to write the lcn change or just
+		 * the lcn in that case?  Not sure as I have never seen this
+		 * case on NT4. - We assume that we just need to write the lcn
+		 * change until someone tells us otherwise... (AIA)
+		 */
+		if (rl->lcn >= 0 || vol->major_ver < 3) {
+			prev_lcn = rl->lcn;
+			if (rl->lcn >= 0)
+				prev_lcn += delta;
+			/* Write change in lcn. */
+			lcn_len = ntfs_write_significant_bytes(dst + 1 +
+					len_len, dst_max, prev_lcn);
+			if (lcn_len < 0)
+				goto size_err;
+		} else
+			lcn_len = 0;
+		dst_next = dst + len_len + lcn_len + 1;
+		if (dst_next > dst_max)
+			goto size_err;
+		/* Update header byte. */
+		*dst = lcn_len << 4 | len_len;
+		/* Position at next mapping pairs array element. */
+		dst = dst_next;
+		/* Go to next runlist element. */
+		rl++;
+	}
+	/* Do the full runs. */
+	for (; rl->length; rl++) {
+		if (rl->length < 0 || rl->lcn < LCN_HOLE)
 			goto err_out;
 		/* Write length. */
 		len_len = ntfs_write_significant_bytes(dst + 1, dst_max,
-				rl[i].length);
+				rl->length);
 		if (len_len < 0)
 			goto size_err;
 		/*
@@ -1399,29 +1514,42 @@ int ntfs_mapping_pairs_build(const ntfs_volume *vol, s8 *dst,
 		 * case on NT4. - We assume that we just need to write the lcn
 		 * change until someone tells us otherwise... (AIA)
 		 */
-		if (rl[i].lcn != LCN_HOLE || vol->major_ver < 3) {
+		if (rl->lcn >= 0 || vol->major_ver < 3) {
+			/* Write change in lcn. */
 			lcn_len = ntfs_write_significant_bytes(dst + 1 +
-					len_len, dst_max, rl[i].lcn - prev_lcn);
+					len_len, dst_max, rl->lcn - prev_lcn);
 			if (lcn_len < 0)
 				goto size_err;
-			prev_lcn = rl[i].lcn;
+			prev_lcn = rl->lcn;
 		} else
 			lcn_len = 0;
+		dst_next = dst + len_len + lcn_len + 1;
+		if (dst_next > dst_max)
+			goto size_err;
 		/* Update header byte. */
 		*dst = lcn_len << 4 | len_len;
-		/* Position ourselves at next mapping pairs array element. */
+		/* Position at next mapping pairs array element. */
 		dst += 1 + len_len + lcn_len;
 	}
-	if (dst <= dst_max) {
-		/* Terminator byte. */
-		*dst = 0;
-		return 0;
-	}
+	/* Set stop vcn. */
+	if (stop_vcn)
+		*stop_vcn = rl->vcn;
+	/* Add terminator byte. */
+	*dst = 0;
+	return 0;
 size_err:
+	/* Set stop vcn. */
+	if (stop_vcn)
+		*stop_vcn = rl->vcn;
+	/* Add terminator byte. */
+	*dst = 0;
 	errno = ENOSPC;
 	return -1;
+val_err:
+	errno = EINVAL;
+	return -1;
 err_out:
-	if (rl[i].lcn == LCN_RL_NOT_MAPPED)
+	if (rl->lcn == LCN_RL_NOT_MAPPED)
 		errno = EINVAL;
 	else
 		errno = EIO;
@@ -1515,3 +1643,412 @@ int ntfs_rl_truncate(runlist **arl, const VCN start_vcn)
 	/* Done! */
 	return 0;
 }
+
+/**
+ * ntfs_rl_sparse - check whether runlist have sparse regions or not.
+ * @rl:		runlist to check
+ *
+ * Return 1 if have, 0 if not, -1 on error with errno set to the error code.
+ */
+int ntfs_rl_sparse(runlist *rl)
+{
+	runlist *rlc;
+
+	if (!rl) {
+		Dprintf("%s(): Invalid argument passed.\n", __FUNCTION__);
+		errno = EINVAL;
+		return -1;
+	}
+
+	for (rlc = rl; rlc->length; rlc++)
+		if (rlc->lcn < 0) {
+			if (rlc->lcn != LCN_HOLE) {
+				Dprintf("%s(): Received unmapped runlist.\n",
+					__FUNCTION__);
+				errno = EINVAL;
+				return -1;
+			}
+			return 1;
+		}
+	return 0;
+}
+
+/**
+ * ntfs_rl_get_compressed_size - calculate length of non sparse regions
+ * @vol:	ntfs volume (need for cluster size)
+ * @rl:		runlist to calculate for
+ *
+ * Return compressed size or -1 on error with errno set to the error code.
+ */
+s64 ntfs_rl_get_compressed_size(ntfs_volume *vol, runlist *rl)
+{
+	runlist *rlc;
+	s64 ret = 0;
+
+	if (!rl) {
+		Dprintf("%s(): Invalid argument passed.\n", __FUNCTION__);
+		errno = EINVAL;
+		return -1;
+	}
+
+	for (rlc = rl; rlc->length; rlc++) {
+		if (rlc->lcn < 0) {
+			if (rlc->lcn != LCN_HOLE) {
+				Dprintf("%s(): Received unmapped runlist.\n",
+					__FUNCTION__);
+				errno = EINVAL;
+				return -1;
+			}
+		} else
+			ret += rlc->length;
+	}
+	return ret << vol->cluster_size_bits;
+}
+
+
+#ifdef NTFS_TEST
+/**
+ * test_rl_helper
+ */
+#define MKRL(R,V,L,S)				\
+	(R)->vcn = V;				\
+	(R)->lcn = L;				\
+	(R)->length = S;
+/*
+}
+*/
+/**
+ * test_rl_dump_runlist
+ */
+static void test_rl_dump_runlist (const runlist_element *rl)
+{
+	int abbr = 0;	/* abbreviate long lists */
+	int len = 0;
+	int i;
+	const char *lcn_str[5] = { "HOLE", "NOTMAP", "ENOENT", "XXXX" };
+
+	if (!rl) {
+		printf("    Run list not present.\n");
+		return;
+	}
+
+	if (abbr)
+		for (len = 0; rl[len].length; len++) ;
+	
+	printf("     VCN      LCN      len\n");
+	for (i = 0; ; i++, rl++) {
+		LCN lcn = rl->lcn;
+
+		if ((abbr) && (len > 20)) {
+			if (i == 4)
+				printf ("     ...\n");
+			if ((i > 3) && (i < (len - 3)))
+				continue;
+		}
+
+		if (lcn < (LCN)0) {
+			int ind = -lcn - 1;
+
+			if (ind > -LCN_ENOENT - 1)
+				ind = 3;
+			printf("%8lld %8s %8lld\n",
+				rl->vcn, lcn_str[ind], rl->length);
+		} else
+			printf("%8lld %8lld %8lld\n",
+				rl->vcn, rl->lcn, rl->length);
+		if (!rl->length)
+			break;
+	}
+	if ((abbr) && (len > 20))
+		printf ("    (%d entries)\n", len+1);
+	printf ("\n");
+}
+
+/**
+ * test_rl_runlists_merge
+ */
+static runlist_element * test_rl_runlists_merge (runlist_element *drl, runlist_element *srl)
+{
+	runlist_element *res = NULL;
+
+	printf ("dst:\n");
+	test_rl_dump_runlist (drl);
+	printf ("src:\n");
+	test_rl_dump_runlist (srl);
+
+	res = ntfs_runlists_merge (drl, srl);
+
+	printf ("res:\n");
+	test_rl_dump_runlist (res);
+
+	return res;
+}
+
+/**
+ * test_rl_read_buffer
+ */
+static int test_rl_read_buffer (const char *file, u8 *buf, int bufsize)
+{
+	FILE *fptr;
+
+	fptr = fopen (file, "r");
+	if (!fptr) {
+		printf ("open %s\n", file);
+		return 0;
+	}
+
+	if (fread (buf, bufsize, 1, fptr) == 99) {
+		printf ("read %s\n", file);
+		return 0;
+	}
+
+	fclose (fptr);
+	return 1;
+}
+
+/**
+ * test_rl_pure_src
+ */
+static runlist_element * test_rl_pure_src (BOOL contig, BOOL multi, int vcn, int len)
+{
+	runlist_element *result;
+	int fudge;
+
+	if (contig)
+		fudge = 0;
+	else
+		fudge = 999;
+
+	result = malloc (4096);
+	if (multi) {
+		MKRL (result+0, vcn + (0*len/4), fudge + vcn + 1000 + (0*len/4), len / 4)
+		MKRL (result+1, vcn + (1*len/4), fudge + vcn + 1000 + (1*len/4), len / 4)
+		MKRL (result+2, vcn + (2*len/4), fudge + vcn + 1000 + (2*len/4), len / 4)
+		MKRL (result+3, vcn + (3*len/4), fudge + vcn + 1000 + (3*len/4), len / 4)
+		MKRL (result+4, vcn + (4*len/4), LCN_RL_NOT_MAPPED,              0)
+	} else {
+		MKRL (result+0, vcn,       fudge + vcn + 1000, len)
+		MKRL (result+1, vcn + len, LCN_RL_NOT_MAPPED,  0)
+	}
+	return result;
+}
+
+/**
+ * test_rl_pure_test
+ */
+static void test_rl_pure_test (int test, BOOL contig, BOOL multi, int vcn, int len, runlist_element *file, int size)
+{
+	runlist_element *src;
+	runlist_element *dst;
+	runlist_element *res;
+
+	src = test_rl_pure_src (contig, multi, vcn, len);
+	dst = malloc (4096);
+
+	memcpy (dst, file, size);
+
+	printf ("Test %2d ----------\n", test);
+	res = test_rl_runlists_merge (dst, src);
+
+	free (res);
+}
+
+/**
+ * test_rl_pure
+ */
+static void test_rl_pure (char *contig, char *multi)
+{
+		/* VCN,  LCN, len */
+	static runlist_element file1[] = {
+		{    0,   -1, 100 },	/* HOLE */
+		{  100, 1100, 100 },	/* DATA */
+		{  200,   -1, 100 },	/* HOLE */
+		{  300, 1300, 100 },	/* DATA */
+		{  400,   -1, 100 },	/* HOLE */
+		{  500,   -3,   0 }	/* NOENT */
+	};
+	static runlist_element file2[] = {
+		{    0, 1000, 100 },	/* DATA */
+		{  100,   -1, 100 },	/* HOLE */
+		{  200,   -3,   0 }	/* NOENT */
+	};
+	static runlist_element file3[] = {
+		{    0, 1000, 100 },	/* DATA */
+		{  100,   -3,   0 }	/* NOENT */
+	};
+	static runlist_element file4[] = {
+		{    0,   -3,   0 }	/* NOENT */
+	};
+	static runlist_element file5[] = {
+		{    0,   -2, 100 },	/* NOTMAP */
+		{  100, 1100, 100 },	/* DATA */
+		{  200,   -2, 100 },	/* NOTMAP */
+		{  300, 1300, 100 },	/* DATA */
+		{  400,   -2, 100 },	/* NOTMAP */
+		{  500,   -3,   0 }	/* NOENT */
+	};
+	static runlist_element file6[] = {
+		{    0, 1000, 100 },	/* DATA */
+		{  100,   -2, 100 },	/* NOTMAP */
+		{  200,   -3,   0 }	/* NOENT */
+	};
+	BOOL c, m;
+
+	if (strcmp (contig, "contig") == 0)
+		c = TRUE;
+	else if (strcmp (contig, "noncontig") == 0)
+		c = FALSE;
+	else {
+		printf ("rl pure [contig|noncontig] [single|multi]\n");
+		return;
+	}
+	if (strcmp (multi, "multi") == 0)
+		m = TRUE;
+	else if (strcmp (multi, "single") == 0)
+		m = FALSE;
+	else {
+		printf ("rl pure [contig|noncontig] [single|multi]\n");
+		return;
+	}
+
+	test_rl_pure_test (1,  c, m,   0,  40, file1, sizeof (file1));
+	test_rl_pure_test (2,  c, m,  40,  40, file1, sizeof (file1));
+	test_rl_pure_test (3,  c, m,  60,  40, file1, sizeof (file1));
+	test_rl_pure_test (4,  c, m,   0, 100, file1, sizeof (file1));
+	test_rl_pure_test (5,  c, m, 200,  40, file1, sizeof (file1));
+	test_rl_pure_test (6,  c, m, 240,  40, file1, sizeof (file1));
+	test_rl_pure_test (7,  c, m, 260,  40, file1, sizeof (file1));
+	test_rl_pure_test (8,  c, m, 200, 100, file1, sizeof (file1));
+	test_rl_pure_test (9,  c, m, 400,  40, file1, sizeof (file1));
+	test_rl_pure_test (10, c, m, 440,  40, file1, sizeof (file1));
+	test_rl_pure_test (11, c, m, 460,  40, file1, sizeof (file1));
+	test_rl_pure_test (12, c, m, 400, 100, file1, sizeof (file1));
+	test_rl_pure_test (13, c, m, 160, 100, file2, sizeof (file2));
+	test_rl_pure_test (14, c, m, 100, 140, file2, sizeof (file2));
+	test_rl_pure_test (15, c, m, 200,  40, file2, sizeof (file2));
+	test_rl_pure_test (16, c, m, 240,  40, file2, sizeof (file2));
+	test_rl_pure_test (17, c, m, 100,  40, file3, sizeof (file3));
+	test_rl_pure_test (18, c, m, 140,  40, file3, sizeof (file3));
+	test_rl_pure_test (19, c, m,   0,  40, file4, sizeof (file4));
+	test_rl_pure_test (20, c, m,  40,  40, file4, sizeof (file4));
+	test_rl_pure_test (21, c, m,   0,  40, file5, sizeof (file5));
+	test_rl_pure_test (22, c, m,  40,  40, file5, sizeof (file5));
+	test_rl_pure_test (23, c, m,  60,  40, file5, sizeof (file5));
+	test_rl_pure_test (24, c, m,   0, 100, file5, sizeof (file5));
+	test_rl_pure_test (25, c, m, 200,  40, file5, sizeof (file5));
+	test_rl_pure_test (26, c, m, 240,  40, file5, sizeof (file5));
+	test_rl_pure_test (27, c, m, 260,  40, file5, sizeof (file5));
+	test_rl_pure_test (28, c, m, 200, 100, file5, sizeof (file5));
+	test_rl_pure_test (29, c, m, 400,  40, file5, sizeof (file5));
+	test_rl_pure_test (30, c, m, 440,  40, file5, sizeof (file5));
+	test_rl_pure_test (31, c, m, 460,  40, file5, sizeof (file5));
+	test_rl_pure_test (32, c, m, 400, 100, file5, sizeof (file5));
+	test_rl_pure_test (33, c, m, 160, 100, file6, sizeof (file6));
+	test_rl_pure_test (34, c, m, 100, 140, file6, sizeof (file6));
+}
+
+/**
+ * test_rl_zero
+ */
+static void test_rl_zero (void)
+{
+	runlist_element *jim = NULL;
+	runlist_element *bob = NULL;
+
+	bob = calloc (3, sizeof (runlist_element));
+	if (!bob)
+		return;
+
+	MKRL(bob+0, 10, 99, 5)
+	MKRL(bob+1, 15, LCN_RL_NOT_MAPPED, 0)
+
+	jim = test_rl_runlists_merge (jim, bob);
+	if (!jim)
+		return;
+
+	free (jim);
+}
+
+/**
+ * test_rl_frag_combine
+ */
+static void test_rl_frag_combine (ntfs_volume *vol, ATTR_RECORD *attr1, ATTR_RECORD *attr2, ATTR_RECORD *attr3)
+{
+	runlist_element *run1;
+	runlist_element *run2;
+	runlist_element *run3;
+
+	run1 = ntfs_mapping_pairs_decompress (vol, attr1, NULL);
+	if (!run1)
+		return;
+
+	run2 = ntfs_mapping_pairs_decompress (vol, attr2, NULL);
+	if (!run2)
+		return;
+
+	run1 = test_rl_runlists_merge (run1, run2);
+
+	run3 = ntfs_mapping_pairs_decompress (vol, attr3, NULL);
+	if (!run3)
+		return;
+
+	run1 = test_rl_runlists_merge (run1, run3);
+
+	free (run1);
+}
+
+/**
+ * test_rl_frag
+ */
+static void test_rl_frag (char *test)
+{
+	ntfs_volume vol;
+	ATTR_RECORD *attr1 = malloc (1024);
+	ATTR_RECORD *attr2 = malloc (1024);
+	ATTR_RECORD *attr3 = malloc (1024);
+
+	if (!attr1 || !attr2 || !attr3)
+		goto out;
+
+	vol.sb = NULL;
+	vol.sector_size_bits = 9;
+	vol.cluster_size = 2048;
+	vol.cluster_size_bits = 11;
+	vol.major_ver = 3;
+
+	if (!test_rl_read_buffer ("runlist-data/attr1.bin", (u8*) attr1, 1024))
+		goto out;
+	if (!test_rl_read_buffer ("runlist-data/attr2.bin", (u8*) attr2, 1024))
+		goto out;
+	if (!test_rl_read_buffer ("runlist-data/attr3.bin", (u8*) attr3, 1024))
+		goto out;
+
+	if      (strcmp (test, "123") == 0)  test_rl_frag_combine (&vol, attr1, attr2, attr3);
+	else if (strcmp (test, "132") == 0)  test_rl_frag_combine (&vol, attr1, attr3, attr2);
+	else if (strcmp (test, "213") == 0)  test_rl_frag_combine (&vol, attr2, attr1, attr3);
+	else if (strcmp (test, "231") == 0)  test_rl_frag_combine (&vol, attr2, attr3, attr1);
+	else if (strcmp (test, "312") == 0)  test_rl_frag_combine (&vol, attr3, attr1, attr2);
+	else if (strcmp (test, "321") == 0)  test_rl_frag_combine (&vol, attr3, attr2, attr1);
+	else printf ("Frag: No such test '%s'\n", test);
+
+out:
+	free (attr1);
+	free (attr2);
+	free (attr3);
+}
+
+/**
+ * test_rl_main
+ */
+int test_rl_main (int argc, char *argv[])
+{
+	if      ((argc == 2) && (strcmp (argv[1], "zero") == 0)) test_rl_zero();
+	else if ((argc == 3) && (strcmp (argv[1], "frag") == 0)) test_rl_frag (argv[2]);
+	else if ((argc == 4) && (strcmp (argv[1], "pure") == 0)) test_rl_pure (argv[2], argv[3]);
+	else printf ("rl [zero|frag|pure] {args}\n");
+
+	return 0;
+}
+
+#endif
+
