@@ -21,11 +21,21 @@
 
 #include "config.h"
 
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_WCHAR_H
 #include <wchar.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
+#ifdef HAVE_ERRNO_H
 #include <errno.h>
+#endif
 
 #include "types.h"
 #include "unistr.h"
@@ -69,7 +79,7 @@ const u8 legal_ansi_char_array[0x40] = {
  *
  * Compare the names @s1 and @s2 and return TRUE (1) if the names are
  * identical, or FALSE (0) if they are not identical. If @ic is IGNORE_CASE,
- * the @upcase table is used to performa a case insensitive comparison.
+ * the @upcase table is used to perform a case insensitive comparison.
  */
 BOOL ntfs_names_are_equal(const ntfschar *s1, size_t s1_len,
 		const ntfschar *s2, size_t s2_len,
@@ -88,12 +98,14 @@ BOOL ntfs_names_are_equal(const ntfschar *s1, size_t s1_len,
 
 /**
  * ntfs_names_collate - collate two Unicode names
+ * @name1:	first Unicode name to compare
+ * @name1_len:	length of first Unicode name to compare
+ * @name2:	second Unicode name to compare
+ * @name2_len:	length of second Unicode name to compare
+ * @err_val:	if @name1 contains an invalid character return this value
+ * @ic:		either CASE_SENSITIVE or IGNORE_CASE
  * @upcase:	upcase table (ignored if @ic is CASE_SENSITIVE)
  * @upcase_len:	upcase table size (ignored if @ic is CASE_SENSITIVE)
- * @name1:	first Unicode name to compare
- * @name2:	second Unicode name to compare
- * @ic:		either CASE_SENSITIVE or IGNORE_CASE
- * @err_val:	if @name1 contains an invalid character return this value
  *
  * ntfs_names_collate() collates two Unicode names and returns:
  *
@@ -113,7 +125,7 @@ int ntfs_names_collate(const ntfschar *name1, const u32 name1_len,
 	ntfschar c1, c2;
 
 #ifdef DEBUG
-	if (!name1 || !name2 || (ic && !upcase && upcase_len)) {
+	if (!name1 || !name2 || (ic && (!upcase || !upcase_len))) {
 		Dputs("ntfs_names_collate received NULL pointer!");
 		exit(1);
 	}
@@ -254,6 +266,37 @@ u32 ntfs_ucsnlen(const ntfschar *s, u32 maxlen)
 	return i;
 }
 
+/**
+ * ntfs_ucsndup - duplicate little endian Unicode string
+ * @s:		pointer to Unicode string
+ * @maxlen:	maximum length of string @s
+ *
+ * Return a pointer to a new little endian Unicode string which is a duplicate
+ * of the string s.  Memory for the new string is obtained with malloc(3), and
+ * can be freed with free(3).
+ *
+ * A maximum of @maxlen Unicode characters are copied and a terminating
+ * (ntfschar)'\0' little endian Unicode character is added.
+ *
+ * This function never looks beyond @s + @maxlen.
+ *
+ * Return a pointer to the new little endian Unicode string on success and NULL
+ * on failure with errno set to the error code.
+ */
+ntfschar *ntfs_ucsndup(const ntfschar *s, u32 maxlen)
+{
+	ntfschar *dst;
+	u32 len;
+
+	len = ntfs_ucsnlen(s, maxlen);
+	dst = malloc((len + 1) * sizeof(ntfschar));
+	if (dst) {
+		memcpy(dst, s, len * sizeof(ntfschar));
+		dst[len] = cpu_to_le16(L'\0');
+	}
+	return dst;
+}
+
 /**
  * ntfs_name_upcase
  */
@@ -281,8 +324,8 @@ void ntfs_file_value_upcase(FILE_NAME_ATTR *file_name_attr,
 /**
  * ntfs_file_values_compare
  */
-int ntfs_file_values_compare(FILE_NAME_ATTR *file_name_attr1,
-		FILE_NAME_ATTR *file_name_attr2,
+int ntfs_file_values_compare(const FILE_NAME_ATTR *file_name_attr1,
+		const FILE_NAME_ATTR *file_name_attr2,
 		const int err_val, const IGNORE_CASE_BOOL ic,
 		const ntfschar *upcase, const u32 upcase_len)
 {
@@ -436,7 +479,7 @@ int ntfs_mbstoucs(const char *ins, ntfschar **outs, int outs_len)
 	ntfschar *ucs;
 	const char *s;
 	wchar_t wc;
-	int i, o, cnt, ins_len, ucs_len;
+	int i, o, cnt, ins_len, ucs_len, ins_size;
 #ifdef HAVE_MBSINIT
 	mbstate_t mbstate;
 #endif
@@ -451,6 +494,8 @@ int ntfs_mbstoucs(const char *ins, ntfschar **outs, int outs_len)
 		errno = ENAMETOOLONG;
 		return -1;
 	}
+	/* Determine the size of the multi-byte string in bytes. */
+	ins_size = strlen(ins);
 	/* Determine the length of the multi-byte string. */
 	s = ins;
 #if defined(HAVE_MBSINIT)
@@ -491,7 +536,7 @@ int ntfs_mbstoucs(const char *ins, ntfschar **outs, int outs_len)
 #else
 	mbtowc(NULL, NULL, 0);
 #endif
-	for (i = o = cnt = 0; o < ins_len; i += cnt, o++) {
+	for (i = o = cnt = 0; i < ins_size; i += cnt, o++) {
 		/* Reallocate memory if necessary or abort. */
 		if (o >= ucs_len) {
 			ntfschar *tc;
@@ -512,9 +557,9 @@ int ntfs_mbstoucs(const char *ins, ntfschar **outs, int outs_len)
 		}
 		/* Convert the multibyte character to a wide character. */
 #ifdef HAVE_MBSINIT
-		cnt = mbrtowc(&wc, ins + i, ins_len - i, &mbstate);
+		cnt = mbrtowc(&wc, ins + i, ins_size - i, &mbstate);
 #else
-		cnt = mbtowc(&wc, ins + i, ins_len - i);
+		cnt = mbtowc(&wc, ins + i, ins_size - i);
 #endif
 		if (!cnt)
 			break;
@@ -556,3 +601,69 @@ err_out:
 	}
 	return -1;
 }
+
+/**
+ * ntfs_upcase_table_build - build the default upcase table for NTFS
+ * @uc:		destination buffer where to store the built table
+ * @uc_len:	size of destination buffer in bytes
+ *
+ * ntfs_upcase_table_build() builds the default upcase table for NTFS and
+ * stores it in the caller supplied buffer @uc of size @uc_len.
+ *
+ * Note, @uc_len must be at least 128kiB in size or bad things will happen!
+ */
+void ntfs_upcase_table_build(ntfschar *uc, u32 uc_len)
+{
+	static int uc_run_table[][3] = { /* Start, End, Add */
+	{0x0061, 0x007B,  -32}, {0x0451, 0x045D, -80}, {0x1F70, 0x1F72,  74},
+	{0x00E0, 0x00F7,  -32}, {0x045E, 0x0460, -80}, {0x1F72, 0x1F76,  86},
+	{0x00F8, 0x00FF,  -32}, {0x0561, 0x0587, -48}, {0x1F76, 0x1F78, 100},
+	{0x0256, 0x0258, -205}, {0x1F00, 0x1F08,   8}, {0x1F78, 0x1F7A, 128},
+	{0x028A, 0x028C, -217}, {0x1F10, 0x1F16,   8}, {0x1F7A, 0x1F7C, 112},
+	{0x03AC, 0x03AD,  -38}, {0x1F20, 0x1F28,   8}, {0x1F7C, 0x1F7E, 126},
+	{0x03AD, 0x03B0,  -37}, {0x1F30, 0x1F38,   8}, {0x1FB0, 0x1FB2,   8},
+	{0x03B1, 0x03C2,  -32}, {0x1F40, 0x1F46,   8}, {0x1FD0, 0x1FD2,   8},
+	{0x03C2, 0x03C3,  -31}, {0x1F51, 0x1F52,   8}, {0x1FE0, 0x1FE2,   8},
+	{0x03C3, 0x03CC,  -32}, {0x1F53, 0x1F54,   8}, {0x1FE5, 0x1FE6,   7},
+	{0x03CC, 0x03CD,  -64}, {0x1F55, 0x1F56,   8}, {0x2170, 0x2180, -16},
+	{0x03CD, 0x03CF,  -63}, {0x1F57, 0x1F58,   8}, {0x24D0, 0x24EA, -26},
+	{0x0430, 0x0450,  -32}, {0x1F60, 0x1F68,   8}, {0xFF41, 0xFF5B, -32},
+	{0}
+	};
+	static int uc_dup_table[][2] = { /* Start, End */
+	{0x0100, 0x012F}, {0x01A0, 0x01A6}, {0x03E2, 0x03EF}, {0x04CB, 0x04CC},
+	{0x0132, 0x0137}, {0x01B3, 0x01B7}, {0x0460, 0x0481}, {0x04D0, 0x04EB},
+	{0x0139, 0x0149}, {0x01CD, 0x01DD}, {0x0490, 0x04BF}, {0x04EE, 0x04F5},
+	{0x014A, 0x0178}, {0x01DE, 0x01EF}, {0x04BF, 0x04BF}, {0x04F8, 0x04F9},
+	{0x0179, 0x017E}, {0x01F4, 0x01F5}, {0x04C1, 0x04C4}, {0x1E00, 0x1E95},
+	{0x018B, 0x018B}, {0x01FA, 0x0218}, {0x04C7, 0x04C8}, {0x1EA0, 0x1EF9},
+	{0}
+	};
+	static int uc_byte_table[][2] = { /* Offset, Value */
+	{0x00FF, 0x0178}, {0x01AD, 0x01AC}, {0x01F3, 0x01F1}, {0x0269, 0x0196},
+	{0x0183, 0x0182}, {0x01B0, 0x01AF}, {0x0253, 0x0181}, {0x026F, 0x019C},
+	{0x0185, 0x0184}, {0x01B9, 0x01B8}, {0x0254, 0x0186}, {0x0272, 0x019D},
+	{0x0188, 0x0187}, {0x01BD, 0x01BC}, {0x0259, 0x018F}, {0x0275, 0x019F},
+	{0x018C, 0x018B}, {0x01C6, 0x01C4}, {0x025B, 0x0190}, {0x0283, 0x01A9},
+	{0x0192, 0x0191}, {0x01C9, 0x01C7}, {0x0260, 0x0193}, {0x0288, 0x01AE},
+	{0x0199, 0x0198}, {0x01CC, 0x01CA}, {0x0263, 0x0194}, {0x0292, 0x01B7},
+	{0x01A8, 0x01A7}, {0x01DD, 0x018E}, {0x0268, 0x0197},
+	{0}
+	};
+	int i, r;
+
+	memset((char*)uc, 0, uc_len);
+	uc_len >>= 1;
+	if (uc_len > 65536)
+		uc_len = 65536;
+	for (i = 0; (u32)i < uc_len; i++)
+		uc[i] = i;
+	for (r = 0; uc_run_table[r][0]; r++)
+		for (i = uc_run_table[r][0]; i < uc_run_table[r][1]; i++)
+			uc[i] += uc_run_table[r][2];
+	for (r = 0; uc_dup_table[r][0]; r++)
+		for (i = uc_dup_table[r][0]; i < uc_dup_table[r][1]; i += 2)
+			uc[i + 1]--;
+	for (r = 0; uc_byte_table[r][0]; r++)
+		uc[uc_byte_table[r][0]] = uc_byte_table[r][1];
+}
