@@ -1,7 +1,8 @@
 /*
  * volume.c - NTFS volume handling code. Part of the Linux-NTFS project.
  *
- * Copyright (c) 2000-2004 Anton Altaparmakov
+ * Copyright (c) 2000-2005 Anton Altaparmakov
+ * Copyright (c) 2002-2005 Szabolcs Szakacsits
  *
  * This program/include file is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as published
@@ -21,14 +22,30 @@
 
 #include "config.h"
 
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
+#ifdef HAVE_FCNTL_H
 #include <fcntl.h>
+#endif
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
+#endif
+#ifdef HAVE_ERRNO_H
 #include <errno.h>
+#endif
+#ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
+#endif
+#ifdef HAVE_LIMITS_H
 #include <limits.h>
+#endif
 
 #include "volume.h"
 #include "attrib.h"
@@ -38,6 +55,12 @@
 #include "debug.h"
 #include "inode.h"
 #include "runlist.h"
+#include "logfile.h"
+#include "dir.h"
+
+#ifndef PATH_MAX
+#define PATH_MAX 4096
+#endif
 
 /**
  * ntfs_volume_alloc -
@@ -72,16 +95,24 @@ ntfs_volume *ntfs_volume_alloc(void)
  */
 static void __ntfs_volume_release(ntfs_volume *v)
 {
+	if (v->lcnbmp_ni && NInoDirty(v->lcnbmp_ni))
+		ntfs_inode_sync(v->lcnbmp_ni);
+	if (v->vol_ni)
+		ntfs_inode_close(v->vol_ni);
 	if (v->lcnbmp_na)
 		ntfs_attr_close(v->lcnbmp_na);
 	if (v->lcnbmp_ni)
 		ntfs_inode_close(v->lcnbmp_ni);
+	if (v->mft_ni && NInoDirty(v->mft_ni))
+		ntfs_inode_sync(v->mft_ni);
 	if (v->mftbmp_na)
 		ntfs_attr_close(v->mftbmp_na);
 	if (v->mft_na)
 		ntfs_attr_close(v->mft_na);
 	if (v->mft_ni)
 		ntfs_inode_close(v->mft_ni);
+	if (v->mftmirr_ni && NInoDirty(v->mftmirr_ni))
+		ntfs_inode_sync(v->mftmirr_ni);
 	if (v->mftmirr_na)
 		ntfs_attr_close(v->mftmirr_na);
 	if (v->mftmirr_ni)
@@ -105,9 +136,6 @@ static void __ntfs_volume_release(ntfs_volume *v)
 	free(v);
 }
 
-/* External declaration for internal function. */
-extern ntfs_inode *ntfs_inode_allocate(ntfs_volume *);
-
 /**
  * Internal:
  *
@@ -156,7 +184,7 @@ static int ntfs_mft_load(ntfs_volume *vol)
 		Dputs("Error: $MFT has invalid magic.");
 		goto io_error_exit;
 	}
-	ctx = ntfs_attr_get_search_ctx(vol->mft_ni, mb);
+	ctx = ntfs_attr_get_search_ctx(vol->mft_ni, NULL);
 	if (!ctx) {
 		Dperror("Failed to allocate attribute search context");
 		goto error_exit;
@@ -193,21 +221,10 @@ static int ntfs_mft_load(ntfs_volume *vol)
 		goto io_error_exit;
 	}
 	if (l != vol->mft_ni->attr_list_size) {
-		Dputs("Error: got unexepected amount of data when reading "
+		Dputs("Error: got unexpected amount of data when reading "
 				"$MFT/$ATTRIBUTE_LIST.");
 		goto io_error_exit;
 	}
-	if (ctx->attr->non_resident) {
-		NInoSetAttrListNonResident(vol->mft_ni);
-		// FIXME: We are duplicating work here! (AIA)
-		vol->mft_ni->attr_list_rl = ntfs_mapping_pairs_decompress(vol,
-				ctx->attr, NULL);
-		if (!vol->mft_ni->attr_list_rl) {
-			Dperror("Error: failed to get runlist for "
-					"$MFT/$ATTRIBUTE_LIST");
-			goto error_exit;
-		}
-	}
 mft_has_no_attr_list:
 	/* We now have a fully setup ntfs inode for $MFT in vol->mft_ni. */
 
@@ -217,9 +234,6 @@ mft_has_no_attr_list:
 		Dperror("Failed to open ntfs attribute");
 		goto error_exit;
 	}
-	/* Set the number of initialized mft records. */
-	vol->nr_mft_records = vol->mft_na->initialized_size >>
-			vol->mft_record_size_bits;
 	/* Read all extents from the $DATA attribute in $MFT. */
 	ntfs_attr_reinit_search_ctx(ctx);
 	last_vcn = vol->mft_na->allocated_size >> vol->cluster_size_bits;
@@ -420,6 +434,15 @@ ntfs_volume *ntfs_volume_startup(struct ntfs_device *dev, unsigned long rwflag)
 	vol = ntfs_volume_alloc();
 	if (!vol)
 		goto error_exit;
+	/* Create the default upcase table. */
+	vol->upcase_len = 65536;
+	vol->upcase = (ntfschar*)malloc(vol->upcase_len * sizeof(ntfschar));
+	if (!vol->upcase) {
+		Dperror("Error allocating memory for upcase table.");
+		goto error_exit;
+	}
+	ntfs_upcase_table_build(vol->upcase,
+			vol->upcase_len * sizeof(ntfschar));
 	if ((rwflag & MS_RDONLY) == MS_RDONLY)
 		NVolSetReadOnly(vol);
 	Dprintf("Reading bootsector... ");
@@ -560,6 +583,166 @@ error_exit:
 	return NULL;
 }
 
+/**
+ * ntfs_volume_check_logfile - check logfile on target volume
+ * @vol:	volume on which to check logfile
+ *
+ * Return 0 on success and -1 on error with errno set error code.
+ */
+static int ntfs_volume_check_logfile(ntfs_volume *vol)
+{
+	ntfs_inode *ni;
+	ntfs_attr *na = NULL;
+	RESTART_PAGE_HEADER *rp = NULL;
+	int err = 0;
+
+	if ((ni = ntfs_inode_open(vol, FILE_LogFile)) == NULL) {
+		Dprintf("Failed to open inode FILE_LogFile.\n");
+		errno = EIO;
+		return -1;
+	}
+	if ((na = ntfs_attr_open(ni, AT_DATA, AT_UNNAMED, 0)) == NULL) {
+		Dprintf("Failed to open $FILE_LogFile/$DATA\n");
+		err = EIO;
+		goto exit;
+	}
+	if (!ntfs_check_logfile(na, &rp) || !ntfs_is_logfile_clean(na, rp))
+		err = EOPNOTSUPP;
+	if (rp)
+		free(rp);
+exit:
+	if (na)
+		ntfs_attr_close(na);
+	ntfs_inode_close(ni);
+	if (err) {
+		errno = err;
+		return -1;
+	}
+	return 0;
+}
+
+/**
+ * ntfs_hiberfile_open - Find and open '/hiberfil.sys'
+ * @vol:    An ntfs volume obtained from ntfs_mount
+ *
+ * Return:  inode  Success, hibefil.sys is valid
+ *	    NULL   hibefil.sys doesn't exist or some other error occurred
+ */
+static ntfs_inode *ntfs_hiberfile_open(ntfs_volume *vol)
+{
+	u64 inode;
+	ntfs_inode *ni_root;
+	ntfs_inode *ni_hibr = NULL;
+	ntfschar   *unicode = NULL;
+	int unicode_len;
+	const char *hiberfile = "hiberfil.sys";
+
+	if (!vol) {
+		errno = EINVAL;
+		return NULL;
+	}
+
+	ni_root = ntfs_inode_open(vol, FILE_root);
+	if (!ni_root) {
+		Dprintf("Couldn't open the root directory.\n");
+		return NULL;
+	}
+
+	unicode_len = ntfs_mbstoucs(hiberfile, &unicode, 0);
+	if (unicode_len < 0) {
+		Dperror("Couldn't convert 'hiberfil.sys' to Unicode");
+		goto out;
+	}
+
+	inode = ntfs_inode_lookup_by_name(ni_root, unicode, unicode_len);
+	if (inode == (u64)-1) {
+		Dprintf("Couldn't find file '%s'.\n", hiberfile);
+		goto out;
+	}
+
+	inode = MREF(inode);
+	ni_hibr = ntfs_inode_open(vol, inode);
+	if (!ni_hibr) {
+		Dprintf("Couldn't open inode %lld.\n", (long long)inode);
+		goto out;
+	}
+out:
+	ntfs_inode_close(ni_root);
+	if (unicode)
+		free(unicode);
+	return ni_hibr;
+}
+
+
+#define NTFS_HIBERFILE_HEADER_SIZE	4096
+
+/**
+ * ntfs_volume_check_hiberfile - check hiberfil.sys whether Windows is
+ *                               hibernated on the target volume
+ * @vol:    volume on which to check hiberfil.sys
+ *
+ * Return:  0 if Windows isn't hibernated for sure
+ *         -1 otherwise and errno is set to the appropriate value
+ */
+static int ntfs_volume_check_hiberfile(ntfs_volume *vol)
+{
+	ntfs_inode *ni;
+	ntfs_attr *na = NULL;
+	int i, bytes_read, ret = -1;
+	char *buf = NULL;
+
+	ni = ntfs_hiberfile_open(vol);
+	if (!ni) {
+		if (errno == ENOENT)
+			return 0;
+		return -1;
+	}
+
+	buf = malloc(NTFS_HIBERFILE_HEADER_SIZE);
+	if (!buf) {
+		Dperror("Error allocating memory for hiberfile.sys header");
+		goto out;
+	}
+
+	na = ntfs_attr_open(ni, AT_DATA, AT_UNNAMED, 0);
+	if (!na) {
+		Dperror("Failed to open hiberfil.sys data attribute");
+		goto out;
+	}
+
+	bytes_read = ntfs_attr_pread(na, 0, NTFS_HIBERFILE_HEADER_SIZE, buf);
+	if (bytes_read == -1) {
+		Dperror("Failed to read hiberfil.sys");
+		goto out;
+	}
+	if (bytes_read < NTFS_HIBERFILE_HEADER_SIZE) {
+		Dprintf("Hibernated non-system partition, refused to mount!\n");
+		errno = EPERM;
+		goto out;
+	}
+	if (memcmp(buf, "hibr", 4) == 0) {
+		Dprintf("Windows is hibernated, refused to mount!\n");
+		errno = EPERM;
+		goto out;
+	}
+	for (i = 0; i < NTFS_HIBERFILE_HEADER_SIZE; i++) {
+		if (buf[i]) {
+			Dprintf("Windows is hibernated, won't mount!\n");
+			errno = EPERM;
+			goto out;
+		}
+	}
+        /* All right, all header bytes are zero */
+	ret = 0;
+out:
+	if (na)
+		ntfs_attr_close(na);
+	if (buf)
+		free(buf);
+	ntfs_inode_close(ni);
+	return ret;
+}
+
 /**
  * ntfs_device_mount - open ntfs volume
  * @dev:	device to open
@@ -733,12 +916,16 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long rwflag)
 		errno = EINVAL;
 		goto error_exit;
 	}
-	vol->upcase_len = na->data_size >> 1;
-	vol->upcase = (ntfschar*)malloc(na->data_size);
-	if (!vol->upcase) {
-		Dputs(FAILED);
-		Dputs("Not enough memory to load $UpCase.");
-		goto error_exit;
+	if (vol->upcase_len != na->data_size >> 1) {
+		vol->upcase_len = na->data_size >> 1;
+		/* Throw away default table. */
+		free(vol->upcase);
+		vol->upcase = (ntfschar*)malloc(na->data_size);
+		if (!vol->upcase) {
+			Dputs(FAILED);
+			Dputs("Not enough memory to load $UpCase.");
+			goto error_exit;
+		}
 	}
 	/* Read in the $DATA attribute value into the buffer. */
 	l = ntfs_attr_pread(na, 0, na->data_size, vol->upcase);
@@ -760,14 +947,14 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long rwflag)
 	 * vol structure accordingly.
 	 */
 	Dprintf("Loading $Volume... ");
-	ni = ntfs_inode_open(vol, FILE_Volume);
-	if (!ni) {
+	vol->vol_ni = ntfs_inode_open(vol, FILE_Volume);
+	if (!vol->vol_ni) {
 		Dputs(FAILED);
 		Dperror("Failed to open inode");
 		goto error_exit;
 	}
 	/* Get a search context for the $Volume/$VOLUME_INFORMATION lookup. */
-	ctx = ntfs_attr_get_search_ctx(ni, NULL);
+	ctx = ntfs_attr_get_search_ctx(vol->vol_ni, NULL);
 	if (!ctx) {
 		Dputs(FAILED);
 		Dperror("Failed to allocate attribute search context");
@@ -877,9 +1064,6 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long rwflag)
 	Dputs(OK);
 	ntfs_attr_put_search_ctx(ctx);
 	ctx = NULL;
-	if (ntfs_inode_close(ni))
-		Dperror("Failed to close inode, leaking memory");
-
 	/* Now load the attribute definitions from $AttrDef. */
 	Dprintf("Loading $AttrDef... ");
 	ni = ntfs_inode_open(vol, FILE_AttrDef);
@@ -924,6 +1108,16 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long rwflag)
 	ntfs_attr_close(na);
 	if (ntfs_inode_close(ni))
 		Dperror("Failed to close inode, leaking memory");
+	/*
+	 * Check for dirty logfile and hibernated Windows.
+	 * We care only about read-write mounts.
+	 */
+	if (!(rwflag & MS_RDONLY)) {
+		if (ntfs_volume_check_logfile(vol) < 0)
+			goto error_exit;
+		if (ntfs_volume_check_hiberfile(vol) < 0)
+			goto error_exit;
+	}
 
 	return vol;
 io_error_exit:
@@ -971,20 +1165,27 @@ ntfs_volume *ntfs_mount(const char *name __attribute__((unused)),
 {
 #ifndef NO_NTFS_DEVICE_DEFAULT_IO_OPS
 	struct ntfs_device *dev;
+	ntfs_volume *vol;
 
 	/* Allocate an ntfs_device structure. */
 	dev = ntfs_device_alloc(name, 0, &ntfs_device_default_io_ops, NULL);
 	if (!dev)
 		return NULL;
 	/* Call ntfs_device_mount() to do the actual mount. */
-	return ntfs_device_mount(dev, rwflag);
+	vol = ntfs_device_mount(dev, rwflag);
+	if (!vol) {
+		int eo = errno;
+		ntfs_device_free(dev);
+		errno = eo;
+	}
+	return vol;
 #else
 	/*
 	 * ntfs_mount() makes no sense if NO_NTFS_DEVICE_DEFAULT_IO_OPS is
 	 * defined as there are no device operations available in libntfs in
 	 * this case.
 	 */
-	errno = ENOTSUP;
+	errno = EOPNOTSUPP;
 	return NULL;
 #endif
 }
@@ -1063,6 +1264,17 @@ int ntfs_umount(ntfs_volume *vol,
 }
 
 #ifdef HAVE_MNTENT_H
+
+#ifndef HAVE_REALPATH
+/* If there is no realpath() on the system, provide a dummy one. */
+static char *realpath(const char *path, char *resolved_path)
+{
+	strncpy(resolved_path, path, PATH_MAX);
+	resolved_path[PATH_MAX] = '\0';
+	return resolved_path;
+}
+#endif
+
 /**
  * Internal:
  *
@@ -1076,16 +1288,35 @@ int ntfs_umount(ntfs_volume *vol,
 static int ntfs_mntent_check(const char *file, unsigned long *mnt_flags)
 {
 	struct mntent *mnt;
+	char *real_file = NULL, *real_fsname = NULL;
 	FILE *f;
+	int err = 0;
 
-	if (!(f = setmntent(MOUNTED, "r")))
+	real_file = malloc(PATH_MAX + 1);
+	if (!real_file)
 		return -1;
-	while ((mnt = getmntent(f)))
-		if (!strcmp(file, mnt->mnt_fsname))
+	real_fsname = malloc(PATH_MAX + 1);
+	if (!real_fsname) {
+		err = errno;
+		goto exit;
+	}
+	if (!realpath(file, real_file)) {
+		err = errno;
+		goto exit;
+	}
+	if (!(f = setmntent(MOUNTED, "r"))) {
+		err = errno;
+		goto exit;
+	}
+	while ((mnt = getmntent(f))) {
+		if (!realpath(mnt->mnt_fsname, real_fsname))
+			continue;
+		if (!strcmp(real_file, real_fsname))
 			break;
+	}
 	endmntent(f);
 	if (!mnt)
-		return 0;
+		goto exit;
 	*mnt_flags = NTFS_MF_MOUNTED;
 	if (!strcmp(mnt->mnt_dir, "/"))
 		*mnt_flags |= NTFS_MF_ISROOT;
@@ -1093,6 +1324,14 @@ static int ntfs_mntent_check(const char *file, unsigned long *mnt_flags)
 	if (hasmntopt(mnt, "ro") && !hasmntopt(mnt, "rw"))
 		*mnt_flags |= NTFS_MF_READONLY;
 #endif
+exit:
+	free(real_file);
+	if (real_fsname)
+		free(real_fsname);
+	if (err) {
+		errno = err;
+		return -1;
+	}
 	return 0;
 }
 #endif /* HAVE_MNTENT_H */
@@ -1118,7 +1357,7 @@ static int ntfs_mntent_check(const char *file, unsigned long *mnt_flags)
  * Finally, check if the file system is mounted read-only, and if so set the
  * NTFS_MF_READONLY flag in *@mnt_flags.
  *
- * On sucess return 0 with *@mnt_flags set to the ntfs mount flags.
+ * On success return 0 with *@mnt_flags set to the ntfs mount flags.
  *
  * On error return -1 with errno set to the error code.
  */
@@ -1141,13 +1380,13 @@ int ntfs_check_if_mounted(const char *file __attribute__((unused)),
  * Version 1.1 and 1.2 are used by Windows NT3.x and NT4.
  * Version 2.x is used by Windows 2000 Betas.
  * Version 3.0 is used by Windows 2000.
- * Version 3.1 is used by Windows XP and Windows Server 2003.
+ * Version 3.1 is used by Windows XP, Windows Server 2003 and Longhorn.
  *
  * Return 0 if NTFS version is supported otherwise -1 with errno set.
  *
  * The following error codes are defined:
- *	ENOTSUP   Unknown NTFS version
- *	EINVAL	  Invalid argument
+ *	EOPNOTSUPP - Unknown NTFS version
+ *	EINVAL	   - Invalid argument
  */
 int ntfs_version_is_supported(ntfs_volume *vol)
 {
@@ -1170,7 +1409,7 @@ int ntfs_version_is_supported(ntfs_volume *vol)
 	if (NTFS_V3_0(major, minor) || NTFS_V3_1(major, minor))
 		return 0;
 
-	errno = ENOTSUP;
+	errno = EOPNOTSUPP;
 	return -1;
 }
 
@@ -1193,8 +1432,6 @@ int ntfs_logfile_reset(ntfs_volume *vol)
 {
 	ntfs_inode *ni;
 	ntfs_attr *na;
-	s64 len, pos, count;
-	char buf[NTFS_BUF_SIZE];
 	int eo;
 
 	if (!vol) {
@@ -1208,73 +1445,28 @@ int ntfs_logfile_reset(ntfs_volume *vol)
 	}
 
 	if ((na = ntfs_attr_open(ni, AT_DATA, AT_UNNAMED, 0)) == NULL) {
+		eo = errno;
 		Dperror("Failed to open $FILE_LogFile/$DATA\n");
 		goto error_exit;
 	}
 
-	/* The $DATA attribute of the $LogFile has to be non-resident. */
-	if (!NAttrNonResident(na)) {
-		Dprintf("$LogFile $DATA attribute is resident!?!\n");
-		errno = EIO;
-		goto io_error_exit;
+	if (ntfs_empty_logfile(na)) {
+		eo = errno;
+		Dperror("Failed to empty $FILE_LogFile/$DATA\n");
+		ntfs_attr_close(na);
+		goto error_exit;
 	}
-
-	/* Get length of $LogFile contents. */
-	len = na->data_size;
-	if (!len) {
-		Dprintf("$LogFile has zero length, no disk write needed.\n");
-		return 0;
-	}
-
-	/* Read $LogFile until its end. We do this as a check for correct
-	   length thus making sure we are decompressing the mapping pairs
-	   array correctly and hence writing below is safe as well. */
-	pos = 0;
-	while ((count = ntfs_attr_pread(na, pos, NTFS_BUF_SIZE, buf)) > 0)
-		pos += count;
-
-	if (count == -1 || pos != len) {
-		Dprintf("Amount of $LogFile data read does not "
-			"correspond to expected length!");
-		if (count != -1)
-			errno = EIO;
-		goto io_error_exit;
-	}
-
-	/* Fill the buffer with 0xff's. */
-	memset(buf, -1, NTFS_BUF_SIZE);
-
-	/* Set the $DATA attribute. */
-	pos = 0;
-	while ((count = len - pos) > 0) {
-		if (count > NTFS_BUF_SIZE)
-			count = NTFS_BUF_SIZE;
-
-		if ((count = ntfs_attr_pwrite(na, pos, count, buf)) <= 0) {
-			Dprintf("Failed to set the $LogFile attribute value.");
-			if (count != -1)
-				errno = EIO;
-			goto io_error_exit;
-		}
-		pos += count;
-	}
-
 	ntfs_attr_close(na);
 	return ntfs_inode_close(ni);
 
-io_error_exit:
-	eo = errno;
-	ntfs_attr_close(na);
-	errno = eo;
 error_exit:
-	eo = errno;
 	ntfs_inode_close(ni);
 	errno = eo;
 	return -1;
 }
 
 /**
- * ntfs_volume_set_flags - set the flags of an ntfs volume
+ * ntfs_volume_write_flags - set the flags of an ntfs volume
  * @vol:	ntfs volume where we set the volume flags
  * @flags:	new flags
  *
@@ -1283,37 +1475,22 @@ error_exit:
  *
  * Return 0 if successful and -1 if not with errno set to the error code.
  */
-int ntfs_volume_set_flags(ntfs_volume *vol, const u16 flags)
+int ntfs_volume_write_flags(ntfs_volume *vol, const u16 flags)
 {
-	MFT_RECORD *m = NULL;
-	ATTR_RECORD *r;
+	ATTR_RECORD *a;
 	VOLUME_INFORMATION *c;
 	ntfs_attr_search_ctx *ctx;
 	int ret = -1;	/* failure */
 
-	if (!vol) {
+	if (!vol || !vol->vol_ni) {
 		errno = EINVAL;
 		return -1;
 	}
-
-	if (ntfs_file_record_read(vol, FILE_Volume, &m, NULL)) {
-		Dperror("Failed to read $Volume");
-		return -1;
-	}
-
-	/* Sanity check */
-	if (!(m->flags & MFT_RECORD_IN_USE)) {
-		Dprintf("Error: $Volume has been deleted. Cannot "
-			"handle this yet. Run chkdsk to fix this.\n");
-		errno = EIO;
-		goto err_exit;
-	}
-
 	/* Get a pointer to the volume information attribute. */
-	ctx = ntfs_attr_get_search_ctx(NULL, m);
+	ctx = ntfs_attr_get_search_ctx(vol->vol_ni, NULL);
 	if (!ctx) {
 		Dperror("Failed to allocate attribute search context");
-		goto err_exit;
+		return -1;
 	}
 	if (ntfs_attr_lookup(AT_VOLUME_INFORMATION, AT_UNNAMED, 0, 0, 0, NULL,
 			0, ctx)) {
@@ -1321,40 +1498,36 @@ int ntfs_volume_set_flags(ntfs_volume *vol, const u16 flags)
 				"$Volume!");
 		goto err_out;
 	}
-	r = ctx->attr;
+	a = ctx->attr;
 	/* Sanity check. */
-	if (r->non_resident) {
+	if (a->non_resident) {
 		Dputs("Error: Attribute $VOLUME_INFORMATION must be resident "
 				"(and it isn't)!");
 		errno = EIO;
 		goto err_out;
 	}
 	/* Get a pointer to the value of the attribute. */
-	c = (VOLUME_INFORMATION*)(le16_to_cpu(r->value_offset) + (char*)r);
+	c = (VOLUME_INFORMATION*)(le16_to_cpu(a->value_offset) + (char*)a);
 	/* Sanity checks. */
-	if ((char*)c + le32_to_cpu(r->value_length) >
-			le16_to_cpu(m->bytes_in_use) + (char*)m ||
-			le16_to_cpu(r->value_offset) +
-			le32_to_cpu(r->value_length) > le32_to_cpu(r->length)) {
+	if ((char*)c + le32_to_cpu(a->value_length) >
+			le16_to_cpu(ctx->mrec->bytes_in_use) +
+			(char*)ctx->mrec || le16_to_cpu(a->value_offset) +
+			le32_to_cpu(a->value_length) > le32_to_cpu(a->length)) {
 		Dputs("Error: Attribute $VOLUME_INFORMATION in $Volume is "
 				"corrupt!");
 		errno = EIO;
 		goto err_out;
 	}
 	/* Set the volume flags. */
-	vol->flags = c->flags = cpu_to_le16(flags);
-
-	if (ntfs_mft_record_write(vol, FILE_Volume, m)) {
+	vol->flags = c->flags = flags & VOLUME_FLAGS_MASK;
+	/* Write them to disk. */
+	ntfs_inode_mark_dirty(vol->vol_ni);
+	if (ntfs_inode_sync(vol->vol_ni)) {
 		Dperror("Error writing $Volume");
 		goto err_out;
 	}
-
 	ret = 0; /* success */
 err_out:
 	ntfs_attr_put_search_ctx(ctx);
-err_exit:
-	if (m)
-		free(m);
 	return ret;
 }
-
