@@ -1,8 +1,8 @@
 /**
  * ntfsclone - Part of the Linux-NTFS project.
  *
- * Copyright (c) 2003-2004 Szabolcs Szakacsits
- * Copyright (c) 2004 Anton Altaparmakov
+ * Copyright (c) 2003-2005 Szabolcs Szakacsits
+ * Copyright (c) 2004-2005 Anton Altaparmakov
  * Special image format support copyright (c) 2004 Per Olofsson
  *
  * Clone NTFS data and/or metadata to a sparse file, image, device or stdout.
@@ -15,20 +15,42 @@
 
 #include "config.h"
 
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
+#endif
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
+#endif
+#ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
+#endif
+#ifdef HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
+#endif
 #ifdef HAVE_SYS_VFS_H
 #	include <sys/vfs.h>
 #endif
+#ifdef HAVE_FCNTL_H
 #include <fcntl.h>
+#endif
+#ifdef HAVE_STDARG_H
 #include <stdarg.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
+#ifdef HAVE_ERRNO_H
 #include <errno.h>
+#endif
+#ifdef HAVE_GETOPT_H
 #include <getopt.h>
+#endif
 
 #include "debug.h"
 #include "types.h"
@@ -43,6 +65,7 @@
 #include "inode.h"
 #include "runlist.h"
 #include "utils.h"
+#include "version.h"
 
 #if defined(linux) && defined(_IO) && !defined(BLKGETSIZE)
 #define BLKGETSIZE	_IO(0x12,96)  /* Get device size in 512-byte blocks. */
@@ -53,6 +76,14 @@
 
 static const char *EXEC_NAME = "ntfsclone";
 
+static const char *bad_sectors_warning_msg =
+"*************************************************************************\n"
+"* WARNING: The disk has bad sector. This means physical damage on the   *\n"
+"* disk surface caused by deterioration, manufacturing faults or other   *\n"
+"* reason. The reliability of the disk may stay stable or degrade fast.  *\n"
+"* Use the --rescue option to efficiently save as much data as possible! *\n"
+"*************************************************************************\n";
+
 struct {
 	int verbose;
 	int quiet;
@@ -60,8 +91,10 @@ struct {
 	int force;
 	int overwrite;
 	int std_out;
-	int blkdev_out;		/* output file is block device */   
+	int blkdev_out;		/* output file is block device */
 	int metadata_only;
+	int ignore_fs_check;
+	int rescue;
 	int save_image;
 	int restore_image;
 	char *output;
@@ -103,12 +136,12 @@ int fd_in;
 int fd_out;
 FILE *msg_out = NULL;
 
-int nr_used_mft_records = 0;
 int wipe = 0;
-int wiped_unused_mft_data = 0;
-int wiped_unused_mft = 0;
-int wiped_resident_data = 0;
-int wiped_timestamp_data = 0;
+unsigned int nr_used_mft_records   = 0;
+unsigned int wiped_unused_mft_data = 0;
+unsigned int wiped_unused_mft      = 0;
+unsigned int wiped_resident_data   = 0;
+unsigned int wiped_timestamp_data  = 0;
 
 #define IMAGE_MAGIC "\0ntfsclone-image"
 #define IMAGE_MAGIC_SIZE 16
@@ -132,6 +165,7 @@ struct {
 #define LAST_METADATA_INODE  	11
 
 #define NTFS_MAX_CLUSTER_SIZE 	65536
+#define NTFS_SECTOR_SIZE 	  512
 
 #define rounded_up_division(a, b) (((a) + (b - 1)) / (b))
 
@@ -207,24 +241,26 @@ static int perr_exit(const char *fmt, ...)
 static void usage(void)	__attribute__((noreturn));
 static void usage(void)
 {
-	Eprintf("\nUsage: %s [options] source\n"
+	Eprintf("\nUsage: %s [OPTIONS] SOURCE\n"
 		"    Efficiently clone NTFS to a sparse file, image, device or standard output.\n"
 		"\n"
 		"    -o, --output FILE      Clone NTFS to the non-existent FILE\n"
 		"    -O, --overwrite FILE   Clone NTFS to FILE, overwriting if exists\n"
-		"    -m, --metadata         Clone *only* metadata (for NTFS experts)\n"
 		"    -s, --save-image       Save to the special image format\n"
 		"    -r, --restore-image    Restore from the special image format\n"
+		"        --rescue           Continue after disk read errors\n"
+		"    -m, --metadata         Clone *only* metadata (for NTFS experts)\n"
+		"        --ignore-fs-check  Ignore the filesystem check result\n"
 		"    -f, --force            Force to progress (DANGEROUS)\n"
 		"    -h, --help             Display this help\n"
 #ifdef DEBUG
 		"    -d, --debug            Show debug information\n"
 #endif
 		"\n"
-		"    If FILE is '-' then send NTFS data to stdout replacing non used\n"
-		"    NTFS and partition space with zeros.\n"
+		"    If FILE is '-' then send the image to the standard output. If SOURCE is '-'\n"
+		"    and --restore-image is used then read the image from the standard input.\n"
 		"\n", EXEC_NAME);
-	Eprintf("%s%s\n", ntfs_bugs, ntfs_home);
+	Eprintf("%s%s", ntfs_bugs, ntfs_home);
 	exit(1);
 }
 
@@ -234,15 +270,17 @@ static void parse_options(int argc, char **argv)
 	static const char *sopt = "-dfhmo:O:rs";
 	static const struct option lopt[] = {
 #ifdef DEBUG
-		{ "debug",	   no_argument,		NULL, 'd' },
+		{ "debug",	      no_argument,	 NULL, 'd' },
 #endif
-		{ "force",	   no_argument,		NULL, 'f' },
-		{ "help",	   no_argument,		NULL, 'h' },
-		{ "metadata",	   no_argument,		NULL, 'm' },
-		{ "output",	   required_argument,	NULL, 'o' },
-		{ "overwrite",	   required_argument,	NULL, 'O' },
-		{ "restore-image", no_argument,		NULL, 'r' },
-		{ "save-image",	   no_argument,		NULL, 's' },
+		{ "force",	      no_argument,	 NULL, 'f' },
+		{ "help",	      no_argument,	 NULL, 'h' },
+		{ "metadata",	      no_argument,	 NULL, 'm' },
+		{ "output",	      required_argument, NULL, 'o' },
+		{ "overwrite",	      required_argument, NULL, 'O' },
+		{ "restore-image",    no_argument,	 NULL, 'r' },
+		{ "ignore-fs-check",  no_argument,	 NULL, 'C' },
+		{ "rescue",           no_argument,	 NULL, 'R' },
+		{ "save-image",	      no_argument,	 NULL, 's' },
 		{ NULL, 0, NULL, 0 }
 	};
 
@@ -279,6 +317,12 @@ static void parse_options(int argc, char **argv)
 		case 'r':
 			opt.restore_image++;
 			break;
+		case 'C':
+			opt.ignore_fs_check++;
+			break;
+		case 'R':
+			opt.rescue++;
+			break;
 		case 's':
 			opt.save_image++;
 			break;
@@ -287,7 +331,7 @@ static void parse_options(int argc, char **argv)
 			usage();
 		}
 	}
-	
+
 	if (opt.output == NULL) {
 		err_printf("You must specify an output file.\n");
 		usage();
@@ -308,20 +352,24 @@ static void parse_options(int argc, char **argv)
 	if (opt.metadata_only && opt.restore_image)
 		err_exit("Restoring only metadata from an image is not "
 			 "supported!\n");
-	
+
 	if (opt.metadata_only && opt.std_out)
 		err_exit("Cloning only metadata to stdout isn't supported!\n");
 
+	if (opt.ignore_fs_check && !opt.metadata_only)
+		err_exit("Filesystem check can be ignored only for metadata "
+			 "cloning!\n");
+
 	if (opt.save_image && opt.restore_image)
 		err_exit("Saving and restoring an image at the same time "
 			 "is not supported!\n");
-	
+
 	if (!opt.std_out) {
 		struct stat st;
-		
+
 		if (stat(opt.output, &st) == -1) {
 			if (errno != ENOENT)
-				perr_exit("Couldn't access '%s'", opt.output); 
+				perr_exit("Couldn't access '%s'", opt.output);
 		} else {
 			if (!opt.overwrite)
 				err_exit("Output file '%s' already exists.\n"
@@ -335,16 +383,16 @@ static void parse_options(int argc, char **argv)
 					     "block device isn't supported!\n");
 			}
 		}
-	}	
-		
+	}
+
 	msg_out = stdout;
-	
+
 	/* FIXME: this is a workaround for loosing debug info if stdout != stderr
 	   and for the uncontrollable verbose messages in libntfs. Ughhh. */
 	if (opt.std_out)
 		msg_out = stderr;
 	else if (opt.debug)
-		stderr = stdout; 
+		stderr = stdout;
 	else
 		if (!(stderr = fopen("/dev/null", "rw")))
 			perr_exit("Couldn't open /dev/null");
@@ -372,36 +420,23 @@ static void progress_update(struct progress_bar *p, u64 current)
 	fflush(msg_out);
 }
 
-/**
- * nr_clusters_to_bitmap_byte_size
- *
- * Take the number of clusters in the volume and calculate the size of $Bitmap.
- * The size will always be a multiple of 8 bytes.
- */
-static s64 nr_clusters_to_bitmap_byte_size(s64 nr_clusters)
-{
-	s64 bm_bsize;
-
-	bm_bsize = rounded_up_division(nr_clusters, 8);
-
-	bm_bsize = (bm_bsize + 7) & ~7;
-	Dprintf("Bitmap byte size  : %lld (%lld clusters)\n",
-	       bm_bsize, rounded_up_division(bm_bsize, vol->cluster_size));
-
-	return bm_bsize;
-}
-
 static s64 is_critical_metadata(ntfs_walk_clusters_ctx *image, runlist *rl)
 {
 	s64 inode = image->ni->mft_no;
 
 	if (inode <= LAST_METADATA_INODE) {
+
+		/* Don't save bad sectors (both $Bad and unnamed are ignored */
+		if (inode == FILE_BadClus && image->ctx->attr->type == AT_DATA)
+			return 0;
+
 		if (inode != FILE_LogFile)
 			return rl->length;
 
 		if (image->ctx->attr->type == AT_DATA) {
-			/* Save at least the first 8 KiB of FILE_LogFile */
-			s64 s = (s64)8192 - rl->vcn * vol->cluster_size;
+
+			/* Save at least the first 16 KiB of FILE_LogFile */
+			s64 s = (s64)16384 - rl->vcn * vol->cluster_size;
 			if (s > 0) {
 				s = rounded_up_division(s, vol->cluster_size);
 				if (rl->length < s)
@@ -411,7 +446,7 @@ static s64 is_critical_metadata(ntfs_walk_clusters_ctx *image, runlist *rl)
 			return 0;
 		}
 	}
-	
+
 	if (image->ctx->attr->type != AT_DATA)
 		return rl->length;
 
@@ -422,8 +457,8 @@ static s64 is_critical_metadata(ntfs_walk_clusters_ctx *image, runlist *rl)
 static int io_all(void *fd, void *buf, int count, int do_write)
 {
 	int i;
-	struct ntfs_device *dev = (struct ntfs_device *)fd;
-	
+	struct ntfs_device *dev = fd;
+
 	while (count > 0) {
 		if (do_write)
 			i = write(*(int *)fd, buf, count);
@@ -443,22 +478,63 @@ static int io_all(void *fd, void *buf, int count, int do_write)
 }
 
 
-static void copy_cluster(void)
+static void rescue_sector(void *fd, off_t pos, void *buff)
+{
+	const char *badsector_magic = "BadSectoR\0";
+	struct ntfs_device *dev = fd;
+
+	if (opt.restore_image) {
+		if (lseek(*(int *)fd, pos, SEEK_SET) == (off_t)-1)
+			perr_exit("lseek");
+	} else {
+		if (vol->dev->d_ops->seek(dev, pos, SEEK_SET) == (off_t)-1)
+			perr_exit("seek input");
+	}
+
+	if (read_all(fd, buff, NTFS_SECTOR_SIZE) == -1) {
+		Printf("WARNING: Can't read sector at %llu, lost data.\n",
+		       (unsigned long long)pos);
+		memset(buff, '?', NTFS_SECTOR_SIZE);
+		memmove(buff, badsector_magic, sizeof(badsector_magic));
+	}
+}
+
+
+static void copy_cluster(int rescue, u64 rescue_lcn)
 {
 	char buff[NTFS_MAX_CLUSTER_SIZE]; /* overflow checked at mount time */
-	u32 csize = opt.restore_image ? image_hdr.cluster_size
-		: vol->cluster_size;
+	/* vol is NULL if opt.restore_image is set */
+	u32 csize = image_hdr.cluster_size;
+	void *fd = (void *)&fd_in;
+	off_t rescue_pos;
 
-	if (read_all(opt.restore_image ? (void *)&fd_in : vol->dev, buff,
-		     csize) == -1)
-		perr_exit("read_all");
+	if (!opt.restore_image) {
+		csize = vol->cluster_size;
+		fd = vol->dev;
+	}
+
+	rescue_pos = (off_t)(rescue_lcn * csize);
+
+	if (read_all(fd, buff, csize) == -1) {
+
+		if (errno != EIO)
+			perr_exit("read_all");
+		else if (rescue){
+			u32 i;
+			for (i = 0; i < csize; i += NTFS_SECTOR_SIZE)
+				rescue_sector(fd, rescue_pos + i, buff + i);
+		} else {
+			Printf("%s", bad_sectors_warning_msg);
+			err_exit("Disk is faulty, can't make full backup!");
+		}
+	}
 
 	if (opt.save_image) {
 		char cmd = 1;
 		if (write_all(&fd_out, &cmd, sizeof(cmd)) == -1)
 			perr_exit("write_all");
 	}
-	
+
 	if (write_all(&fd_out, buff, csize) == -1) {
 		int err = errno;
 		perr_printf("Write failed");
@@ -474,15 +550,15 @@ static void copy_cluster(void)
 static void lseek_to_cluster(s64 lcn)
 {
 	off_t pos;
-	
+
 	pos = (off_t)(lcn * vol->cluster_size);
-	
+
 	if (vol->dev->d_ops->seek(vol->dev, pos, SEEK_SET) == (off_t)-1)
 		perr_exit("lseek input");
 
 	if (opt.std_out || opt.save_image)
 		return;
-	
+
 	if (lseek(fd_out, pos, SEEK_SET) == (off_t)-1)
 		perr_exit("lseek output");
 }
@@ -494,7 +570,7 @@ static void image_skip_clusters(s64 count)
 
 		buff[0] = 0;
 		memcpy(buff + 1, &count, sizeof(count));
-		
+
 		if (write_all(&fd_out, buff, sizeof(buff)) == -1)
 			perr_exit("write_all");
 	}
@@ -503,24 +579,23 @@ static void image_skip_clusters(s64 count)
 static void dump_clusters(ntfs_walk_clusters_ctx *image, runlist *rl)
 {
 	s64 i, len; /* number of clusters to copy */
-	
+
 	if (opt.std_out || !opt.metadata_only)
 		return;
-	
+
 	if (!(len = is_critical_metadata(image, rl)))
 		return;
-	
+
 	lseek_to_cluster(rl->lcn);
-	
+
 	/* FIXME: this could give pretty suboptimal performance */
 	for (i = 0; i < len; i++)
-		copy_cluster();
+		copy_cluster(opt.rescue, rl->lcn + i);
 }
 
 static void clone_ntfs(u64 nr_clusters)
 {
-	s64 i, pos, count, last_cl;
-	u8 bm[NTFS_BUF_SIZE];
+	u64 cl, last_cl;  /* current and last used cluster */
 	void *buf;
 	u32 csize = vol->cluster_size;
 	u64 p_counter = 0;
@@ -532,56 +607,34 @@ static void clone_ntfs(u64 nr_clusters)
 		Printf("Cloning NTFS ...\n");
 
 	if ((buf = calloc(1, csize)) == NULL)
-		perr_exit("dump_to_stdout");
+		perr_exit("clone_ntfs");
 
 	progress_init(&progress, p_counter, nr_clusters, 100);
 
 	if (opt.save_image) {
-		if (write_all(&fd_out, &image_hdr, sizeof(image_hdr))
-		    == -1)
+		if (write_all(&fd_out, &image_hdr, sizeof(image_hdr)) == -1)
 			perr_exit("write_all");
 	}
-	
-	pos = 0;
-	last_cl = 0;
-	while (1) {
-		count = ntfs_attr_pread(vol->lcnbmp_na, pos, NTFS_BUF_SIZE, bm);
-		if (count == -1)
-			perr_exit("Couldn't read $Bitmap (pos = %lld)\n", pos);
 
-		if (count == 0) {
-			image_skip_clusters(pos * 8 - last_cl - 1);
-			return;
+	for (last_cl = cl = 0; cl < (u64)vol->nr_clusters; cl++) {
+
+		if (ntfs_bit_get(lcn_bitmap.bm, cl)) {
+			progress_update(&progress, ++p_counter);
+			lseek_to_cluster(cl);
+			image_skip_clusters(cl - last_cl - 1);
+
+			copy_cluster(opt.rescue, cl);
+			last_cl = cl;
+			continue;
 		}
 
-		for (i = 0; i < count; i++, pos++) {
-			s64 cl;  /* current cluster */	  
-
-			for (cl = pos * 8; cl < (pos + 1) * 8; cl++) {
-
-				if (cl > vol->nr_clusters - 1) {
-					image_skip_clusters(cl - last_cl - 1);
-					return;
-				}
-				
-				if (ntfs_bit_get(bm, i * 8 + cl % 8)) {
-					progress_update(&progress, ++p_counter);
-					lseek_to_cluster(cl);
-					image_skip_clusters(cl - last_cl - 1);
-					
-					copy_cluster();
-					last_cl = cl;
-					continue;
-				}
-				
-				if (opt.std_out && !opt.save_image) {
-					progress_update(&progress, ++p_counter);
-					if (write_all(&fd_out, buf, csize) == -1)
-						perr_exit("write_all");
-				}
-			}
+		if (opt.std_out && !opt.save_image) {
+			progress_update(&progress, ++p_counter);
+			if (write_all(&fd_out, buf, csize) == -1)
+				perr_exit("write_all");
 		}
 	}
+	image_skip_clusters(cl - last_cl - 1);
 }
 
 static void write_empty_clusters(s32 csize, s64 count,
@@ -589,7 +642,7 @@ static void write_empty_clusters(s32 csize, s64 count,
 {
 	s64 i;
 	char buff[NTFS_MAX_CLUSTER_SIZE];
-	
+
 	memset(buff, 0, csize);
 
 	for (i = 0; i < count; i++) {
@@ -625,11 +678,11 @@ static void restore_image(void)
 			else {
 				if (lseek(fd_out, count * csize, SEEK_CUR)
 				    == (off_t)-1)
-					perr_exit("lseek output");
+					perr_exit("restore_image: lseek");
 			}
 			pos += count;
 		} else if (cmd == 1) {
-			copy_cluster();
+			copy_cluster(0, 0);
 			pos++;
 			progress_update(&progress, ++p_counter);
 		} else
@@ -654,14 +707,14 @@ do {								\
 static void wipe_timestamps(ntfs_walk_clusters_ctx *image)
 {
 	ATTR_RECORD *a = image->ctx->attr;
-	
+
 	if (image->ni->mft_no <= LAST_METADATA_INODE)
 		return;
 
-	if (a->type == AT_FILE_NAME) 
+	if (a->type == AT_FILE_NAME)
 		WIPE_TIMESTAMPS(FILE_NAME_ATTR, a);
 
-	else if (a->type == AT_STANDARD_INFORMATION)	
+	else if (a->type == AT_STANDARD_INFORMATION)
 		WIPE_TIMESTAMPS(STANDARD_INFORMATION, a);
 }
 
@@ -674,23 +727,44 @@ static void wipe_resident_data(ntfs_walk_clusters_ctx *image)
 
 	a = image->ctx->attr;
 	p = (u8*)a + le16_to_cpu(a->value_offset);
-	
+
 	if (image->ni->mft_no <= LAST_METADATA_INODE)
 		return;
-	
+
 	if (a->type != AT_DATA)
 		return;
-	
+
 	for (i = 0; i < le32_to_cpu(a->value_length); i++) {
 		if (p[i]) {
 			p[i] = 0;
 			n++;
-		}	
-	}		
-			
+		}
+	}
+
 	wiped_resident_data += n;
 }
 
+static void clone_logfile_parts(ntfs_walk_clusters_ctx *image, runlist *rl)
+{
+	s64 offset = 0, lcn, vcn;
+
+	while (1) {
+
+		vcn = offset / image->ni->vol->cluster_size;
+		lcn = ntfs_rl_vcn_to_lcn(rl, vcn);
+		if (lcn < 0)
+			break;
+
+		lseek_to_cluster(lcn);
+		copy_cluster(opt.rescue, lcn);
+
+		if (offset == 0)
+			offset = NTFS_BLOCK_SIZE >> 1;
+		else
+			offset <<= 1;
+	}
+}
+
 static void walk_runs(struct ntfs_walk_cluster *walk)
 {
 	int i, j;
@@ -725,10 +799,10 @@ static void walk_runs(struct ntfs_walk_cluster *walk)
 				 "%llx length %llx\n", ctx->ntfs_ino->mft_no,
 				 (unsigned int)le32_to_cpu(a->type), lcn,
 				 lcn_length);
-		
+
 		if (!wipe)
 			dump_clusters(walk->image, rl + i);
-		
+
 		for (j = 0; j < lcn_length; j++) {
 			u64 k = (u64)lcn + j;
 			if (ntfs_bit_get_and_set(lcn_bitmap.bm, k, 1))
@@ -739,7 +813,11 @@ static void walk_runs(struct ntfs_walk_cluster *walk)
 
 		walk->image->inuse += lcn_length;
 	}
-	
+	if (!wipe && !opt.std_out && opt.metadata_only &&
+	    walk->image->ni->mft_no == FILE_LogFile &&
+	    walk->image->ctx->attr->type == AT_DATA)
+		clone_logfile_parts(walk->image, rl);
+
 	free(rl);
 }
 
@@ -779,15 +857,17 @@ static void compare_bitmaps(struct bitmap *a)
 			perr_exit("Couldn't get $Bitmap $DATA");
 
 		if (count == 0) {
-			if (a->size != pos)
-				err_exit("$Bitmap file size doesn't match "
-					 "calculated size (%lld != %lld)\n",
-					 a->size, pos);
+			if (a->size > pos)
+				err_exit("$Bitmap size is smaller than expected"
+					 " (%lld != %lld)\n", a->size, pos);
 			break;
 		}
 
 		for (i = 0; i < count; i++, pos++) {
-			s64 cl;  /* current cluster */	  
+			s64 cl;  /* current cluster */
+
+			if (a->size <= pos)
+				goto done;
 
 			if (a->bm[pos] == bm[i])
 				continue;
@@ -799,24 +879,32 @@ static void compare_bitmaps(struct bitmap *a)
 				if (bit == ntfs_bit_get(bm, i * 8 + cl % 8))
 					continue;
 
+				if (opt.ignore_fs_check) {
+					lseek_to_cluster(cl);
+					copy_cluster(opt.rescue, cl);
+				}
+
 				if (++mismatch > 10)
 					continue;
 
 				Printf("Cluster accounting failed at %lld "
-						"(0x%llx): %s cluster in "
-						"$Bitmap\n", (long long)cl,
-						(unsigned long long)cl,
-						bit ? "missing" : "extra");
+				       "(0x%llx): %s cluster in $Bitmap\n",
+				       (long long)cl, (unsigned long long)cl,
+				       bit ? "missing" : "extra");
 			}
 		}
 	}
-
+done:
 	if (mismatch) {
-		Printf("Totally %d cluster accounting mismatches.\n", 
-		       mismatch);
+		Printf("Totally %d cluster accounting mismatches.\n", mismatch);
+		if (opt.ignore_fs_check) {
+			Printf("WARNING: The NTFS inconsistency was overruled "
+			       "by the --ignore-fs-check option.\n");
+			return;
+		}
 		err_exit("Filesystem check failed! Windows wasn't shutdown "
 			 "properly or inconsistent\nfilesystem. Please run "
-			 "chkdsk on Windows.\n");
+			 "chkdsk /f on Windows then reboot it TWICE.\n");
 	}
 }
 
@@ -824,14 +912,13 @@ static void compare_bitmaps(struct bitmap *a)
 static int wipe_data(char *p, int pos, int len)
 {
 	int wiped = 0;
-	
-	p += pos;
-	for (; len > 0; len--) {
+
+	for (p += pos; --len >= 0;) {
 		if (p[len]) {
 			p[len] = 0;
 			wiped++;
-		}	
-	}		
+		}
+	}
 
 	return wiped;
 }
@@ -840,11 +927,11 @@ static void wipe_unused_mft_data(ntfs_inode *ni)
 {
 	int unused;
 	MFT_RECORD *m = ni->mrec;
-	
+
 	/* FIXME: broken MFTMirr update was fixed in libntfs, check if OK now */
 	if (ni->mft_no <= LAST_METADATA_INODE)
 		return;
-	
+
 	unused = le32_to_cpu(m->bytes_allocated) - le32_to_cpu(m->bytes_in_use);
 	wiped_unused_mft_data += wipe_data((char *)m,
 			le32_to_cpu(m->bytes_in_use), unused);
@@ -854,11 +941,11 @@ static void wipe_unused_mft(ntfs_inode *ni)
 {
 	int unused;
 	MFT_RECORD *m = ni->mrec;
-	
+
 	/* FIXME: broken MFTMirr update was fixed in libntfs, check if OK now */
 	if (ni->mft_no <= LAST_METADATA_INODE)
 		return;
-	
+
 	unused = le32_to_cpu(m->bytes_in_use) - sizeof(MFT_RECORD);
 	wiped_unused_mft += wipe_data((char *)m, sizeof(MFT_RECORD), unused);
 }
@@ -873,11 +960,12 @@ static int walk_clusters(ntfs_volume *volume, struct ntfs_walk_cluster *walk)
 
 	Printf("Scanning volume ...\n");
 
-	last_mft_rec = volume->nr_mft_records - 1;
+	last_mft_rec = (volume->mft_na->initialized_size >>
+			volume->mft_record_size_bits) - 1;
 	progress_init(&progress, inode, last_mft_rec, 100);
 
 	for (; inode <= last_mft_rec; inode++) {
-		
+
 		int err, deleted_inode;
 		MFT_REF mref = (MFT_REF)inode;
 
@@ -888,14 +976,14 @@ static int walk_clusters(ntfs_volume *volume, struct ntfs_walk_cluster *walk)
 		ni = (ntfs_inode*)calloc(1, sizeof(ntfs_inode));
 		if (!ni)
 			perr_exit("walk_clusters");
-		
+
 		ni->vol = volume;
 
 		err = ntfs_file_record_read(volume, mref, &ni->mrec, NULL);
 		if (err == -1) {
 			free(ni);
 			continue;
-		}	
+		}
 
 	        deleted_inode = !(ni->mrec->flags & MFT_RECORD_IN_USE);
 
@@ -907,16 +995,16 @@ static int walk_clusters(ntfs_volume *volume, struct ntfs_walk_cluster *walk)
 				wipe_unused_mft_data(ni);
 				if (ntfs_mft_record_write(volume, ni->mft_no, ni->mrec))
 					perr_exit("ntfs_mft_record_write");
-			}		
+			}
 		}
 
 	       	if (ni->mrec)
 	        	free(ni->mrec);
 		free(ni);
 
-	        if (deleted_inode) 
+	        if (deleted_inode)
 			continue;
-		
+
 		if ((ni = ntfs_inode_open(volume, mref)) == NULL) {
 			/* FIXME: continue only if it make sense, e.g.
 			   MFT record not in use based on $MFT bitmap */
@@ -933,17 +1021,17 @@ static int walk_clusters(ntfs_volume *volume, struct ntfs_walk_cluster *walk)
 
 		walk->image->ni = ni;
 		walk_attributes(walk);
-out:		
+out:
 		if (wipe) {
 			wipe_unused_mft_data(ni);
 			if (ntfs_mft_record_write(volume, ni->mft_no, ni->mrec))
 				perr_exit("ntfs_mft_record_write");
-		}		
+		}
 
 		if (ntfs_inode_close(ni))
 			perr_exit("ntfs_inode_close for inode %lld", inode);
 	}
-	
+
 	return 0;
 }
 
@@ -967,7 +1055,7 @@ static void bitmap_file_data_fixup(s64 cluster, struct bitmap *bm)
 static void setup_lcn_bitmap(void)
 {
 	/* Determine lcn bitmap byte size and allocate it. */
-	lcn_bitmap.size = nr_clusters_to_bitmap_byte_size(vol->nr_clusters);
+	lcn_bitmap.size = rounded_up_division(vol->nr_clusters, 8);
 
 	if (!(lcn_bitmap.bm = (unsigned char *)calloc(1, lcn_bitmap.size)))
 		perr_exit("Failed to allocate internal buffer");
@@ -1020,16 +1108,16 @@ static void print_image_info(void)
 static void check_if_mounted(const char *device, unsigned long new_mntflag)
 {
 	unsigned long mntflag;
-	
+
 	if (ntfs_check_if_mounted(device, &mntflag))
 		perr_exit("Failed to check '%s' mount state", device);
 
 	if (mntflag & NTFS_MF_MOUNTED) {
 		if (!(mntflag & NTFS_MF_READONLY))
-			err_exit("Device %s is mounted read-write. "
+			err_exit("Device '%s' is mounted read-write. "
 				 "You must 'umount' it first.\n", device);
 		if (!new_mntflag)
-			err_exit("Device %s is mounted. "
+			err_exit("Device '%s' is mounted. "
 				 "You must 'umount' it first.\n", device);
 	}
 }
@@ -1042,12 +1130,12 @@ static void check_if_mounted(const char *device, unsigned long new_mntflag)
 static void mount_volume(unsigned long new_mntflag)
 {
 	check_if_mounted(opt.volume, new_mntflag);
-	
+
 	if (!(vol = ntfs_mount(opt.volume, new_mntflag))) {
 
 		int err = errno;
 
-		perr_printf("ntfs_mount failed");
+		perr_printf("Opening '%s' as NTFS failed", opt.volume);
 		if (err == EINVAL) {
 			Printf("Apparently device '%s' doesn't have a "
 			       "valid NTFS. Maybe you selected\nthe whole "
@@ -1061,7 +1149,7 @@ static void mount_volume(unsigned long new_mntflag)
 		if (opt.force-- <= 0)
 			err_exit("Volume is dirty. Run chkdsk and "
 				 "please try again (or see -f option).\n");
-	
+
 	if (NTFS_MAX_CLUSTER_SIZE < vol->cluster_size)
 		err_exit("Cluster size %u is too large!\n",
 				(unsigned int)vol->cluster_size);
@@ -1076,7 +1164,7 @@ static void mount_volume(unsigned long new_mntflag)
 			  volume_size(vol, vol->nr_clusters));
 }
 
-struct ntfs_walk_cluster backup_clusters = { NULL, NULL }; 
+struct ntfs_walk_cluster backup_clusters = { NULL, NULL };
 
 static int device_offset_valid(int fd, s64 ofs)
 {
@@ -1095,8 +1183,8 @@ static s64 device_size_get(int fd)
 
 		if (ioctl(fd, BLKGETSIZE64, &size) >= 0) {
 			Dprintf("BLKGETSIZE64 nr bytes = %llu (0x%llx)\n",
-					(unsigned long long)size,
-					(unsigned long long)size);
+				(unsigned long long)size,
+				(unsigned long long)size);
 			return (s64)size;
 		}
 	}
@@ -1106,7 +1194,7 @@ static s64 device_size_get(int fd)
 
 		if (ioctl(fd, BLKGETSIZE, &size) >= 0) {
 			Dprintf("BLKGETSIZE nr 512 byte blocks = %lu "
-					"(0x%lx)\n", size, size);
+				"(0x%lx)\n", size, size);
 			return (s64)size * 512;
 		}
 	}
@@ -1116,7 +1204,7 @@ static s64 device_size_get(int fd)
 
 		if (ioctl(fd, FDGETPRM, &this_floppy) >= 0) {
 			Dprintf("FDGETPRM nr 512 byte blocks = %lu (0x%lx)\n",
-					this_floppy.size, this_floppy.size);
+				this_floppy.size, this_floppy.size);
 			return (s64)this_floppy.size * 512;
 		}
 	}
@@ -1149,15 +1237,20 @@ static void fsync_clone(int fd)
 
 static void set_filesize(s64 filesize)
 {
+	long fs_type = 0; /* Unknown filesystem type */
+
 	if (fstatfs(fd_out, &opt.stfs) == -1)
 		Printf("WARNING: Couldn't get filesystem type: "
 		       "%s\n", strerror(errno));
-	else if (opt.stfs.f_type == 0x52654973)
+	else 
+		fs_type = opt.stfs.f_type;
+		
+	if (fs_type == 0x52654973)
 		Printf("WARNING: You're using ReiserFS, it has very poor "
 		       "performance creating\nlarge sparse files. The next "
 		       "operation might take a very long time!\n"
 		       "Creating sparse output file ...\n");
-	else if (opt.stfs.f_type == 0x517b)
+	else if (fs_type == 0x517b)
 		Printf("WARNING: You're using SMBFS and if the remote share "
 		       "isn't Samba but a Windows\ncomputer then the clone "
 		       "operation will be very inefficient and may fail!\n");
@@ -1165,15 +1258,24 @@ static void set_filesize(s64 filesize)
 	if (ftruncate(fd_out, filesize) == -1) {
 		int err = errno;
 		perr_printf("ftruncate failed for file '%s'", opt.output);
+		if (fs_type)
+			Printf("Destination filesystem type is 0x%lx.\n",
+			       (unsigned long)fs_type);
 		if (err == E2BIG) {
 			Printf("Your system or the destination filesystem "
 			       "doesn't support large files.\n");
-			if (opt.stfs.f_type == 0x517b) {
+			if (fs_type == 0x517b) {
 				Printf("SMBFS needs minimum Linux kernel "
 				       "version 2.4.25 and\n the 'lfs' option"
 				       "\nfor smbmount to have large "
 				       "file support.\n");
 			}
+		} else if (err == EPERM) {
+			Printf("Apparently the destination filesystem doesn't "
+			       "support sparse files.\nYou can overcome this "
+			       "by using the more efficient --save-image "
+			       "option\nof ntfsclone. Use the --restore-image "
+			       "option to restore the image.\n");
 		}
 		exit(1);
 	}
@@ -1182,7 +1284,7 @@ static void set_filesize(s64 filesize)
 static s64 open_image(void)
 {
 	if (strcmp(opt.volume, "-") == 0) {
-		if ((fd_in = fileno(stdin)) == -1) 
+		if ((fd_in = fileno(stdin)) == -1)
 			perr_exit("fileno for stdout failed");
 	} else {
 		if ((fd_in = open(opt.volume, O_RDONLY)) == -1)
@@ -1201,15 +1303,15 @@ static s64 open_image(void)
 static s64 open_volume(void)
 {
 	s64 device_size;
-	
+
 	mount_volume(MS_RDONLY);
-	
+
 	device_size = ntfs_device_size_get(vol->dev, 1);
 	if (device_size <= 0)
 		err_exit("Couldn't get device size (%lld)!\n", device_size);
-	
+
 	print_volume_size("Current device size", device_size);
-	
+
 	if (device_size < vol->nr_clusters * vol->cluster_size)
 		err_exit("Current NTFS volume size is bigger than the device "
 			 "size (%lld)!\nCorrupt partition table or incorrect "
@@ -1236,25 +1338,118 @@ static void check_output_device(s64 input_size)
 		if (dest_size < input_size)
 			err_exit("Output device is too small (%lld) to fit the "
 				 "NTFS image (%lld).\n", dest_size, input_size);
-		
+
 		check_if_mounted(opt.output, 0);
 	} else
 		set_filesize(input_size);
 }
 
+static ntfs_attr_search_ctx *attr_get_search_ctx(ntfs_inode *ni)
+{
+	ntfs_attr_search_ctx *ret;
+
+	if ((ret = ntfs_attr_get_search_ctx(ni, NULL)) == NULL)
+		perr_printf("ntfs_attr_get_search_ctx");
+
+	return ret;
+}
+
+static int str2unicode(const char *aname, ntfschar **ustr, int *len)
+{
+	if (aname && ((*len = ntfs_mbstoucs(aname, ustr, 0)) == -1)) {
+		perr_printf("Unable to convert '%s' to Unicode", aname);
+		return -1;
+	}
+
+	if (!*ustr || !*len) {
+		*ustr = AT_UNNAMED;
+		*len = 0;
+	}
+
+	return 0;
+}
+
+/**
+ * lookup_data_attr
+ *
+ * Find the $DATA attribute (with or without a name) for the given ntfs inode.
+ */
+static ntfs_attr_search_ctx *lookup_data_attr(ntfs_inode *ni, const char *aname)
+{
+	ntfs_attr_search_ctx *ctx;
+	ntfschar *ustr = NULL;
+	int len = 0;
+
+	if ((ctx = attr_get_search_ctx(ni)) == NULL)
+		return NULL;
+
+	if (str2unicode(aname, &ustr, &len) == -1)
+		goto error_out;
+
+	if (ntfs_attr_lookup(AT_DATA, ustr, len, 0, 0, NULL, 0, ctx)) {
+		perr_printf("ntfs_attr_lookup");
+		goto error_out;
+	}
+	if (ustr != AT_UNNAMED)
+		free(ustr);
+
+	return ctx;
+error_out:
+	ntfs_attr_put_search_ctx(ctx);
+	return NULL;
+}
+
+static void ignore_bad_clusters(ntfs_walk_clusters_ctx *image)
+{
+	ntfs_inode *ni;
+	ntfs_attr_search_ctx *ctx = NULL;
+	runlist *rl, *rl_bad;
+	s64 nr_bad_clusters = 0;
+
+	if (!(ni = ntfs_inode_open(vol, FILE_BadClus)))
+		perr_exit("ntfs_open_inode");
+
+	if ((ctx = lookup_data_attr(ni, "$Bad")) == NULL)
+		exit(1);
+
+	if (!(rl_bad = ntfs_mapping_pairs_decompress(vol, ctx->attr, NULL)))
+		perr_exit("ntfs_mapping_pairs_decompress");
+
+	for (rl = rl_bad; rl->length; rl++) {
+		s64 lcn = rl->lcn;
+
+		if (lcn == LCN_HOLE || lcn < 0)
+			continue;
+
+		for (; lcn < rl->lcn + rl->length; lcn++, nr_bad_clusters++) {
+			if (ntfs_bit_get_and_set(lcn_bitmap.bm, lcn, 0))
+				image->inuse--;
+		}
+	}
+	if (nr_bad_clusters)
+		Printf("WARNING: The disk has %lld or more bad sectors"
+		       " (hardware faults).\n", nr_bad_clusters);
+	free(rl_bad);
+
+	ntfs_attr_put_search_ctx(ctx);
+	if (ntfs_inode_close(ni))
+		perr_exit("ntfs_inode_close failed for $BadClus");
+}
+
 int main(int argc, char **argv)
 {
 	ntfs_walk_clusters_ctx image;
 	s64 device_size;        /* input device size in bytes */
 	s64 ntfs_size;
-	int wiped_total = 0;
+	unsigned int wiped_total = 0;
 
 	/* print to stderr, stdout can be an NTFS image ... */
-	Eprintf("%s v%s\n", EXEC_NAME, VERSION);
+	Eprintf("%s v%s (libntfs %s)\n", EXEC_NAME, VERSION,
+			ntfs_libntfs_version());
 	msg_out = stderr;
 
 	parse_options(argc, argv);
-	
+
 	utils_set_locale();
 
 	if (opt.restore_image) {
@@ -1267,19 +1462,19 @@ int main(int argc, char **argv)
 	ntfs_size += 512; /* add backup boot sector */
 
 	if (opt.std_out) {
-	       if ((fd_out = fileno(stdout)) == -1) 
+	       if ((fd_out = fileno(stdout)) == -1)
 		       perr_exit("fileno for stdout failed");
 	} else {
 	        /* device_size_get() might need to read() */
 		int flags = O_RDWR;
-		
+
 		if (!opt.blkdev_out) {
 			flags |= O_CREAT | O_TRUNC;
 			if (!opt.overwrite)
 				flags |= O_EXCL;
 		}
 
-		if ((fd_out = open(opt.output, flags, S_IRWXU)) == -1) 
+		if ((fd_out = open(opt.output, flags, S_IRWXU)) == -1)
 			perr_exit("Opening file '%s' failed", opt.output);
 
 		if (!opt.save_image)
@@ -1292,59 +1487,62 @@ int main(int argc, char **argv)
 		fsync_clone(fd_out);
 		exit(0);
 	}
-	
+
 	setup_lcn_bitmap();
 	memset(&image, 0, sizeof(image));
-	backup_clusters.image = &image; 
+	backup_clusters.image = &image;
 
 	walk_clusters(vol, &backup_clusters);
 	compare_bitmaps(&lcn_bitmap);
 	print_disk_usage(vol->cluster_size, vol->nr_clusters, image.inuse);
-	
-	free(lcn_bitmap.bm);
+
+	ignore_bad_clusters(&image);
 
 	if (opt.save_image)
 		initialise_image_hdr(device_size, image.inuse);
-	
+
 	/* FIXME: save backup boot sector */
 
 	if (opt.std_out || !opt.metadata_only) {
-		s64 nr_clusters = (opt.std_out && !opt.save_image)
-			? vol->nr_clusters : image.inuse;
-		
-		clone_ntfs(nr_clusters);
+		s64 nr_clusters_to_save = image.inuse;
+		if (opt.std_out && !opt.save_image)
+			nr_clusters_to_save = vol->nr_clusters;
+
+		clone_ntfs(nr_clusters_to_save);
 		fsync_clone(fd_out);
 		exit(0);
 	}
-	
-	wipe = 1;	
+
+	wipe = 1;
 	opt.volume = opt.output;
-	/* 'force' again mount for dirty volumes (e.g. after resize). 
+	/* 'force' again mount for dirty volumes (e.g. after resize).
 	   FIXME: use mount flags to avoid potential side-effects in future */
 	opt.force++;
 	mount_volume(0);
 
+	free(lcn_bitmap.bm);
 	setup_lcn_bitmap();
 	memset(&image, 0, sizeof(image));
-	backup_clusters.image = &image; 
+	backup_clusters.image = &image;
 
 	walk_clusters(vol, &backup_clusters);
-	
-	Printf("Num of MFT records       = %8lld\n",
-			(long long)vol->nr_mft_records); 
-	Printf("Num of used MFT records  = %8d\n", nr_used_mft_records); 
-	
-	Printf("Wiped unused MFT data    = %8d\n", wiped_unused_mft_data); 
-	Printf("Wiped deleted MFT data   = %8d\n", wiped_unused_mft); 
-	Printf("Wiped resident user data = %8d\n", wiped_resident_data);
-	Printf("Wiped timestamp data     = %8d\n", wiped_timestamp_data);
-	
+
+	Printf("Num of MFT records       = %10lld\n",
+			(long long)vol->mft_na->initialized_size >>
+			vol->mft_record_size_bits);
+	Printf("Num of used MFT records  = %10u\n", nr_used_mft_records);
+
+	Printf("Wiped unused MFT data    = %10u\n", wiped_unused_mft_data);
+	Printf("Wiped deleted MFT data   = %10u\n", wiped_unused_mft);
+	Printf("Wiped resident user data = %10u\n", wiped_resident_data);
+	Printf("Wiped timestamp data     = %10u\n", wiped_timestamp_data);
+
 	wiped_total += wiped_unused_mft_data;
 	wiped_total += wiped_unused_mft;
 	wiped_total += wiped_resident_data;
 	wiped_total += wiped_timestamp_data;
-	Printf("Wiped totally            = %8d\n", wiped_total);
-	
+	Printf("Wiped totally            = %10u\n", wiped_total);
+
 	fsync_clone(fd_out);
 	exit(0);
 }
