@@ -1,6 +1,7 @@
 /**
  * ntfscluster - Part of the Linux-NTFS project.
  *
+ * Copyright (c) 2005 Anton Altaparmakov
  * Copyright (c) 2002-2003 Richard Russon
  *
  * This utility will locate the owner of any given sector or cluster.
@@ -23,11 +24,21 @@
 
 #include "config.h"
 
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_GETOPT_H
 #include <getopt.h>
+#endif
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
+#ifdef HAVE_LIMITS_H
 #include <limits.h>
+#endif
 
 #include "ntfscluster.h"
 #include "types.h"
@@ -37,6 +48,7 @@
 #include "debug.h"
 #include "dir.h"
 #include "cluster.h"
+#include "version.h"
 
 static const char *EXEC_NAME = "ntfscluster";
 static struct options opts;
@@ -54,8 +66,9 @@ GEN_PRINTF (Qprintf, stdout, &opts.quiet,   FALSE)
  */
 static void version (void)
 {
-	printf ("\n%s v%s - Find the owner of any given sector or cluster.\n\n",
-		EXEC_NAME, VERSION);
+	printf("\n%s v%s (libntfs %s) - Find the owner of any given sector or "
+			"cluster.\n\n", EXEC_NAME, VERSION,
+			ntfs_libntfs_version());
 	printf ("Copyright (c) 2002-2003 Richard Russon\n");
 	printf ("\n%s\n%s%s\n", ntfs_gpl, ntfs_bugs, ntfs_home);
 }
@@ -269,7 +282,7 @@ static int info (ntfs_volume *vol)
 
 		inuse++;
 
-		a_ctx = ntfs_attr_get_search_ctx (NULL, m_ctx->inode->mrec);
+		a_ctx = ntfs_attr_get_search_ctx (m_ctx->inode, NULL);
 
 		while ((rec = find_attribute (AT_UNUSED, a_ctx))) {
 
@@ -311,7 +324,7 @@ static int info (ntfs_volume *vol)
 	d = vol->nr_clusters << cb;
 	e = vol->nr_clusters;
 	f = vol->nr_clusters >> cps;
-	g = vol->nr_mft_records;
+	g = vol->mft_na->initialized_size >> vol->mft_record_size_bits;
 	h = inuse;
 	i = h * 100 / g;
 	j = fc;
@@ -357,7 +370,7 @@ static int info (ntfs_volume *vol)
  */
 static int dump_file (ntfs_volume *vol, ntfs_inode *ino)
 {
-	u8 buffer[1024];
+	char buffer[1024];
 	ntfs_attr_search_ctx *ctx;
 	ATTR_RECORD *rec;
 	int i;
@@ -487,7 +500,7 @@ int main (int argc, char *argv[])
 			result = cluster_find (vol, opts.range_begin, opts.range_end, (cluster_cb*)&print_match, NULL);
 			break;
 		case act_file:
-			ino = utils_pathname_to_inode (vol, NULL, opts.filename);
+			ino = ntfs_pathname_to_inode(vol, NULL, opts.filename);
 			if (ino)
 				result = dump_file (vol, ino);
 			break;
