@@ -1,7 +1,7 @@
 /**
  * ntfsdump_logfile - Part of the Linux-NTFS project.
  *
- * Copyright (c) 2000-2004 Anton Altaparmakov
+ * Copyright (c) 2000-2005 Anton Altaparmakov
  *
  * This utility will interpret the contents of the journal ($LogFile) of an
  * NTFS partition and display the results on stdout.  Errors will be output to
@@ -25,22 +25,40 @@
 /* TODO:
  *	- Remove the need for clipping at 64MiB.
  *	- Add normal command line switchs (use getopt_long()).
- *	- For a volume: allow dumping only uncommited records.
+ *	- For a volume: allow dumping only uncommitted records.
  *	- For a file:   get an optional command line parameter for the last SN.
  *	- Sanity checks.
  */
 
 #include "config.h"
 
+#ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
+#endif
+#ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
+#endif
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
+#endif
+#ifdef HAVE_STDARG_H
 #include <stdarg.h>
+#endif
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
+#ifdef HAVE_ERRNO_H
 #include <errno.h>
+#endif
+#ifdef HAVE_FCNTL_H
 #include <fcntl.h>
+#endif
 
 #include "types.h"
 #include "endians.h"
@@ -51,6 +69,7 @@
 #include "logfile.h"
 #include "mst.h"
 #include "utils.h"
+#include "version.h"
 
 typedef struct {
 	BOOL is_volume;
@@ -153,19 +172,20 @@ void log_err_exit(u8 *buf, const char *fmt, ...)
 void usage(const char *exec_name) __attribute__ ((noreturn));
 void usage(const char *exec_name)
 {
-	Eprintf("%s v%s - Interpret and display information about the "
-		"journal\n($LogFile) of an NTFS volume.\n"
-		"Copyright (c) 2000-2004 Anton Altaparmakov.\n"
-		"%s is free software, released under the GNU General "
-		"Public License\nand you are welcome to redistribute "
-		"it under certain conditions.\n%s comes with "
-		"ABSOLUTELY NO WARRANTY; for details read the GNU\n"
-		"General Public License to be found in the file "
-		"COPYING in the main Linux-NTFS\ndistribution "
-		"directory.\nUsage: %s device\n    e.g. %s /dev/hda6\n"
-		"Alternative usage: %s -f file\n    e.g. %s -f "
-		"MyCopyOfTheLogFile\n", exec_name, VERSION, exec_name,
-		exec_name, exec_name, exec_name, exec_name, exec_name);
+	Eprintf("%s v%s (libntfs %s) - Interpret and display information "
+			"about the journal\n($LogFile) of an NTFS volume.\n"
+			"Copyright (c) 2000-2005 Anton Altaparmakov.\n"
+			"%s is free software, released under the GNU General "
+			"Public License\nand you are welcome to redistribute "
+			"it under certain conditions.\n%s comes with "
+			"ABSOLUTELY NO WARRANTY; for details read the GNU\n"
+			"General Public License to be found in the file "
+			"COPYING in the main Linux-NTFS\ndistribution "
+			"directory.\nUsage: %s device\n    e.g. %s /dev/hda6\n"
+			"Alternative usage: %s -f file\n    e.g. %s -f "
+			"MyCopyOfTheLogFile\n", exec_name, VERSION,
+			ntfs_libntfs_version(), exec_name, exec_name,
+			exec_name, exec_name, exec_name, exec_name);
 	exit(1);
 }
 
@@ -295,7 +315,7 @@ static void restart_header_sanity(RESTART_PAGE_HEADER *rstr, u8 *buf)
 	/* Verify the location and size of the update sequence array. */
 	usa_end_ofs = le16_to_cpu(rstr->usa_ofs) +
 			le16_to_cpu(rstr->usa_count) * sizeof(u16);
-	if (page_size / NTFS_SECTOR_SIZE + 1 != le16_to_cpu(rstr->usa_count))
+	if (page_size / NTFS_BLOCK_SIZE + 1 != le16_to_cpu(rstr->usa_count))
 		log_err_exit(buf, "Restart page header in $LogFile is "
 				"corrupt:  Update sequence array size is "
 				"wrong.  Cannot handle this yet.\n");
@@ -304,18 +324,18 @@ static void restart_header_sanity(RESTART_PAGE_HEADER *rstr, u8 *buf)
 				"corrupt:  Update sequence array overlaps "
 				"restart page header.  Cannot handle this "
 				"yet.\n");
-	if (usa_end_ofs > NTFS_SECTOR_SIZE - sizeof(u16))
+	if (usa_end_ofs > NTFS_BLOCK_SIZE - sizeof(u16))
 		log_err_exit(buf, "Restart page header in $LogFile is "
 				"corrupt:  Update sequence array overlaps or "
 				"is behind first protected sequence number.  "
 				"Cannot handle this yet.\n");
-	if (usa_end_ofs > le16_to_cpu(rstr->restart_offset))
+	if (usa_end_ofs > le16_to_cpu(rstr->restart_area_offset))
 		log_err_exit(buf, "Restart page header in $LogFile is "
 				"corrupt:  Update sequence array overlaps or "
 				"is behind restart area.  Cannot handle this "
 				"yet.\n");
 	/* Finally, verify the offset of the restart area. */
-	if (le16_to_cpu(rstr->restart_offset) & 7)
+	if (le16_to_cpu(rstr->restart_area_offset) & 7)
 		log_err_exit(buf, "Restart page header in $LogFile is "
 				"corrupt:  Restart area offset is not aligned "
 				"to 8-byte boundary.  Cannot handle this "
@@ -341,8 +361,8 @@ static void dump_restart_areas_header(RESTART_PAGE_HEADER *rstr)
 			(unsigned int)le32_to_cpu(rstr->log_page_size),
 			(unsigned int)le32_to_cpu(rstr->log_page_size));
 	printf("restart_offset = %u (0x%x)\n",
-			le16_to_cpu(rstr->restart_offset),
-			le16_to_cpu(rstr->restart_offset));
+			le16_to_cpu(rstr->restart_area_offset),
+			le16_to_cpu(rstr->restart_area_offset));
 }
 
 static void dump_restart_areas_area(RESTART_PAGE_HEADER *rstr)
@@ -351,7 +371,8 @@ static void dump_restart_areas_area(RESTART_PAGE_HEADER *rstr)
 	RESTART_AREA *ra;
 	int client;
 
-	ra = (RESTART_AREA*)((u8*)rstr + le16_to_cpu(rstr->restart_offset));
+	ra = (RESTART_AREA*)((u8*)rstr +
+			le16_to_cpu(rstr->restart_area_offset));
 	printf("current_lsn = %lli (0x%llx)\n",
 			(long long)sle64_to_cpu(ra->current_lsn),
 			(unsigned long long)sle64_to_cpu(ra->current_lsn));
@@ -379,13 +400,15 @@ static void dump_restart_areas_area(RESTART_PAGE_HEADER *rstr)
 	printf("last_lsn_data_length = %u (0x%x)\n",
 			(unsigned int)le32_to_cpu(ra->last_lsn_data_length),
 			(unsigned int)le32_to_cpu(ra->last_lsn_data_length));
-	printf("record_length = %u (0x%x)\n", le16_to_cpu(ra->record_length),
-			le16_to_cpu(ra->record_length));
+	printf("log_record_header_length = %u (0x%x)\n",
+			le16_to_cpu(ra->log_record_header_length),
+			le16_to_cpu(ra->log_record_header_length));
 	printf("log_page_data_offset = %u (0x%x)\n",
 			le16_to_cpu(ra->log_page_data_offset),
 			le16_to_cpu(ra->log_page_data_offset));
-	printf("unknown = %u (0x%x)\n", le16_to_cpu(ra->unknown),
-			le16_to_cpu(ra->unknown));
+	printf("restart_log_open_count = %u (0x%x)\n",
+			(unsigned)le32_to_cpu(ra->restart_log_open_count),
+			(unsigned)le32_to_cpu(ra->restart_log_open_count));
 	lcr = (LOG_CLIENT_RECORD*)((u8*)ra +
 			le16_to_cpu(ra->client_array_offset));
 	for (client = 0; client < le16_to_cpu(ra->log_clients); client++) {
@@ -467,11 +490,11 @@ rstr_pass_loc:
 
 		/* Exclude the usa from the comparison. */
 		ra = (RESTART_AREA*)((u8*)rstr1 +
-				le16_to_cpu(rstr1->restart_offset));
+				le16_to_cpu(rstr1->restart_area_offset));
 		if (!memcmp(rstr1, rstr, le16_to_cpu(rstr1->usa_ofs)) &&
 				!memcmp((u8*)rstr1 + le16_to_cpu(
-				rstr1->restart_offset), (u8*)rstr +
-				le16_to_cpu(rstr->restart_offset),
+				rstr1->restart_area_offset), (u8*)rstr +
+				le16_to_cpu(rstr->restart_area_offset),
 				le16_to_cpu(ra->restart_area_length))) {
 			puts("\nSkipping analysis of second restart page "
 					"because it fully matches the first "
@@ -502,7 +525,7 @@ skip_rstr_pass:
 		++pass;
 		goto rstr_pass_loc;
 	}
-	
+
 	return rstr;
 }
 
@@ -711,7 +734,7 @@ int main(int argc, char **argv)
 		puts("$LogFile is not initialized.");
 		return 0;
 	}
-	
+
 	/*
 	 * First, verify the restart page header for consistency.
 	 */
@@ -736,7 +759,7 @@ int main(int argc, char **argv)
 	 */
 	rcrd = (RECORD_PAGE_HEADER*)dump_restart_areas(rstr, buf, page_size);
 	printf("\n\nFinished with restart pages.  "
- 		"Beginning with log pages.\n");
+		"Beginning with log pages.\n");
 
 	/*
 	 * Dump the log areas.
