@@ -2,9 +2,10 @@
  * ntfsinfo - Part of the Linux-NTFS project.
  *
  * Copyright (c) 2002-2004 Matthew J. Fanto
- * Copyright (c) 2002-2004 Anton Altaparmakov
- * Copyright (c) 2002-2003 Richard Russon
- * Copyright (c) 2004 Yura Pakhuchiy
+ * Copyright (c) 2002-2005 Anton Altaparmakov
+ * Copyright (c) 2002-2005 Richard Russon
+ * Copyright (c) 2004-2005 Yura Pakhuchiy
+ * Copyright (c)      2005 Cristian Klein
  *
  * This utility will dump a file's attributes.
  *
@@ -39,19 +40,27 @@
  *
  *	Still need to do:
  *	    $REPARSE_POINT/$SYMBOLIC_LINK
- *	    $EA_INFORMATION
- *	    $EA
  *	    $PROPERTY_SET
  *	    $LOGGED_UTILITY_STREAM
  */
 
 #include "config.h"
 
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
+#ifdef HAVE_TIME_H
 #include <time.h>
+#endif
+#ifdef HAVE_GETOPT_H
 #include <getopt.h>
+#endif
 
 #ifdef HAVE_ERRNO_H
 #include <errno.h>
@@ -64,6 +73,10 @@
 #include "inode.h"
 #include "utils.h"
 #include "security.h"
+#include "mst.h"
+#include "dir.h"
+#include "ntfstime.h"
+#include "version.h"
 
 static const char *EXEC_NAME = "ntfsinfo";
 
@@ -92,13 +105,15 @@ GEN_PRINTF (Qprintf, stdout, &opts.quiet,   FALSE)
  */
 static void version (void)
 {
-	printf ("\n%s v%s - Display information about an NTFS Volume.\n\n",
-		EXEC_NAME, VERSION);
+	printf ("\n%s v%s (libntfs %s) - Display information about an NTFS "
+			"Volume.\n\n", EXEC_NAME, VERSION,
+			ntfs_libntfs_version());
 	printf ("Copyright (c)\n");
 	printf ("    2002-2004 Matthew J. Fanto\n");
-	printf ("    2002-2004 Anton Altaparmakov\n");
+	printf ("    2002-2005 Anton Altaparmakov\n");
 	printf ("    2002-2003 Richard Russon\n");
 	printf ("    2003      Leonard Norrgård\n");
+	printf ("    2004-2005 Yura Pakhuchiy\n");
 	printf ("\n%s\n%s%s\n", ntfs_gpl, ntfs_bugs, ntfs_home);
 }
 
@@ -200,7 +215,7 @@ static int parse_options (int argc, char *argv[])
 			opts.notime++;
 			break;
 		case 'T':
-			/* 'T' is depreceted, notify */
+			/* 'T' is deprecated, notify */
 			Eprintf ("Option 'T' is deprecated, it was replaced by 't'.\n");
 			err++;
 			break;
@@ -277,13 +292,13 @@ static int parse_options (int argc, char *argv[])
  */
 static char *ntfsinfo_time_to_str(const s64 sle_ntfs_clock)
 {
-	time_t unix_clock = ntfs2utc(sle64_to_cpu(sle_ntfs_clock));
+	time_t unix_clock = ntfs2utc(sle_ntfs_clock);
 	return ctime(&unix_clock);
 }
 
 /**
  * ntfs_attr_get_name()
- * @attr:	a vaild attribute record
+ * @attr:	a valid attribute record
  *
  * return multi-byte string containing the attribute name if exist. the user
  *             is then responsible of freeing that memory.
@@ -296,7 +311,7 @@ static char *ntfs_attr_get_name(ATTR_RECORD *attr)
 	ntfschar *ucs_attr_name;
 	char *mbs_attr_name = NULL;
 	int mbs_attr_name_size;
-	
+
 	/* calculate name position */
 	ucs_attr_name = (ntfschar *)((char *)attr + le16_to_cpu(attr->name_offset));
 	/* convert unicode to printable format */
@@ -326,7 +341,7 @@ static void ntfs_dump_volume(ntfs_volume *vol)
 	printf("\tCluster Size: %u\n", (unsigned int)vol->cluster_size);
 	printf("\tVolume Size in Clusters: %lld\n",
 			(long long)vol->nr_clusters);
-	
+
 	printf("MFT Information \n");
 	printf("\tMFT Record Size: %u\n", (unsigned int)vol->mft_record_size);
 	printf("\tMFT Zone Multiplier: %u\n", vol->mft_zone_multiplier);
@@ -338,8 +353,6 @@ static void ntfs_dump_volume(ntfs_volume *vol)
 			(long long)vol->data1_zone_pos);
 	printf("\tCurrent Position in Second Data Zone: %lld\n",
 			(long long)vol->data2_zone_pos);
-	printf("\tNumber of Initialized Records in MFT: %lld\n",
-			(long long)vol->nr_mft_records);
 	printf("\tLCN of Data Attribute for FILE_MFT: %lld\n",
 			(long long)vol->mft_lcn);
 	printf("\tFILE_MFTMirr Size: %d\n", vol->mftmirr_size);
@@ -347,7 +360,7 @@ static void ntfs_dump_volume(ntfs_volume *vol)
 			(long long)vol->mftmirr_lcn);
 	printf("\tSize of Attribute Definition Table: %d\n",
 			(int)vol->attrdef_len);
-	
+
 	printf("FILE_Bitmap Information \n");
 	printf("\tFILE_Bitmap MFT Record Number: %llu\n",
 			(unsigned long long)vol->lcnbmp_ni->mft_no);
@@ -359,7 +372,7 @@ static void ntfs_dump_volume(ntfs_volume *vol)
 			(int)vol->lcnbmp_ni->nr_extents);
 	/* FIXME: need to add code for the union if nr_extens != 0, but
 	   i dont know if it will ever != 0 with FILE_Bitmap */
-	
+
 	printf("FILE_Bitmap Data Attribute Information\n");
 	printf("\tDecompressed Runlist: not done yet\n");
 	printf("\tBase Inode: %llu\n",
@@ -383,10 +396,108 @@ static void ntfs_dump_volume(ntfs_volume *vol)
 			vol->lcnbmp_na->compression_block_size_bits);
 	printf("\tCompression Block Clusters: %u\n",
 			vol->lcnbmp_na->compression_block_clusters);
-		
+
 	//TODO: Still need to add a few more attributes
 }
 
+/**
+ * ntfs_dump_flags - Dump flags for STANDARD_INFORMATION and FILE_NAME.
+ * @type:	dump flags for this attribute type
+ * @flags:	flags for dumping
+ */
+static void ntfs_dump_flags(ATTR_TYPES type, u32 flags)
+{
+	printf("\tFile attributes:\t");
+	if (flags & FILE_ATTR_READONLY) {
+		printf(" READONLY");
+		flags &= ~FILE_ATTR_READONLY;
+	}
+	if (flags & FILE_ATTR_HIDDEN) {
+		printf(" HIDDEN");
+		flags &= ~FILE_ATTR_HIDDEN;
+	}
+	if (flags & FILE_ATTR_SYSTEM) {
+		printf(" SYSTEM");
+		flags &= ~FILE_ATTR_SYSTEM;
+	}
+	if (flags & FILE_ATTR_ARCHIVE) {
+		printf(" ARCHIVE");
+		flags &= ~FILE_ATTR_ARCHIVE;
+	}
+	if (flags & FILE_ATTR_DEVICE) {
+		printf(" DEVICE");
+		flags &= ~FILE_ATTR_DEVICE;
+	}
+	if (flags & FILE_ATTR_NORMAL) {
+		printf(" NORMAL");
+		flags &= ~FILE_ATTR_NORMAL;
+	}
+	if (flags & FILE_ATTR_TEMPORARY) {
+		printf(" TEMPORARY");
+		flags &= ~FILE_ATTR_TEMPORARY;
+	}
+	if (flags & FILE_ATTR_SPARSE_FILE) {
+		printf(" SPARSE_FILE");
+		flags &= ~FILE_ATTR_SPARSE_FILE;
+	}
+	if (flags & FILE_ATTR_REPARSE_POINT) {
+		printf(" REPARSE_POINT");
+		flags &= ~FILE_ATTR_REPARSE_POINT;
+	}
+	if (flags & FILE_ATTR_COMPRESSED) {
+		printf(" COMPRESSED");
+		flags &= ~FILE_ATTR_COMPRESSED;
+	}
+	if (flags & FILE_ATTR_OFFLINE) {
+		printf(" OFFLINE");
+		flags &= ~FILE_ATTR_OFFLINE;
+	}
+	if (flags & FILE_ATTR_NOT_CONTENT_INDEXED) {
+		printf(" NOT_CONTENT_INDEXED");
+		flags &= ~FILE_ATTR_NOT_CONTENT_INDEXED;
+	}
+	if (flags & FILE_ATTR_ENCRYPTED) {
+		printf(" ENCRYPTED");
+		flags &= ~FILE_ATTR_ENCRYPTED;
+	}
+	if (type == AT_FILE_NAME) {
+		if (flags & FILE_ATTR_DUP_FILE_NAME_INDEX_PRESENT) {
+			printf(" FILE_NAME_INDEX");
+			flags &= ~FILE_ATTR_DUP_FILE_NAME_INDEX_PRESENT;
+		}
+		if (flags & FILE_ATTR_DUP_VIEW_INDEX_PRESENT) {
+			printf(" VIEW_INDEX");
+			flags &= ~FILE_ATTR_DUP_VIEW_INDEX_PRESENT;
+		}
+	}
+	if (flags)
+		printf(" UNKNOWN: 0x%x", (unsigned int)le32_to_cpu(flags));
+	printf("\n");
+}
+
+static void ntfs_dump_namespace(u8 file_name_type)
+{
+	const char *mbs_file_type;
+
+	/* name space */
+	switch (file_name_type) {
+	case FILE_NAME_POSIX:
+		mbs_file_type = "POSIX";
+		break;
+	case FILE_NAME_WIN32:
+		mbs_file_type = "Win32";
+		break;
+	case FILE_NAME_DOS:
+		mbs_file_type = "DOS";
+		break;
+	case FILE_NAME_WIN32_AND_DOS:
+		mbs_file_type = "Win32 & DOS";
+		break;
+	default:
+		mbs_file_type = "(unknown)";
+	}
+	printf("\tNamespace:\t\t %s\n", mbs_file_type);
+}
 
 /* *************** functions for dumping attributes ******************** */
 /**
@@ -402,29 +513,30 @@ static void ntfs_dump_attr_standard_information(ATTR_RECORD *attr)
 
 	printf("Dumping attribute $STANDARD_INFORMATION (0x10)\n");
 
+	printf("\tAttribute instance:\t %u\n", le16_to_cpu(attr->instance));
+
 	/* let's start with mandatory? fields */
 
 	/* time conversion stuff */
 	if (!opts.notime) {
 		char *ntfs_time_str = NULL;
-		
+
 		ntfs_time_str = ntfsinfo_time_to_str(standard_attr->creation_time);
 		printf("\tFile Creation Time:\t %s",ntfs_time_str);
-		
+
 		ntfs_time_str = ntfsinfo_time_to_str(
 			standard_attr->last_data_change_time);
 		printf("\tFile Altered Time:\t %s",ntfs_time_str);
-		
+
 		ntfs_time_str = ntfsinfo_time_to_str(
 			standard_attr->last_mft_change_time);
 		printf("\tMFT Changed Time:\t %s",ntfs_time_str);
-		
+
 		ntfs_time_str = ntfsinfo_time_to_str(standard_attr->last_access_time);
 		printf("\tLast Accessed Time:\t %s",ntfs_time_str);
 	}
-	
-	/* TODO: file_attributes - Flags describing the file. */
-	
+	ntfs_dump_flags(attr->type, standard_attr->file_attributes);
+
 	printf("\tMax Number of Versions:\t %u \n",
 		(unsigned int)le32_to_cpu(standard_attr->maximum_versions));
 	printf("\tVersion Number:\t\t %u \n",
@@ -432,7 +544,6 @@ static void ntfs_dump_attr_standard_information(ATTR_RECORD *attr)
 	printf("\tClass ID:\t\t %u \n",
 		(unsigned int)le32_to_cpu(standard_attr->class_id));
 
-	
 	value_length = le32_to_cpu(attr->value_length);
 	if (value_length == 48) {
 /*		printf("\t$STANDARD_INFORMATION fields owner_id, security_id, quota \n"
@@ -447,14 +558,17 @@ static void ntfs_dump_attr_standard_information(ATTR_RECORD *attr)
 			"either 72 or 48, something is wrong...\n",
 			(unsigned int)value_length);
 	}
-	
 }
 
 /**
  * ntfs_dump_attr_list()
  */
-static void ntfs_dump_attr_list(ATTR_RECORD *attr)
+static void ntfs_dump_attr_list(ATTR_RECORD *attr, ntfs_volume *vol)
 {
+	ATTR_LIST_ENTRY *entry;
+	u8 *value;
+	s64 l;
+
 	printf("Dumping attribute AT_ATTRIBUTE_LIST (0x20)\n");
 
 	/* Dump list's name */
@@ -466,7 +580,7 @@ static void ntfs_dump_attr_list(ATTR_RECORD *attr)
 			printf("\tList name:\t\t '%s'\n",stream_name);
 			free(stream_name);
 		} else {
-			/* an error occured, errno holds the reason - notify the user */
+			/* an error occurred, errno holds the reason - notify the user */
 			fprintf(stderr, "ntfsinfo error: could not parse stream name: %s\n",
 				strerror(errno));
 		}
@@ -474,6 +588,8 @@ static void ntfs_dump_attr_list(ATTR_RECORD *attr)
 		printf("\tList name:\t\t unnamed\n");
 	}
 
+	printf("\tAttribute instance:\t %u\n", le16_to_cpu(attr->instance));
+
 	/* Dump list's size */
 	if (attr->non_resident) {
 		printf("\tAllocated size:\t\t %llu\n",
@@ -487,6 +603,60 @@ static void ntfs_dump_attr_list(ATTR_RECORD *attr)
 		printf("\tList's size:\t\t %u bytes\n",
 			(unsigned int)le32_to_cpu(attr->value_length));
 	}
+
+	if (!opts.verbose)
+		return;
+
+	l = ntfs_get_attribute_value_length(attr);
+	if (!l) {
+		perror("ntfs_get_attribute_value_length failed");
+		return;
+	}
+	value = malloc(l);
+	if (!value) {
+		perror("malloc failed");
+		return;
+	}
+	l = ntfs_get_attribute_value(vol, attr, value);
+	if (!l) {
+		perror("ntfs_get_attribute_value failed");
+		free(value);
+		return;
+	}
+	printf("\tDumping attribute list:");
+	entry = (ATTR_LIST_ENTRY *) value;
+	for(;(u8 *)entry < (u8 *) value + l; entry = (ATTR_LIST_ENTRY *)
+				((u8 *) entry + le16_to_cpu(entry->length))) {
+		printf("\n");
+		printf("\t\tAttribute type:\t0x%x\n",
+				(unsigned int)le32_to_cpu(entry->type));
+		printf("\t\tRecord length:\t%u\n",
+				le16_to_cpu(entry->length));
+		printf("\t\tName length:\t%u\n", entry->name_length);
+		printf("\t\tName offset:\t%u\n", entry->name_offset);
+		printf("\t\tStarting VCN:\t%lld\n",
+				sle64_to_cpu(entry->lowest_vcn));
+		printf("\t\tMFT reference:\t%lld\n",
+				MREF_LE(entry->mft_reference));
+		printf("\t\tInstance:\t%u\n", le16_to_cpu(entry->instance));
+		printf("\t\tName:\t\t");
+		if (entry->name_length) {
+			char *name = NULL;
+			int name_size;
+
+			name_size = ntfs_ucstombs(entry->name,
+					entry->name_length, &name, 0);
+
+			if (name_size > 0) {
+				printf("%s\n", name);
+				free(name);
+			} else
+				perror("ntfs_ucstombs failed");
+		} else
+			printf("unnamed\n");
+	}
+	free(value);
+	printf("\tEnd of attribute list reached.\n");
 }
 
 /**
@@ -500,7 +670,7 @@ static void ntfs_dump_attr_file_name(ATTR_RECORD *attr)
 		le16_to_cpu(attr->value_offset));
 
 	printf("Dumping attribute $FILE_NAME (0x30)\n");
-	
+
 	/* let's start with the obvious - file name */
 
 	if (file_name_attr->file_name_length>0) {
@@ -516,7 +686,7 @@ static void ntfs_dump_attr_file_name(ATTR_RECORD *attr)
 			printf("\tFile Name:\t\t '%s'\n", mbs_file_name);
 			free(mbs_file_name);
 		} else {
-			/* an error occured, errno holds the reason - notify the user */
+			/* an error occurred, errno holds the reason - notify the user */
 			fprintf(stderr, "ntfsinfo error: could not parse file name: %s\n",
 				strerror(errno));
 		}
@@ -525,17 +695,24 @@ static void ntfs_dump_attr_file_name(ATTR_RECORD *attr)
 	} else {
 		printf("\tFile Name:\t\t unnamed?!?\n");
 	}
-	
+
+	ntfs_dump_namespace(file_name_attr->file_name_type);
+
+	printf("\tAttribute instance:\t %u\n", le16_to_cpu(attr->instance));
+
 	/* other basic stuff about the file */
 	printf("\tAllocated File Size:\t %lld\n",
 		(long long)sle64_to_cpu(file_name_attr->allocated_size));
 	printf("\tReal File Size:\t\t %lld\n",
 		(long long)sle64_to_cpu(file_name_attr->data_size));
+	printf("\tParent directory:\t %lld\n",
+		(long long)MREF_LE(file_name_attr->parent_directory));
+	ntfs_dump_flags(attr->type, file_name_attr->file_attributes);
 
 	/* time stuff stuff */
 	if (!opts.notime) {
 		char *ntfs_time_str;
-	
+
 		ntfs_time_str = ntfsinfo_time_to_str(file_name_attr->creation_time);
 		printf("\tFile Creation Time:\t %s",ntfs_time_str);
 
@@ -566,6 +743,8 @@ static void ntfs_dump_attr_object_id(ATTR_RECORD *attr,ntfs_volume *vol)
 
 	printf("Dumping attribute $OBJECT_ID (0x40)\n");
 
+	printf("\tAttribute instance:\t %u\n", le16_to_cpu(attr->instance));
+
 	if (vol->major_ver >= 3.0) {
 		u32 value_length;
 		char printable_GUID[37];
@@ -618,7 +797,7 @@ static void ntfs_dump_acl(const char *prefix,ACL *acl)
 	unsigned int i;
 	u16 ace_count;
 	ACCESS_ALLOWED_ACE *ace;
-	
+
 	printf("%sRevision\t %u\n",prefix,acl->revision);
 
 	/* don't recalc le16_to_cpu every iteration (minor speedup on big-endians */
@@ -631,7 +810,7 @@ static void ntfs_dump_acl(const char *prefix,ACL *acl)
 	for (i=0;i<acl->ace_count;i++) {
 		const char *ace_type;
 		char *sid;
-		
+
 		/* set ace_type. */
 		switch (ace->type) {
 		case ACCESS_ALLOWED_ACE_TYPE:
@@ -647,7 +826,7 @@ static void ntfs_dump_acl(const char *prefix,ACL *acl)
 			ace_type = "unknown";
 			break;
 		}
-		
+
 		printf("%sACE:\t\t type:%s  flags:0x%x  access:0x%x\n",prefix,ace_type,
 			(unsigned int)le16_to_cpu(ace->flags),(unsigned int)le32_to_cpu(ace->mask));
 		/* get a SID string */
@@ -655,7 +834,7 @@ static void ntfs_dump_acl(const char *prefix,ACL *acl)
 		printf("%s\t\t SID: %s\n",prefix,sid);
 		if (sid)
 			free(sid);
-			
+
 		/* proceed to next ACE */
 		ace = (ACCESS_ALLOWED_ACE *)(((char *)ace) + le32_to_cpu(ace->size));
 	}
@@ -670,25 +849,34 @@ static void ntfs_dump_attr_security_descriptor(ATTR_RECORD *attr, ntfs_volume *v
 {
 	SECURITY_DESCRIPTOR_ATTR *sec_desc_attr;
 	char *sid;
-	
+
 	printf("Dumping attribute $SECURITY_DESCRIPTOR (0x50)\n");
 
+	printf("\tAttribute instance:\t %u\n", le16_to_cpu(attr->instance));
+
 	if (attr->non_resident) {
+		/* FIXME: We don't handle fragmented mapping pairs case. */
 		runlist *rl = ntfs_mapping_pairs_decompress(vol, attr, 0);
 		if (rl) {
 			s64 data_size, bytes_read;
 
 			data_size = sle64_to_cpu(attr->data_size);
 			sec_desc_attr = malloc(data_size);
+			if (!sec_desc_attr) {
+				perror("malloc failed");
+				free(rl);
+				return;
+			}
 			bytes_read = ntfs_rl_pread(vol, rl, 0,
 						data_size, sec_desc_attr);
 			if (bytes_read != data_size) {
 				Eprintf("ntfsinfo error: could not "
-						"read secutiry descriptor\n");
+						"read security descriptor\n");
+				free(rl);
 				free(sec_desc_attr);
 				return;
 			}
-			free (rl);
+			free(rl);
 		} else {
 			Eprintf("ntfsinfo error: could not "
 						"decompress runlist\n");
@@ -725,7 +913,7 @@ static void ntfs_dump_attr_security_descriptor(ATTR_RECORD *attr, ntfs_volume *v
 	} else {
 		printf("missing\n");
 	}
-	
+
 	printf("\tDiscretionary ACL:\t\t ");
 	if (sec_desc_attr->control & SE_DACL_PRESENT) {
 		if (sec_desc_attr->control & SE_SACL_DEFAULTED) {
@@ -737,7 +925,7 @@ static void ntfs_dump_attr_security_descriptor(ATTR_RECORD *attr, ntfs_volume *v
 	} else {
 		printf("missing\n");
 	}
-	
+
 	if (attr->non_resident) free(sec_desc_attr);
 }
 
@@ -752,6 +940,8 @@ static void ntfs_dump_attr_volume_name(ATTR_RECORD *attr)
 
 	printf("Dumping attribute $VOLUME_NAME (0x60)\n");
 
+	printf("\tAttribute instance:\t %u\n", le16_to_cpu(attr->instance));
+
 	if (attr->value_length>0) {
 		char *mbs_vol_name = NULL;
 		int mbs_vol_name_size;
@@ -768,7 +958,7 @@ static void ntfs_dump_attr_volume_name(ATTR_RECORD *attr)
 			printf("\tVolume Name:\t\t '%s'\n",mbs_vol_name);
 			free(mbs_vol_name);
 		} else {
-			/* an error occured, errno holds the reason - notify the user */
+			/* an error occurred, errno holds the reason - notify the user */
 			fprintf(stderr,"ntfsinfo error: could not parse volume name: %s\n",
 				strerror(errno));
 		}
@@ -786,13 +976,15 @@ static void ntfs_dump_attr_volume_name(ATTR_RECORD *attr)
 static void ntfs_dump_attr_volume_information(ATTR_RECORD *attr)
 {
 	VOLUME_INFORMATION *vol_information = NULL;
-	
+
 	vol_information = (VOLUME_INFORMATION*)((char *)attr+
 		le16_to_cpu(attr->value_offset));
 
 	printf("Dumping attribute $VOLUME_INFORMATION (0x70)\n");
 
-	printf("\tVolume Version:\t %d.%d\n", vol_information->major_ver,
+	printf("\tAttribute instance:\t %u\n", le16_to_cpu(attr->instance));
+
+	printf("\tVolume Version:\t\t %d.%d\n", vol_information->major_ver,
 		vol_information->minor_ver);
 	printf("\tFlags:\t\t\t ");
 	if (vol_information->flags & VOLUME_IS_DIRTY)
@@ -837,7 +1029,7 @@ static void ntfs_dump_attr_data(ATTR_RECORD *attr, ntfs_volume *vol)
 			printf("\tStream name:\t\t '%s'\n",stream_name);
 			free(stream_name);
 		} else {
-			/* an error occured, errno holds the reason - notify the user */
+			/* an error occurred, errno holds the reason - notify the user */
 			fprintf(stderr, "ntfsinfo error: could not parse stream name: %s\n",
 				strerror(errno));
 		}
@@ -845,9 +1037,11 @@ static void ntfs_dump_attr_data(ATTR_RECORD *attr, ntfs_volume *vol)
 		printf("\tStream name:\t\t unnamed\n");
 	}
 
+	printf("\tAttribute instance:\t %u\n", le16_to_cpu(attr->instance));
+
 	/* TODO: parse the flags */
 	printf("\tFlags:\t\t\t 0x%04hx\n",le16_to_cpu(attr->flags));
-	
+
 	/* fork by residence */
 	if (attr->non_resident) {
 /*		VCN lowest_vcn;	  Lowest valid virtual cluster number
@@ -894,6 +1088,83 @@ static void ntfs_dump_attr_data(ATTR_RECORD *attr, ntfs_volume *vol)
 	}
 }
 
+/**
+ * ntfs_dump_index_entries()
+ *
+ * dump sequence of index_entries and return number of entries dumped.
+ */
+static int ntfs_dump_index_entries(INDEX_ENTRY *entry, ATTR_TYPES type)
+{
+	int numb_entries = 1;
+	char *name = NULL;
+
+	while(1) {
+		if (!opts.verbose) {
+			if (entry->flags & INDEX_ENTRY_END)
+				break;
+			entry = (INDEX_ENTRY *)((u8 *)entry +
+						le16_to_cpu(entry->length));
+			numb_entries++;
+			continue;
+		}
+		Vprintf("\n");
+		Vprintf("\t\tEntry length:\t\t %u\n",
+				le16_to_cpu(entry->length));
+		Vprintf("\t\tKey length:\t\t %u\n",
+				le16_to_cpu(entry->key_length));
+		Vprintf("\t\tFlags:\t\t\t 0x%02x\n", le16_to_cpu(entry->flags));
+
+		if (entry->flags & INDEX_ENTRY_NODE)
+			Vprintf("\t\tSubnode VCN:\t\t %lld\n", le64_to_cpu(
+					*((u8*)entry + le16_to_cpu(
+					entry->length) - sizeof(VCN))));
+		if (entry->flags & INDEX_ENTRY_END)
+			break;
+
+		switch(type) {
+			case(AT_FILE_NAME):
+				Vprintf("\t\tFILE record number:\t %llu\n",
+						MREF_LE(entry->indexed_file));
+				if (opts.verbose) {
+					printf("\t");
+					ntfs_dump_flags(type, entry->key.
+						file_name.file_attributes);
+					printf("\t");
+					ntfs_dump_namespace(entry->key.
+						file_name.file_name_type);
+				}
+				ntfs_ucstombs(entry->key.file_name.file_name,
+					entry->key.file_name.file_name_length,
+					&name, 0);
+				Vprintf("\t\tName:\t\t\t %s\n", name);
+				free(name);
+				name = NULL;
+				Vprintf("\t\tParent directory:\t %lld\n",
+					 MREF_LE(entry->
+					 key.file_name.parent_directory));
+				Vprintf("\t\tData size:\t\t %lld\n",
+					sle64_to_cpu(
+					entry->key.file_name.data_size));
+				Vprintf("\t\tAllocated size:\t\t %lld\n",
+					sle64_to_cpu(
+					entry->key.file_name.allocated_size));
+				break;
+			default:
+				// TODO: determine more attribute types
+				Vprintf("\t\tData offset:\t\t %u\n",
+					le16_to_cpu(entry->data_offset));
+				Vprintf("\t\tData length:\t\t %u\n",
+					le16_to_cpu(entry->data_length));
+				break;
+		}
+		entry = (INDEX_ENTRY *)((u8 *)entry +
+						le16_to_cpu(entry->length));
+		numb_entries++;
+	}
+	Vprintf("\tEnd of index block reached\n");
+	return numb_entries;
+}
+
 /**
  * ntfs_dump_attr_index_root()
  *
@@ -903,21 +1174,22 @@ static void ntfs_dump_attr_index_root(ATTR_RECORD *attr)
 {
 	unsigned int type;
 	INDEX_ROOT *index_root = NULL;
-	
+	INDEX_ENTRY *entry;
+
 	index_root = (INDEX_ROOT*)((u8*)attr + le16_to_cpu(attr->value_offset));
-	
+
 	printf("Dumping attribute $INDEX_ROOT (0x90)\n");
-	
+
 	/* Dump index name */
 	if (attr->name_length) {
 		char *index_name = NULL;
 		index_name = ntfs_attr_get_name(attr);
-		
+
 		if (index_name) {
 			printf("\tIndex name:\t\t '%s'\n",index_name);
 			free(index_name);
 		} else {
-			/* an error occured, errno holds the reason - notify the user */
+			/* an error occurred, errno holds the reason - notify the user */
 			fprintf(stderr, "ntfsinfo error: could not parse index name: %s\n",
 				strerror(errno));
 		}
@@ -925,12 +1197,14 @@ static void ntfs_dump_attr_index_root(ATTR_RECORD *attr)
 		printf("\tIndex name:\t\t unnamed\n");
 	}
 
+	printf("\tAttribute instance:\t %u\n", le16_to_cpu(attr->instance));
+
 	/* attr_type dumping */
 	printf("\tIndexed Attr Type:\t ");
 	type = le32_to_cpu(index_root->type);
 	if (type) {
 		if (index_root->type != AT_FILE_NAME) {
-			/* wierd, this should be illgeal */
+			/* weird, this should be illgeal */
 			printf("0x%0X\n", type);
 			fprintf(stderr, "ntfsinfo error: Unknown Indexed Attr Type: 0x%0X\n",
 				type);
@@ -953,7 +1227,7 @@ static void ntfs_dump_attr_index_root(ATTR_RECORD *attr)
 		(unsigned int)le32_to_cpu(index_root->index_block_size));
 	printf("\tClusters Per Block:\t %u\n",
 		index_root->clusters_per_index_block);
-	
+
 	/* index header starts here */
 	printf("\tAllocated Size:\t\t %u\n",
 		(unsigned int)le32_to_cpu(index_root->index.allocated_size));
@@ -962,7 +1236,152 @@ static void ntfs_dump_attr_index_root(ATTR_RECORD *attr)
 
 	/* the flags are 8bit long, no need for byte-order handling */
 	printf("\tFlags:\t\t\t 0x%02x\n",index_root->index.flags);
-	/* printf("\tIndex Entries Following\t %u\n", ???? );*/
+
+	entry = (INDEX_ENTRY *)((u8 *)index_root +
+			le32_to_cpu(index_root->index.entries_offset) + 0x10);
+	Vprintf("\tDumping index block:");
+	printf("\tIndex entries total:\t %d\n",
+			ntfs_dump_index_entries(entry, index_root->type));
+}
+
+/**
+ * ntfs_dump_attr_index_allocation()
+ *
+ * determine size and type of INDX record
+ */
+static int get_type_and_size_of_indx(ntfs_inode *ni, ATTR_RECORD *attr,
+						ATTR_TYPES *type, u32 *size)
+{
+	ntfs_attr_search_ctx *ctx;
+	ntfschar *name = 0;
+	INDEX_ROOT *root;
+
+	if (attr->name_length) {
+		name = malloc(attr->name_length * sizeof(ntfschar));
+		if (!name) {
+			perror("malloc failed");
+			return -1;
+		}
+		memcpy(name, (u8 *)attr + attr->name_offset,
+				attr->name_length * sizeof(ntfschar));
+	}
+	ctx = ntfs_attr_get_search_ctx(ni, NULL);
+	if (!ctx) {
+		perror("ntfs_get_search_ctx failed");
+		free(name);
+		return -1;
+	}
+	if (ntfs_attr_lookup(AT_INDEX_ROOT, name, attr->name_length, 0,
+							0, NULL, 0, ctx)) {
+		perror("ntfs_attr_lookup failed");
+		ntfs_attr_put_search_ctx(ctx);
+		free(name);
+		return -1;
+	}
+
+	root = (INDEX_ROOT*)((u8*)ctx->attr +
+				le16_to_cpu(ctx->attr->value_offset));
+	*size = le32_to_cpu(root->index_block_size);
+	*type = root->type;
+	ntfs_attr_put_search_ctx(ctx);
+	free(name);
+	return 0;
+}
+
+/**
+ * ntfs_dump_attr_index_allocation()
+ *
+ * dump context of the index_allocation attribute
+ */
+static void ntfs_dump_index_allocation(ATTR_RECORD *attr, ntfs_inode *ni)
+{
+	INDEX_ALLOCATION *allocation, *tmp_alloc;
+	INDEX_ENTRY *entry;
+	ATTR_TYPES type;
+	u32 indx_record_size;
+	int total_entries = 0;
+	int total_indx_blocks = 0;
+	ntfs_attr *na;
+	u8 *bitmap, *byte;
+	int bit;
+	ntfschar *name;
+
+	if (get_type_and_size_of_indx(ni, attr, &type, &indx_record_size))
+		return;
+
+	name = (ntfschar*)((u8*)attr + attr->name_offset);
+	na = ntfs_attr_open(ni, AT_BITMAP, name, attr->name_length);
+	if (!na) {
+		perror("ntfs_attr_open failed");
+		return;
+	}
+	bitmap = malloc(na->data_size);
+	if (!bitmap) {
+		perror("malloc failed");
+		return;
+	}
+	if (ntfs_attr_pread(na, 0, na->data_size, bitmap) != na->data_size) {
+		perror("ntfs_attr_pread failed");
+		free(bitmap);
+		return;
+	}
+	ntfs_attr_close(na);
+	byte = bitmap;
+
+	na = ntfs_attr_open(ni, AT_INDEX_ALLOCATION, name, attr->name_length);
+	if (!na) {
+		perror("ntfs_attr_open failed");
+		free(bitmap);
+		return;
+	}
+	allocation = malloc(na->data_size);
+	if (!allocation) {
+		perror("malloc failed");
+		free(bitmap);
+		return;
+	}
+	if (ntfs_attr_pread(na, 0, na->data_size, allocation)
+							 != na->data_size) {
+		perror("ntfs_attr_pread failed");
+		free(allocation);
+		free(bitmap);
+		return;
+	}
+	ntfs_attr_close(na);
+	tmp_alloc = allocation;
+
+	bit = 0;
+	while((u8 *)tmp_alloc < (u8 *)allocation + na->data_size) {
+		if (*byte & (1 << bit)) {
+			if (ntfs_mst_post_read_fixup((NTFS_RECORD *) tmp_alloc,
+						indx_record_size)) {
+				perror("Damaged INDX record");
+				goto free;
+			}
+			entry = (INDEX_ENTRY *)((u8 *)tmp_alloc + le32_to_cpu(
+				tmp_alloc->index.entries_offset) + 0x18);
+			Vprintf("\tDumping index block "
+					"(VCN %lld, used %u/%u):", le64_to_cpu(
+					tmp_alloc->index_block_vcn),
+					le32_to_cpu(tmp_alloc->index.
+					index_length), le32_to_cpu(tmp_alloc->
+					index.allocated_size));
+			total_entries += ntfs_dump_index_entries(entry, type);
+			total_indx_blocks++;
+		}
+		tmp_alloc = (INDEX_ALLOCATION *)((u8 *)tmp_alloc +
+							indx_record_size);
+		bit++;
+		if (bit > 7) {
+			bit = 0;
+			byte++;
+		}
+	}
+	printf("\tIndex entries total:\t %d\n", total_entries);
+	printf("\tINDX blocks total:\t %d\n", total_indx_blocks);
+free:
+	free(allocation);
+	free(bitmap);
 }
 
 /**
@@ -970,7 +1389,7 @@ static void ntfs_dump_attr_index_root(ATTR_RECORD *attr)
  *
  * dump the index_allocation attribute
  */
-static void ntfs_dump_attr_index_allocation(ATTR_RECORD *attr)
+static void ntfs_dump_attr_index_allocation(ATTR_RECORD *attr, ntfs_inode *ni)
 {
 	printf("Dumping attribute $INDEX_ALLOCATION (0xA0)\n");
 
@@ -978,19 +1397,24 @@ static void ntfs_dump_attr_index_allocation(ATTR_RECORD *attr)
 	if (attr->name_length) {
 		char *index_name = NULL;
 		index_name = ntfs_attr_get_name(attr);
-		
+
 		if (index_name) {
 			printf("\tIndex name:\t\t '%s'\n",index_name);
 			free(index_name);
 		} else {
-			/* an error occured, errno holds the reason - notify the user */
-			fprintf(stderr, "ntfsinfo error: could not parse index name: %s\n",
-				strerror(errno));
+			/*
+			 * An error occurred, errno holds the reason -
+			 * notify the user
+			 */
+			fprintf(stderr, "ntfsinfo error: could not parse "
+					"index name: %s\n", strerror(errno));
 		}
 	} else {
 		printf("\tIndex name:\t\t unnamed\n");
 	}
 
+	printf("\tAttribute instance:\t %u\n", le16_to_cpu(attr->instance));
+
 	/* dump index's size */
 	if (attr->non_resident) {
 		/* print only the non resident part's size */
@@ -999,12 +1423,11 @@ static void ntfs_dump_attr_index_allocation(ATTR_RECORD *attr)
 		printf("\tUsed data size:\t\t %llu\n",
 			(unsigned long long)le64_to_cpu(attr->data_size));
 	} else {
-		/* print only the payload's size */
-		printf("\tValue's size:\t\t %u\n",
-			(unsigned int)le32_to_cpu(attr->value_length));
+		Eprintf("Invalid $INDEX_ALLOCATION attribute. Should be be"
+						    " non-resident\n");
 	}
-	
-	/* TODO: parse how many records does this B-*+/Tree contains */
+
+	ntfs_dump_index_allocation(attr, ni);
 }
 
 /**
@@ -1020,12 +1443,12 @@ static void ntfs_dump_attr_bitmap(ATTR_RECORD *attr)
 	if (attr->name_length) {
 		char *bitmap_name = NULL;
 		bitmap_name = ntfs_attr_get_name(attr);
-		
+
 		if (bitmap_name) {
 			printf("\tBitmap name:\t\t '%s'\n",bitmap_name);
 			free(bitmap_name);
 		} else {
-			/* an error occured, errno holds the reason - notify the user */
+			/* an error occurred, errno holds the reason - notify the user */
 			fprintf(stderr, "ntfsinfo error: could not parse bitmap name: %s\n",
 				strerror(errno));
 		}
@@ -1033,6 +1456,8 @@ static void ntfs_dump_attr_bitmap(ATTR_RECORD *attr)
 		printf("\tBitmap name:\t\t unnamed\n");
 	}
 
+	printf("\tAttribute instance:\t %u\n", le16_to_cpu(attr->instance));
+
 	/* dump bitmap size */
 	if (attr->non_resident) {
 		/* print only the non resident part's size */
@@ -1065,10 +1490,18 @@ static void ntfs_dump_attr_reparse_point(ATTR_RECORD *attr __attribute__((unused
  *
  * dump the ea_information attribute
  */
-static void ntfs_dump_attr_ea_information(ATTR_RECORD *attr __attribute__((unused)))
+static void ntfs_dump_attr_ea_information(ATTR_RECORD *attr)
 {
+	EA_INFORMATION *ea_info;
+	
+	ea_info = (EA_INFORMATION*)((u8*)attr +
+			le16_to_cpu(attr->value_offset));
 	printf("Dumping attribute $EA_INFORMATION (0xD0)\n");
-	printf("\tTODO\n");
+	printf("\tPacked EA length:\t %u\n", le16_to_cpu(ea_info->ea_length));
+	printf("\tNEED_EA count:\t\t %u\n",
+			le16_to_cpu(ea_info->need_ea_count));
+	printf("\tUnpacked EA length:\t %u\n",
+			(unsigned)le32_to_cpu(ea_info->ea_query_length));
 }
 
 /**
@@ -1076,10 +1509,77 @@ static void ntfs_dump_attr_ea_information(ATTR_RECORD *attr __attribute__((unuse
  *
  * dump the ea attribute
  */
-static void ntfs_dump_attr_ea(ATTR_RECORD *attr __attribute__((unused)))
+static void ntfs_dump_attr_ea(ATTR_RECORD *attr, ntfs_volume *vol)
 {
+	EA_ATTR *ea;
+	u8 *buf = NULL;
+	s64 data_size;
+
 	printf("Dumping attribute $EA (0xE0)\n");
-	printf("\tTODO\n");
+	if (attr->non_resident) {
+		runlist *rl;
+
+		data_size = sle64_to_cpu(attr->data_size);	
+		printf("\tIs resident? \t\t No\n");
+		printf("\tData size:\t\t %lld\n", data_size);
+		if (!opts.verbose)
+			return;
+		/* FIXME: We don't handle fragmented mapping pairs case. */
+		rl = ntfs_mapping_pairs_decompress(vol, attr, 0);
+		if (rl) {
+			s64 bytes_read;
+
+			buf = malloc(data_size);
+			if (!buf) {
+				perror("malloc failed");
+				free(rl);
+				return;
+			}
+			bytes_read = ntfs_rl_pread(vol, rl, 0, data_size, buf);
+			if (bytes_read != data_size) {
+				perror("ntfs_rl_pread failed");
+				free(buf);
+				free(rl);
+				return;
+			}
+			free(rl);
+			ea = (EA_ATTR*)buf;
+		} else {
+			perror("ntfs_mapping_pairs_decompress failed");
+			return;
+		}
+	} else {
+		data_size = le32_to_cpu(attr->value_length);
+		printf("\tIs resident? \t\t Yes\n");
+		printf("\tAttribute value length:\t %lld\n", data_size);
+		if (!opts.verbose)
+			return;
+		ea = (EA_ATTR*)((u8*)attr + le16_to_cpu(attr->value_offset));
+	}
+	while (1) {
+		printf("\n\tFlags:\t\t ");
+		if (ea->flags) {
+			if (ea->flags == NEED_EA)
+				printf("NEED_EA\n");
+			else
+				printf("Unknown (0x%02x)\n", ea->flags);
+		} else
+			printf("\n");
+		printf("\tName length:\t %d\n", ea->name_length);
+		printf("\tValue length:\t %d\n",
+				le16_to_cpu(ea->value_length));
+		printf("\tName:\t\t '%s'\n", ea->name);
+		printf("\tValue:\t\t '%s'\n", ea->value + ea->name_length + 1);
+		if (ea->next_entry_offset)
+			ea = (EA_ATTR*)((u8*)ea +
+					le32_to_cpu(ea->next_entry_offset));
+		else
+			break;
+		if ((u8*)ea - buf >= data_size)
+			break;
+	}
+	if (buf)
+		free(buf);
 }
 
 /**
@@ -1121,12 +1621,12 @@ static void ntfs_hex_dump(void *buf,unsigned int length)
 			unsigned char c = *((char *)buf + j);
 			printf("%02hhX ",c);
 		}
-		
+
 		/* realign */
 		for (;j<i+16;j++) {
 			printf("   ");
 		}
-		
+
 		/* char content */
 		for (j=i;(j<length) && (j<i+16);j++) {
 			unsigned char c = *((char *)buf + j);
@@ -1136,7 +1636,7 @@ static void ntfs_hex_dump(void *buf,unsigned int length)
 			}
 			printf("%c",c);
 		}
-		
+
 		/* end line */
 		printf("\n");
 		i=j;
@@ -1170,7 +1670,7 @@ static void ntfs_dump_attr_unknown(ATTR_RECORD *attr)
 			printf("\tAttribute name:\t '%s'\n",attr_name);
 			free(attr_name);
 		} else {
-			/* an error occured, errno holds the reason - notify the user */
+			/* an error occurred, errno holds the reason - notify the user */
 			fprintf(stderr, "ntfsinfo error: could not parse attribute name: %s\n",
 				strerror(errno));
 		}
@@ -1220,9 +1720,9 @@ static void ntfs_dump_attr_unknown(ATTR_RECORD *attr)
 static void ntfs_dump_inode_general_info(ntfs_inode *inode)
 {
 	u16 inode_flags = inode->mrec->flags;
-	
+
 	printf("Dumping Inode #%llu\n",(long long)inode->mft_no);
-	
+
 	printf("Update Sequence Array Count:\t %hu\n",
 		le16_to_cpu(inode->mrec->usa_count));
 	printf("$LogFile seqNum for this Inode:\t 0x%llx\n",
@@ -1252,7 +1752,7 @@ static void ntfs_dump_inode_general_info(ntfs_inode *inode)
 		(unsigned int)le32_to_cpu(inode->mrec->bytes_in_use));
 	printf("Size - Allocated:\t\t %u bytes\n",
 		(unsigned int)le32_to_cpu(inode->mrec->bytes_allocated));
-		
+
 	if (inode->mrec->base_mft_record) {
 		printf("base MFT record:\t\t %llu\n",
 			MREF_LE(inode->mrec->base_mft_record));
@@ -1269,7 +1769,7 @@ static void ntfs_dump_file_attributes(ntfs_inode *inode)
 	ntfs_attr_search_ctx *ctx = NULL;
 
 	/* then start enumerating attributes
-	   see ntfs_attr_lookup documentation for detailed explenation */
+	   see ntfs_attr_lookup documentation for detailed explanation */
 	ctx = ntfs_attr_get_search_ctx(inode, NULL);
 	while (!ntfs_attr_lookup(AT_UNUSED, NULL, 0, 0, 0, NULL, 0, ctx)) {
 		switch (ctx->attr->type) {
@@ -1281,7 +1781,7 @@ static void ntfs_dump_file_attributes(ntfs_inode *inode)
 			ntfs_dump_attr_standard_information(ctx->attr);
 			break;
 		case AT_ATTRIBUTE_LIST:
-			ntfs_dump_attr_list(ctx->attr);
+			ntfs_dump_attr_list(ctx->attr, inode->vol);
 			break;
 		case AT_FILE_NAME:
 			ntfs_dump_attr_file_name(ctx->attr);
@@ -1305,7 +1805,7 @@ static void ntfs_dump_file_attributes(ntfs_inode *inode)
 			ntfs_dump_attr_index_root(ctx->attr);
 			break;
 		case AT_INDEX_ALLOCATION:
-			ntfs_dump_attr_index_allocation(ctx->attr);
+			ntfs_dump_attr_index_allocation(ctx->attr, inode);
 			break;
 		case AT_BITMAP:
 			ntfs_dump_attr_bitmap(ctx->attr);
@@ -1317,7 +1817,7 @@ static void ntfs_dump_file_attributes(ntfs_inode *inode)
 			ntfs_dump_attr_ea_information(ctx->attr);
 			break;
 		case AT_EA:
-			ntfs_dump_attr_ea(ctx->attr);
+			ntfs_dump_attr_ea(ctx->attr, inode->vol);
 			break;
 		case AT_PROPERTY_SET:
 			ntfs_dump_attr_property_set(ctx->attr);
@@ -1332,7 +1832,7 @@ static void ntfs_dump_file_attributes(ntfs_inode *inode)
 			ntfs_dump_attr_unknown(ctx->attr);
 		}
 	}
-	
+
 	/* if we exited the loop before we're done - notify the user */
 	if (errno != ENOENT) {
 		fprintf(stderr, "ntfsinfo error: stopped before finished "
@@ -1344,7 +1844,7 @@ static void ntfs_dump_file_attributes(ntfs_inode *inode)
 	/* close all data-structures we used */
 	ntfs_attr_put_search_ctx(ctx);
 	ntfs_inode_close(inode);
-	
+
 	/* happily exit */
 }
 
@@ -1380,7 +1880,7 @@ int main(int argc, char **argv)
 		ntfs_inode *inode;
 		/* obtain the inode */
 		if (opts.filename) {
-			inode = utils_pathname_to_inode (vol, NULL, opts.filename);
+			inode = ntfs_pathname_to_inode(vol, NULL, opts.filename);
 		} else {
 			inode = ntfs_inode_open(vol, MK_LE_MREF(opts.inode, 0));
 		}
