@@ -2,7 +2,7 @@
  * ntfsls - Part of the Linux-NTFS project.
  *
  * Copyright (c) 2003 Lode Leroy
- * Copyright (c) 2003 Anton Altaparmakov
+ * Copyright (c) 2003-2005 Anton Altaparmakov
  * Copyright (c) 2003 Richard Russon
  * Copyright (c) 2004 Carmelo Kintana
  * Copyright (c) 2004 Giang Nguyen
@@ -26,11 +26,21 @@
  */
 #include "config.h"
 
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_TIME_H
 #include <time.h>
+#endif
+#ifdef HAVE_GETOPT_H
 #include <getopt.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
 
 #include "types.h"
 #include "mft.h"
@@ -40,6 +50,8 @@
 #include "utils.h"
 #include "dir.h"
 #include "list.h"
+#include "ntfstime.h"
+#include "version.h"
 
 static const char *EXEC_NAME = "ntfsls";
 
@@ -124,10 +136,11 @@ GEN_PRINTF(Qprintf, stdout, &opts.quiet,   FALSE)
  */
 static void version(void)
 {
-	printf("\n%s v%s - Display information about an NTFS Volume.\n\n",
-			EXEC_NAME, VERSION);
+	printf("\n%s v%s (libntfs %s) - Display information about an NTFS "
+			"Volume.\n\n", EXEC_NAME, VERSION,
+			ntfs_libntfs_version());
 	printf("Copyright (c) 2003 Lode Leroy\n");
-	printf("Copyright (c) 2003 Anton Altaparmakov\n");
+	printf("Copyright (c) 2003-2005 Anton Altaparmakov\n");
 	printf("Copyright (c) 2003 Richard Russon\n");
 	printf("Copyright (c) 2004 Carmelo Kintana\n");
 	printf("Copyright (c) 2004 Giang Nguyen\n");
@@ -404,7 +417,7 @@ static int readdir_recursive(ntfs_inode * ni, s64 * pos, ntfsls_dirent * dirent)
 			dir_list_insert_pos = &dirs.list;
 			if (!subdir->ni) {
 				subdir->ni =
-				    utils_pathname_to_inode(ni->vol, ni,
+				    ntfs_pathname_to_inode(ni->vol, ni,
 							    subdir->name);
 
 				if (!subdir->ni) {
@@ -447,7 +460,7 @@ static int readdir_recursive(ntfs_inode * ni, s64 * pos, ntfsls_dirent * dirent)
 	if (tofree)
 		free_dir(tofree);
 
-	/* if at the outter-most readdir_recursive, then clean up */
+	/* if at the outer-most readdir_recursive, then clean up */
 	if (ni_depth == 0) {
 		free_dirs(&dirs.list);
 	}
@@ -481,7 +494,7 @@ static int list_dir_entry(ntfsls_dirent * dirent, const ntfschar * name,
 	}
 
 	result = 0;					// These are successful
-	if ((filename[0] == '$') && (!opts.system))
+	if ((MREF(mref) < FILE_first_user) && (!opts.system))
 		goto free;
 	if (name_type == FILE_NAME_POSIX && !opts.all)
 		goto free;
@@ -533,7 +546,7 @@ static int list_dir_entry(ntfsls_dirent * dirent, const ntfschar * name,
 		if (!ni)
 			goto release;
 
-		ctx = ntfs_attr_get_search_ctx(ni, ni->mrec);
+		ctx = ntfs_attr_get_search_ctx(ni, NULL);
 		if (!ctx)
 			goto release;
 
@@ -547,10 +560,10 @@ static int list_dir_entry(ntfsls_dirent * dirent, const ntfschar * name,
 		if (!file_name_attr)
 			goto release;
 
-		ntfs_time = ntfs2utc(sle64_to_cpu(
-				file_name_attr->last_data_change_time));
+		ntfs_time = ntfs2utc(file_name_attr->last_data_change_time);
 		strcpy(t_buf, ctime(&ntfs_time));
-		t_buf[16] = '\0';
+		memmove(t_buf+16, t_buf+19, 5);
+		t_buf[21] = '\0';
 
 		if (dt_type != NTFS_DT_DIR) {
 			if (!ntfs_attr_lookup(AT_DATA, AT_UNNAMED, 0, 0, 0,
@@ -575,7 +588,7 @@ static int list_dir_entry(ntfsls_dirent * dirent, const ntfschar * name,
 
 		result = 0;
 release:
-		/* Release atrtibute search context and close the inode. */
+		/* Release attribute search context and close the inode. */
 		if (ctx)
 			ntfs_attr_put_search_ctx(ctx);
 		if (ni)
@@ -628,7 +641,7 @@ int main(int argc, char **argv)
 		return 2;
 	}
 
-	ni = utils_pathname_to_inode (vol, NULL, opts.path);
+	ni = ntfs_pathname_to_inode (vol, NULL, opts.path);
 	if (!ni) {
 		// FIXME: Print error... (AIA)
 		ntfs_umount(vol, FALSE);
@@ -657,7 +670,7 @@ int main(int argc, char **argv)
 		ntfschar *name = NULL;
 		int name_len = 0;;
 
-		ctx = ntfs_attr_get_search_ctx (NULL, ni->mrec);
+		ctx = ntfs_attr_get_search_ctx(ni, NULL);
 		if (!ctx)
 			return -1;
 
