@@ -2,7 +2,7 @@
  * ntfsmove - Part of the Linux-NTFS project.
  *
  * Copyright (c) 2003 Richard Russon
- * Copyright (c) 2003 Anton Altaparmakov
+ * Copyright (c) 2003-2005 Anton Altaparmakov
  *
  * This utility will move files on an NTFS volume.
  *
@@ -24,10 +24,18 @@
 
 #include "config.h"
 
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_GETOPT_H
 #include <getopt.h>
+#endif
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
 
 #include "types.h"
 #include "attrib.h"
@@ -37,6 +45,7 @@
 #include "dir.h"
 #include "bitmap.h"
 #include "ntfsmove.h"
+#include "version.h"
 
 static const char *EXEC_NAME = "ntfsmove";
 static struct options opts;
@@ -59,8 +68,9 @@ GEN_PRINTF (Qprintf, stdout, &opts.quiet,   FALSE)
  */
 static void version (void)
 {
-	printf ("\n%s v%s - Move files and directories on an NTFS volume.\n\n",
-		EXEC_NAME, VERSION);
+	printf ("\n%s v%s (libntfs %s) - Move files and directories on an "
+			"NTFS volume.\n\n", EXEC_NAME, VERSION,
+			ntfs_libntfs_version());
 	printf ("Copyright (c) 2003 Richard Russon\n");
 	printf ("\n%s\n%s%s\n", ntfs_gpl, ntfs_bugs, ntfs_home);
 }
@@ -365,7 +375,8 @@ static void dump_runs (u8 *buffer, int len)
 /**
  * find_unused
  */
-static runlist * find_unused (ntfs_volume *vol, s64 size, u64 loc, int flags)
+static runlist * find_unused (ntfs_volume *vol, s64 size, u64 loc
+	__attribute__((unused)), int flags __attribute__((unused)))
 {
 	const int bufsize = 8192;
 	u8 *buffer;
@@ -639,7 +650,7 @@ static s64 move_runlist (ntfs_volume *vol, runlist_element *from,
 
 // got to get the runlist out of this function
 //      requires a mrec arg, not an ino (ino->mrec will do for now)
-// check size of new runlist before allocting / moving
+// check size of new runlist before allocating / moving
 // replace one datarun with another (by hand)
 static s64 move_datarun (ntfs_volume *vol, ntfs_inode *ino, ATTR_RECORD *rec,
 	runlist_element *run, u64 loc, int flags)
@@ -669,7 +680,7 @@ static s64 move_datarun (ntfs_volume *vol, ntfs_inode *ino, ATTR_RECORD *rec,
 
 	printf ("move %lld,%lld,%lld to %lld,%lld,%lld\n", run->vcn, run->lcn, run->length, to->vcn, to->lcn, to->length);
 
-	need_from = ntfs_get_size_for_mapping_pairs (vol, from);
+	need_from = ntfs_get_size_for_mapping_pairs(vol, from, 0);
 	printf ("orig data run = %d bytes\n", need_from);
 
 	//ntfs_debug_runlist_dump2 (from, 5, "\t");
@@ -683,7 +694,7 @@ static s64 move_datarun (ntfs_volume *vol, ntfs_inode *ino, ATTR_RECORD *rec,
 
 	//ntfs_debug_runlist_dump2 (from, 5, "\t");
 
-	need_to = ntfs_get_size_for_mapping_pairs (vol, from);
+	need_to = ntfs_get_size_for_mapping_pairs(vol, from, 0);
 	printf ("new  data run = %d bytes\n", need_to);
 
 	need_from = calc_attr_length (rec, need_from);
@@ -703,12 +714,13 @@ static s64 move_datarun (ntfs_volume *vol, ntfs_inode *ino, ATTR_RECORD *rec,
 		Eprintf ("!move_runlist\n");
 		return -1;
 	}
-	
+
 	// wipe orig runs
 	memset (((u8*)rec) +rec->mapping_pairs_offset, 0, need_to - rec->mapping_pairs_offset);
 
 	// update data runs
-	ntfs_mapping_pairs_build (vol, ((u8*)rec) + rec->mapping_pairs_offset, need_to, from);
+	ntfs_mapping_pairs_build(vol, ((u8*)rec) + rec->mapping_pairs_offset,
+			need_to, from, 0, NULL);
 
 	// commit
 	ntfs_inode_mark_dirty (ino);
@@ -857,7 +869,7 @@ int main (int argc, char *argv[])
 		return 1;
 	}
 
-	inode = utils_pathname_to_inode (vol, NULL, opts.file);
+	inode = ntfs_pathname_to_inode(vol, NULL, opts.file);
 	if (!inode) {
 		printf ("!inode\n");
 		return 1;
@@ -865,7 +877,8 @@ int main (int argc, char *argv[])
 
 	count = move_file (vol, inode, opts.location, 0);
 	if ((count > 0) && (!opts.nodirty)) {
-		if (ntfs_volume_set_flags (vol, VOLUME_IS_DIRTY) < 0) {
+		if (ntfs_volume_write_flags (vol, vol->flags | VOLUME_IS_DIRTY) <
+				0) {
 			Eprintf ("Couldn't mark volume dirty\n");
 		}
 		printf ("Relocated %lld bytes\n", count);
@@ -882,4 +895,3 @@ int main (int argc, char *argv[])
 	ntfs_umount (vol, FALSE);
 	return result;
 }
-
