@@ -1,7 +1,7 @@
 /**
  * ntfstruncate - Part of the Linux-NTFS project.
  *
- * Copyright (c) 2002-2003 Anton Altaparmakov
+ * Copyright (c) 2002-2005 Anton Altaparmakov
  *
  * This utility will truncate a specified attribute belonging to a
  * specified inode, i.e. file or directory, to a specified length.
@@ -42,14 +42,18 @@
 #ifdef HAVE_ERRNO_H
 #	include <errno.h>
 #endif
+#ifdef HAVE_TIME_H
 #include <time.h>
+#endif
 #ifdef HAVE_GETOPT_H
 #	include <getopt.h>
 #else
 	extern char *optarg;
 	extern int optind;
 #endif
+#ifdef HAVE_LIMITS_H
 #include <limits.h>
+#endif
 #ifndef LLONG_MAX
 #	define LLONG_MAX 9223372036854775807LL
 #endif
@@ -60,6 +64,7 @@
 #include "layout.h"
 #include "volume.h"
 #include "utils.h"
+#include "version.h"
 
 extern const unsigned char attrdef_ntfs12_array[2400];
 
@@ -93,10 +98,11 @@ struct {
 } opts;
 
 /**
- * Dprintf - debugging output (-vv); overriden by quiet (-q)
+ * mkDprintf - debugging output (-vv); overridden by quiet (-q)
  */
-static void Dprintf(const char *fmt, ...) __attribute__((format(printf, 1, 2)));
-static void Dprintf(const char *fmt, ...)
+static void mkDprintf(const char *fmt, ...)
+		__attribute__((format(printf, 1, 2)));
+static void mkDprintf(const char *fmt, ...)
 {
 	va_list ap;
 
@@ -151,14 +157,14 @@ static void err_exit(const char *fmt, ...)
  */
 static void copyright(void)
 {
-	fprintf(stderr, "Copyright (c) 2002-2003 Anton Altaparmakov\n"
+	fprintf(stderr, "Copyright (c) 2002-2005 Anton Altaparmakov\n"
 			"Copyright (c) 2003 Richard Russon\n"
 			"Truncate a specified attribute of a specified "
 			"inode.\n");
 }
 
 /**
- * license - print licese statement
+ * license - print license statement
  */
 static void license(void)
 {
@@ -201,7 +207,8 @@ static void parse_options(int argc, char *argv[])
 
 	if (argc && *argv)
 		EXEC_NAME = *argv;
-	fprintf(stderr, "%s v%s\n", EXEC_NAME, VERSION);
+	fprintf(stderr, "%s v%s (libntfs %s)\n", EXEC_NAME, VERSION,
+			ntfs_libntfs_version());
 	while ((c = getopt(argc, argv, "fh?nqvVl")) != EOF)
 		switch (c) {
 		case 'f':
@@ -233,7 +240,7 @@ static void parse_options(int argc, char *argv[])
 
 	/* Get the device. */
 	dev_name = argv[optind++];
-	Dprintf("device name = %s\n", dev_name);
+	mkDprintf("device name = %s\n", dev_name);
 
 	if (optind == argc)
 		usage();
@@ -243,7 +250,7 @@ static void parse_options(int argc, char *argv[])
 	if (*s || !ll || (ll >= LLONG_MAX && errno == ERANGE))
 		err_exit("Invalid inode number: %s\n", argv[optind - 1]);
 	inode = ll;
-	Dprintf("inode = %lli\n", (long long)inode);
+	mkDprintf("inode = %lli\n", (long long)inode);
 
 	if (optind == argc)
 		usage();
@@ -283,11 +290,11 @@ static void parse_options(int argc, char *argv[])
 			attr_name_len = 0;
 		}
 	}
-	Dprintf("attribute type = 0x%x\n", (unsigned int)attr_type);
+	mkDprintf("attribute type = 0x%x\n", (unsigned int)attr_type);
 	if (attr_name == AT_UNNAMED)
-		Dprintf("attribute name = \"\" (UNNAMED)\n");
+		mkDprintf("attribute name = \"\" (UNNAMED)\n");
 	else
-		Dprintf("attribute name = \"%s\" (length %u Unicode "
+		mkDprintf("attribute name = \"%s\" (length %u Unicode "
 				"characters)\n", s2,
 				(unsigned int)attr_name_len);
 
@@ -296,7 +303,7 @@ static void parse_options(int argc, char *argv[])
 	if (*s2 || ll < 0 || (ll >= LLONG_MAX && errno == ERANGE))
 		err_exit("Invalid new length: %s\n", s);
 	new_len = ll;
-	Dprintf("new length = %lli\n", new_len);
+	mkDprintf("new length = %lli\n", new_len);
 }
 
 /**
@@ -502,7 +509,8 @@ static void dump_resident_attr(ATTR_RECORD *a)
 /**
  * dump_mapping_pairs_array
  */
-static void dump_mapping_pairs_array(char *b, unsigned int max_len)
+static void dump_mapping_pairs_array(char *b __attribute__((unused)),
+	unsigned int max_len __attribute__((unused)))
 {
 	// TODO
 	return;
@@ -791,7 +799,7 @@ int main(int argc, char **argv)
 				(unsigned int)attr_type, strerror(errno));
 
 	if (!opts.quiet && opts.verbose > 1) {
-		Dprintf("Dumping mft record before calling "
+		mkDprintf("Dumping mft record before calling "
 				"ntfs_attr_truncate():\n");
 		dump_mft_record(ni->mrec);
 	}
@@ -803,7 +811,7 @@ int main(int argc, char **argv)
 				(unsigned int)attr_type, strerror(errno));
 
 	if (!opts.quiet && opts.verbose > 1) {
-		Dprintf("Dumping mft record after calling "
+		mkDprintf("Dumping mft record after calling "
 				"ntfs_attr_truncate():\n");
 		dump_mft_record(ni->mrec);
 	}
