@@ -1,6 +1,7 @@
 /**
  * ntfswipe - Part of the Linux-NTFS project.
  *
+ * Copyright (c) 2005 Anton Altaparmakov
  * Copyright (c) 2002-2003 Richard Russon
  * Copyright (c) 2004 Yura Pakhuchiy
  *
@@ -24,13 +25,27 @@
 
 #include "config.h"
 
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_ERRNO_H
 #include <errno.h>
+#endif
+#ifdef HAVE_STDARG_H
 #include <stdarg.h>
+#endif
+#ifdef HAVE_GETOPT_H
 #include <getopt.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
+#endif
 
 #include "ntfswipe.h"
 #include "types.h"
@@ -39,6 +54,7 @@
 #include "debug.h"
 #include "dir.h"
 #include "mst.h"
+#include "version.h"
 
 static const char *EXEC_NAME = "ntfswipe";
 static struct options opts;
@@ -56,8 +72,9 @@ GEN_PRINTF (Qprintf, stdout, &opts.quiet,   FALSE)
  */
 static void version (void)
 {
-	printf ("\n%s v%s - Overwrite the unused space on an NTFS Volume.\n\n",
-		EXEC_NAME, VERSION);
+	printf("\n%s v%s (libntfs %s) - Overwrite the unused space on an NTFS "
+			"Volume.\n\n", EXEC_NAME, VERSION,
+			ntfs_libntfs_version());
 	printf ("Copyright (c) 2002-2003 Richard Russon\n");
 	printf ("Copyright (c) 2004 Yura Pakhuchiy\n");
 	printf ("\n%s\n%s%s\n", ntfs_gpl, ntfs_bugs, ntfs_home);
@@ -352,7 +369,7 @@ static int parse_options (int argc, char *argv[])
  *
  * Read $Bitmap and wipe any clusters that are marked as not in use.
  *
- * Return: >0  Success, the atrribute was wiped
+ * Return: >0  Success, the attribute was wiped
  *          0  Nothing to wipe
  *         -1  Error, something went wrong
  */
@@ -406,7 +423,7 @@ free:
  * @act:	Wipe, test or info
  * @na:		Opened ntfs attribute
  *
- * Return: >0  Success, the atrribute was wiped
+ * Return: >0  Success, the attribute was wiped
  *          0  Nothing to wipe
  *         -1  Error, something went wrong
  */
@@ -419,7 +436,7 @@ static s64 wipe_compressed_attribute (ntfs_volume *vol, int byte,
 	VCN cur_vcn = 0;
 	runlist *rlc = na->rl;
 	s64 cu_mask = na->compression_block_clusters - 1;
-		
+
 	while (rlc->length) {
 		cur_vcn += rlc->length;
 		if ((cur_vcn & cu_mask) ||
@@ -465,20 +482,20 @@ static s64 wipe_compressed_attribute (ntfs_volume *vol, int byte,
 			size = na->allocated_size - na->data_size;
 			offset = (cur_vcn << vol->cluster_size_bits) - size;
 		}
-		
+
 		if (size < 0) {
 			Vprintf ("Internal error\n");
 			Eprintf ("bug or damaged fs: we want "
 				"allocate buffer size %lld bytes", size);
 			return -1;
 		}
-		
+
 		if ((act == act_info) || (!size)) {
 			wiped += size;
 			rlc++;
 			continue;
 		}
-		
+
 		buf = malloc (size);
 		if (!buf) {
 			Vprintf ("Not enough memory\n");
@@ -487,7 +504,7 @@ static s64 wipe_compressed_attribute (ntfs_volume *vol, int byte,
 			return -1;
 		}
 		memset (buf, byte, size);
-		
+
 		ret = ntfs_rl_pwrite (vol, na->rl, offset, size, buf);
 		free (buf);
 		if (ret != size) {
@@ -500,7 +517,7 @@ static s64 wipe_compressed_attribute (ntfs_volume *vol, int byte,
 next:
 		rlc++;
 	}
-	
+
 	return wiped;
 }
 
@@ -511,11 +528,11 @@ next:
  * @act:	Wipe, test or info
  * @na:		Opened ntfs attribute
  *
- * Return: >0  Success, the atrribute was wiped
+ * Return: >0  Success, the attribute was wiped
  *          0  Nothing to wipe
  *         -1  Error, something went wrong
  */
-static s64 wipe_attribute (ntfs_volume *vol, int byte, enum action act, 
+static s64 wipe_attribute (ntfs_volume *vol, int byte, enum action act,
 								ntfs_attr *na)
 {
 	unsigned char *buf;
@@ -528,7 +545,7 @@ static s64 wipe_attribute (ntfs_volume *vol, int byte, enum action act,
 	if (NAttrEncrypted(na))
 		offset = (((offset - 1) >> 10) + 1) << 10;
 	size = (vol->cluster_size - offset) % vol->cluster_size;
-	
+
 	if (act == act_info)
 		return size;
 
@@ -566,14 +583,17 @@ static s64 wipe_attribute (ntfs_volume *vol, int byte, enum action act,
 static s64 wipe_tails (ntfs_volume *vol, int byte, enum action act)
 {
 	s64 total = 0;
-	s64 inode_num;
+	s64 nr_mft_records, inode_num;
 	ntfs_inode *ni;
 	ntfs_attr *na;
 
 	if (!vol || (byte < 0))
 		return -1;
 
-	for (inode_num = 16; inode_num < vol->nr_mft_records; inode_num++) {
+	nr_mft_records = vol->mft_na->initialized_size >>
+			vol->mft_record_size_bits;
+
+	for (inode_num = 16; inode_num < nr_mft_records; inode_num++) {
 		s64 wiped;
 
 		Vprintf ("Inode %lld - ", inode_num);
@@ -595,7 +615,7 @@ static s64 wipe_tails (ntfs_volume *vol, int byte, enum action act)
 		}
 
 		if (!NAttrNonResident(na)) {
-			Vprintf ("Resident $DATA atrribute. Skipping.\n");
+			Vprintf ("Resident $DATA attribute. Skipping.\n");
 			goto close_attr;
 		}
 
@@ -614,7 +634,7 @@ static s64 wipe_tails (ntfs_volume *vol, int byte, enum action act)
 			Eprintf (" (inode %lld)\n", inode_num);
 			goto close_attr;
 		}
-		
+
 		if (wiped) {
 			Vprintf ("Wiped %llu bytes\n", wiped);
 			total += wiped;
@@ -646,7 +666,7 @@ static s64 wipe_mft (ntfs_volume *vol, int byte, enum action act)
 {
 	// by considering the individual attributes we might be able to
 	// wipe a few more bytes at the attr's tail.
-	s64 i;
+	s64 nr_mft_records, i;
 	s64 total = 0;
 	s64 result = 0;
 	int size = 0;
@@ -661,7 +681,10 @@ static s64 wipe_mft (ntfs_volume *vol, int byte, enum action act)
 		return -1;
 	}
 
-	for (i = 0; i < vol->nr_mft_records; i++) {
+	nr_mft_records = vol->mft_na->initialized_size >>
+			vol->mft_record_size_bits;
+
+	for (i = 0; i < nr_mft_records; i++) {
 		if (utils_mftrec_in_use (vol, i)) {
 			result = ntfs_attr_mst_pread (vol->mft_na, vol->mft_record_size * i,
 				1, vol->mft_record_size, buffer);
@@ -767,15 +790,17 @@ free:
  * @byte:		Overwrite with this value
  * @act:		Wipe, test or info
  * @naa:		Opened ntfs $INDEX_ALLOCATION attribute
- * @nab:		Opened ntfs $BIMTAP attribute
+ * @nab:		Opened ntfs $BITMAP attribute
  * @indx_record_size:	Size of INDX record
  *
  * Return: >0  Success, the clusters were wiped
  *          0  Nothing to wipe
  *         -1  Error, something went wrong
  */
-static s64 wipe_index_allocation (ntfs_volume *vol, int byte, enum action act,
-			ntfs_attr *naa, ntfs_attr *nab, u32 indx_record_size) {
+static s64 wipe_index_allocation (ntfs_volume *vol, int byte, enum action act
+	__attribute__((unused)), ntfs_attr *naa, ntfs_attr *nab,
+	u32 indx_record_size)
+{
 	s64 total = 0;
 	s64 wiped = 0;
 	s64 offset = 0;
@@ -786,14 +811,14 @@ static s64 wipe_index_allocation (ntfs_volume *vol, int byte, enum action act,
 	u8 mask;
 	u8 *bitmap;
 	u8 *buf;
-	
+
 	bitmap = malloc (nab->data_size);
 	if (!bitmap) {
 		Vprintf ("malloc failed\n");
 		Eprintf ("Couldn't allocate %lld bytes", nab->data_size);
 		return -1;
 	}
-	
+
 	if (ntfs_attr_pread (nab, 0, nab->data_size, bitmap)
 						!= nab->data_size) {
 		Vprintf ("Internal error\n");
@@ -801,7 +826,7 @@ static s64 wipe_index_allocation (ntfs_volume *vol, int byte, enum action act,
 		total = -1;
 		goto free_bitmap;
 	}
-	
+
 	buf = malloc (indx_record_size);
 	if (!buf) {
 		Vprintf ("malloc failed\n");
@@ -810,12 +835,12 @@ static s64 wipe_index_allocation (ntfs_volume *vol, int byte, enum action act,
 		total = -1;
 		goto free_bitmap;
 	}
-	
+
 	while (offset < naa->allocated_size) {
 		mask = 1 << obit;
 		if (bitmap[obyte] & mask) {
 			INDEX_ALLOCATION *indx;
-			
+
 			s64 ret = ntfs_rl_pread (vol, naa->rl,
 					offset, indx_record_size, buf);
 			if (ret != indx_record_size) {
@@ -824,13 +849,13 @@ static s64 wipe_index_allocation (ntfs_volume *vol, int byte, enum action act,
 				total = -1;
 				goto free_buf;
 			}
-			
+
 			indx = (INDEX_ALLOCATION *) buf;
 			if (ntfs_mst_post_read_fixup ((NTFS_RECORD *)buf,
 								indx_record_size))
 				Eprintf ("damaged fs: mst_post_read_fixup failed");
-    			
-			if ((le32_to_cpu(indx->index.allocated_size) + 0x18) != 
+
+			if ((le32_to_cpu(indx->index.allocated_size) + 0x18) !=
 							indx_record_size) {
 				Vprintf ("Internal error\n");
 				Eprintf ("INDX record should be %u bytes",
@@ -838,7 +863,7 @@ static s64 wipe_index_allocation (ntfs_volume *vol, int byte, enum action act,
 				total = -1;
 				goto free_buf;
 			}
-			
+
 			wipe_offset = le32_to_cpu(indx->index.index_length) + 0x18;
 			wipe_size = indx_record_size - wipe_offset;
 			memset (buf + wipe_offset, byte, wipe_size);
@@ -853,7 +878,7 @@ static s64 wipe_index_allocation (ntfs_volume *vol, int byte, enum action act,
 			if (opts.verbose > 1)
 				Vprintf ("x");
 		}
-		
+
 		wiped = ntfs_rl_pwrite (vol, naa->rl, offset, indx_record_size, buf);
 		if (wiped != indx_record_size) {
 			Vprintf ("ntfs_rl_pwrite failed\n");
@@ -889,13 +914,13 @@ free_bitmap:
 static u32 get_indx_record_size (ntfs_attr *nar)
 {
 	u32 indx_record_size;
-	
+
 	if (ntfs_attr_pread (nar, 8, 4, &indx_record_size) != 4) {
 		Vprintf ("Couldn't determine size of INDX record\n");
 		Eprintf ("ntfs_attr_pread failed");
 		return 0;
 	}
-	
+
 	indx_record_size = le32_to_cpu (indx_record_size);
 	if (!indx_record_size) {
 		Vprintf ("Internal error\n");
@@ -920,7 +945,7 @@ static u32 get_indx_record_size (ntfs_attr *nar)
 static s64 wipe_directory (ntfs_volume *vol, int byte, enum action act)
 {
 	s64 total = 0;
-	s64 inode_num;
+	s64 nr_mft_records, inode_num;
 	ntfs_inode *ni;
 	ntfs_attr *naa;
 	ntfs_attr *nab;
@@ -929,7 +954,10 @@ static s64 wipe_directory (ntfs_volume *vol, int byte, enum action act)
 	if (!vol || (byte < 0))
 		return -1;
 
-	for (inode_num = 5; inode_num < vol->nr_mft_records; inode_num++) {
+	nr_mft_records = vol->mft_na->initialized_size >>
+			vol->mft_record_size_bits;
+
+	for (inode_num = 5; inode_num < nr_mft_records; inode_num++) {
 		u32 indx_record_size;
 		s64 wiped;
 
@@ -959,7 +987,7 @@ static s64 wipe_directory (ntfs_volume *vol, int byte, enum action act)
 				Vprintf ("\r");
 			goto close_inode;
 		}
-		
+
 		if (!NAttrNonResident(naa)) {
 			Vprintf ("Resident $INDEX_ALLOCATION\n");
 			Eprintf ("damaged fs: Resident $INDEX_ALLOCATION "
@@ -982,7 +1010,7 @@ static s64 wipe_directory (ntfs_volume *vol, int byte, enum action act)
 					"name (inode %lld)\n", inode_num);
 			goto close_attr_allocation;
 		}
-		
+
 		nar = ntfs_attr_open (ni, AT_INDEX_ROOT, I30, 4);
 		if (!nar) {
 			Vprintf ("Couldn't open $INDEX_ROOT\n");
@@ -991,27 +1019,27 @@ static s64 wipe_directory (ntfs_volume *vol, int byte, enum action act)
 					" (inode %lld)\n", inode_num);
 			goto close_attr_bitmap;
 		}
-		
+
 		if (NAttrNonResident(nar)) {
 			Vprintf ("Not resident $INDEX_ROOT\n");
 			Eprintf ("damaged fs: Not resident $INDEX_ROOT "
 					"(inode %lld)\n", inode_num);
 			goto close_attr_root;
 		}
-		
+
 		indx_record_size = get_indx_record_size (nar);
 		if (!indx_record_size) {
 			Eprintf (" (inode %lld)\n", inode_num);
 			goto close_attr_root;
 		}
-		
+
 		wiped = wipe_index_allocation (vol, byte, act,
 						naa, nab, indx_record_size);
 		if (wiped == -1) {
 			Eprintf (" (inode %lld)\n", inode_num);
 			goto close_attr_root;
 		}
-		
+
 		if (wiped) {
 			Vprintf ("Wiped %llu bytes\n", wiped);
 			total += wiped;
@@ -1038,13 +1066,14 @@ close_inode:
  * @act:   Wipe, test or info
  *
  * The logfile journals the metadata to give the volume fault-tolerance.  If the
- * volume is in a consistant state, then this information can be erased.
+ * volume is in a consistent state, then this information can be erased.
  *
  * Return: >0  Success, the clusters were wiped
  *          0  Nothing to wipe
  *         -1  Error, something went wrong
  */
-static s64 wipe_logfile (ntfs_volume *vol, int byte, enum action act)
+static s64 wipe_logfile (ntfs_volume *vol, int byte, enum action act
+	__attribute__((unused)))
 {
 	const int NTFS_BUF_SIZE2 = 8192;
 	//FIXME(?): We might need to zero the LSN field of every single mft
@@ -1056,6 +1085,9 @@ static s64 wipe_logfile (ntfs_volume *vol, int byte, enum action act)
 	char buf[NTFS_BUF_SIZE2];
 	int eo;
 
+	/* We can wipe logfile only with 0xff. */
+	byte = 0xff;
+
 	if (!vol || (byte < 0))
 		return -1;
 
@@ -1148,7 +1180,8 @@ error_exit:
  *          0  Nothing to wipe
  *         -1  Error, something went wrong
  */
-static s64 wipe_pagefile (ntfs_volume *vol, int byte, enum action act)
+static s64 wipe_pagefile (ntfs_volume *vol, int byte, enum action act
+	__attribute__((unused)))
 {
 	// wipe completely, chkdsk doesn't do anything, booting writes header
 	const int NTFS_BUF_SIZE2 = 4096;
@@ -1162,8 +1195,8 @@ static s64 wipe_pagefile (ntfs_volume *vol, int byte, enum action act)
 		return -1;
 
 	//Qprintf ("wipe_pagefile (not implemented) 0x%02x\n", byte);
-	
-	ni = utils_pathname_to_inode (vol, NULL, "pagefile.sys");
+
+	ni = ntfs_pathname_to_inode(vol, NULL, "pagefile.sys");
 	if (!ni) {
 		Dprintf("Failed to open inode of pagefile.sys.\n");
 		return 0;
@@ -1181,7 +1214,7 @@ static s64 wipe_pagefile (ntfs_volume *vol, int byte, enum action act)
 		goto io_error_exit;
 	}
 
-	/* Get length of pagfile.sys contents. */
+	/* Get length of pagefile.sys contents. */
 	len = na->data_size;
 	if (!len) {
 		Dprintf("pagefile.sys has zero length, no disk write needed.\n");
@@ -1375,18 +1408,6 @@ int main (int argc, char *argv[])
 
 		printf ("%lld bytes were wiped\n", (long long)total);
 	}
-	/*
-	 * We need to reset the logfile so Windows can boot and so journal
-	 * replay does not cause corruption.
-	 */
-	if (act != act_info) {
-		printf ("Resetting logfile.\n");
-		ntfs_logfile_reset(vol);
-	}
-	if (ntfs_volume_set_flags (vol, VOLUME_IS_DIRTY) < 0) {
-		Eprintf ("Couldn't mark volume dirty\n");
-	}
-
 	result = 0;
 umount:
 	ntfs_umount (vol, FALSE);
@@ -1395,4 +1416,3 @@ free:
 		free (opts.bytes);
 	return result;
 }
-
