@@ -1,8 +1,8 @@
 /**
  * utils.c - Part of the Linux-NTFS project.
  *
- * Copyright (c) 2002-2003 Richard Russon
- * Copyright (c) 2003-2004 Anton Altaparmakov
+ * Copyright (c) 2002-2005 Richard Russon
+ * Copyright (c) 2003-2005 Anton Altaparmakov
  * Copyright (c) 2003 Lode Leroy
  *
  * A set of shared functions for ntfs utilities
@@ -25,20 +25,42 @@
 
 #include "config.h"
 
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_STDARG_H
 #include <stdarg.h>
+#endif
+#ifdef HAVE_ERRNO_H
 #include <errno.h>
+#endif
+#ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
+#endif
+#ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
+#endif
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
+#ifdef HAVE_LOCALE_H
 #include <locale.h>
+#endif
 #ifdef HAVE_LIBINTL_H
 #	include <libintl.h>
 #endif
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_LIMITS_H
 #include <limits.h>
+#endif
+#ifdef HAVE_CTYPE_H
 #include <ctype.h>
+#endif
 
 #include "config.h"
 #include "utils.h"
@@ -46,8 +68,9 @@
 #include "volume.h"
 #include "debug.h"
 #include "dir.h"
+#include "version.h"
 
-const char *ntfs_bugs = "Please report bugs to linux-ntfs-dev@lists.sourceforge.net\n";
+const char *ntfs_bugs = "Developers' email address: linux-ntfs-dev@lists.sourceforge.net\n";
 const char *ntfs_home = "Linux NTFS homepage: http://linux-ntfs.sourceforge.net\n";
 const char *ntfs_gpl = "This program is free software, released under the GNU "
 	"General Public License\nand you are welcome to redistribute it under "
@@ -56,13 +79,6 @@ const char *ntfs_gpl = "This program is free software, released under the GNU "
 	"\"COPYING\" distributed with this program, or online at:\n"
 	"http://www.gnu.org/copyleft/gpl.html\n";
 
-#define NTFS_TIME_OFFSET ((s64)(369 * 365 + 89) * 24 * 3600 * 10000000)
-
-/* These utilities require the following functions */
-extern int Eprintf (const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-extern int Vprintf (const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-extern int Qprintf (const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-
 /**
  * utils_set_locale
  */
@@ -161,7 +177,19 @@ ntfs_volume * utils_mount_volume (const char *device, unsigned long flags, BOOL
 
 	vol = ntfs_mount (device, flags);
 	if (!vol) {
-		Eprintf ("Couldn't mount device '%s': %s\n", device, strerror (errno));
+		int err;
+
+		err = errno;
+		Eprintf("Couldn't mount device '%s': %s\n", device,
+				strerror(err));
+		if (err == EPERM)
+			Eprintf("Windows was hibernated.  Try to mount volume "
+					"in windows, shut down and try "
+					"again.\n");
+		if (err == EOPNOTSUPP)
+			Eprintf("Windows did not shut down properly.  Try to "
+					"mount volume in windows, shut down "
+					"and try again.\n");
 		return NULL;
 	}
 
@@ -299,41 +327,6 @@ int utils_parse_range (const char *string, s64 *start, s64 *finish, BOOL scale)
 	return 1;
 }
 
-/**
- * ntfs2utc - Convert an NTFS time to Unix time
- * @ntfs_time:  An NTFS time in 100ns units since 1601
- *
- * NTFS stores times as the number of 100ns intervals since January 1st 1601 at
- * 00:00 UTC.  This system will not suffer from Y2K problems until ~57000AD.
- *
- * Return:  n  A Unix time (number of seconds since 1970)
- */
-time_t ntfs2utc (s64 ntfs_time)
-{
-	return (ntfs_time - (NTFS_TIME_OFFSET)) / 10000000;
-}
-
-/**
- * utc2ntfs - Convert Linux time to NTFS time
- * @utc_time:  Linux time to convert to NTFS
- *
- * Convert the Linux time @utc_time to its corresponding NTFS time.
- *
- * Linux stores time in a long at present and measures it as the number of
- * 1-second intervals since 1st January 1970, 00:00:00 UTC.
- *
- * NTFS uses Microsoft's standard time format which is stored in a s64 and is
- * measured as the number of 100 nano-second intervals since 1st January 1601,
- * 00:00:00 UTC.
- *
- * Return:  n  An NTFS time (100ns units since Jan 1601)
- */
-s64 utc2ntfs (time_t utc_time)
-{
-	/* Convert to 100ns intervals and then add the NTFS time offset. */
-	return (s64)utc_time * 10000000 + NTFS_TIME_OFFSET;
-}
-
 /**
  * find_attribute - Find an attribute of the given type
  * @type:  An attribute type, e.g. AT_FILE_NAME
@@ -466,7 +459,7 @@ int utils_inode_get_name (ntfs_inode *inode, char *buffer, int bufsize)
 			}
 
 			if (ntfs_ucstombs (attr->file_name, attr->file_name_length,
-			    &names[i], attr->file_name_length) < 0) {
+			    &names[i], 0) < 0) {
 				char *temp;
 				Eprintf ("Couldn't translate filename to current locale.\n");
 				temp = malloc (30);
@@ -547,7 +540,7 @@ int utils_attr_get_name (ntfs_volume *vol, ATTR_RECORD *attr, char *buffer, int
 	if (attrdef) {
 		name    = NULL;
 		namelen = ntfs_ucsnlen (attrdef->name, sizeof (attrdef->name));
-		if (ntfs_ucstombs (attrdef->name, namelen, &name, namelen) < 0) {
+		if (ntfs_ucstombs (attrdef->name, namelen, &name, 0) < 0) {
 			Eprintf ("Couldn't translate attribute type to current locale.\n");
 			// <UNKNOWN>?
 			return 0;
@@ -572,8 +565,8 @@ int utils_attr_get_name (ntfs_volume *vol, ATTR_RECORD *attr, char *buffer, int
 
 	name    = NULL;
 	namelen = attr->name_length;
-	if (ntfs_ucstombs ((uchar_t *)((char *)attr + attr->name_offset),
-	    namelen, &name, namelen) < 0) {
+	if (ntfs_ucstombs ((ntfschar *)((char *)attr + attr->name_offset),
+	    namelen, &name, 0) < 0) {
 		Eprintf ("Couldn't translate attribute name to current locale.\n");
 		// <UNKNOWN>?
 		len = snprintf (buffer, bufsize, "<UNKNOWN>");
@@ -624,7 +617,7 @@ int utils_cluster_in_use (ntfs_volume *vol, long long lcn)
 	/* Does lcn lie in the section of $Bitmap we already have cached? */
 	if ((lcn < bmplcn) || (lcn >= (bmplcn + (sizeof (buffer) << 3)))) {
 		Dprintf ("Bit lies outside cache.\n");
-		attr = ntfs_attr_open (vol->lcnbmp_ni, AT_DATA, NULL, 0);
+		attr = ntfs_attr_open (vol->lcnbmp_ni, AT_DATA, AT_UNNAMED, 0);
 		if (!attr) {
 			Eprintf ("Couldn't open $Bitmap: %s\n", strerror (errno));
 			return -1;
@@ -706,98 +699,6 @@ int utils_mftrec_in_use (ntfs_volume *vol, MFT_REF mref)
 	return (buffer[byte] & bit);
 }
 
-/**
- * utils_pathname_to_inode - Find the inode which represents the given pathname
- * @vol:       An ntfs volume obtained from ntfs_mount
- * @parent:    A directory inode to begin the search (may be NULL)
- * @pathname:  Pathname to be located
- *
- * Take an ASCII pathname and find the inode that represents it.  The function
- * splits the path and then descends the directory tree.  If @parent is NULL,
- * then the root directory '.' will be used as the base for the search.
- *
- * Return:  inode  Success, the pathname was valid
- *	    NULL   Error, the pathname was invalid, or some other error occurred
- */
-ntfs_inode * utils_pathname_to_inode (ntfs_volume *vol, ntfs_inode *parent, const char *pathname)
-{
-	u64 inum;
-	int len;
-	char *p, *q;
-	ntfs_inode *ni;
-	ntfs_inode *result  = NULL;
-	uchar_t    *unicode = NULL;
-	char       *ascii   = NULL;
-
-	if (!vol || !pathname) {
-		errno = EINVAL;
-		return NULL;
-	}
-
-	if (parent) {
-		ni = parent;
-	} else {
-		ni = ntfs_inode_open (vol, FILE_root);
-		if (!ni) {
-			Eprintf ("Couldn't open the inode of the root directory.\n");
-			goto close;
-		}
-	}
-
-	unicode = calloc (1, MAX_PATH);
-	ascii   = strdup (pathname);		// Work with a r/w copy
-	if (!unicode || !ascii) {
-		Eprintf ("Out of memory.\n");
-		goto close;
-	}
-
-	p = ascii;
-	while (p && *p && *p == PATH_SEP)	// Remove leading /'s
-		p++;
-	while (p && *p) {
-		q = strchr (p, PATH_SEP);	// Find the end of the first token
-		if (q != NULL) {
-			*q = '\0';
-			q++;
-		}
-
-		len = ntfs_mbstoucs (p, &unicode, MAX_PATH);
-		if (len < 0) {
-			Eprintf ("Couldn't convert name to Unicode: %s.\n", p);
-			goto close;
-		}
-
-		inum = ntfs_inode_lookup_by_name (ni, unicode, len);
-		if (inum == (u64)-1) {
-			Eprintf ("Couldn't find name '%s' in pathname '%s'.\n", p, pathname);
-			goto close;
-		}
-
-		if (ni != parent)
-			ntfs_inode_close (ni);
-
-		ni = ntfs_inode_open (vol, inum);
-		if (!ni) {
-			Eprintf ("Cannot open inode %llu: %s.\n",
-					(unsigned long long)inum, p);
-			goto close;
-		}
-
-		p = q;
-		while (p && *p && *p == PATH_SEP)
-			p++;
-	}
-
-	result = ni;
-	ni = NULL;
-close:
-	if (ni && (ni != parent))
-		ntfs_inode_close (ni);
-	free (ascii);
-	free (unicode);
-	return result;
-}
-
 /**
  * __metadata
  */
@@ -873,43 +774,63 @@ int utils_is_metadata (ntfs_inode *inode)
  * @buf:     Buffer to be displayed
  * @start:   Offset into @buf to start from
  * @length:  Number of bytes to display
- * @ascii:   Whether or not to display the ascii values
+ * @flags:   Options to change the style of the output
  *
  * Display a block of memory in a tradition hex-dump manner.
  * Optionally the ascii part can be turned off.
+ *
+ * The flags, described fully in utils.h, default to 0 (DM_DEFAULTS).
+ * Examples are: DM_INDENT (indent the output by one tab); DM_RED (colour the
+ * output); DM_NO_ASCII (only print the hex values).
  */
-void utils_dump_mem (u8 *buf, int start, int length, int ascii)
+void utils_dump_mem (void *buf, int start, int length, int flags)
 {
-	int off, i, s, e;
+	int off, i, s, e, col;
+	u8 *mem = buf;
 
 	s =  start                & ~15;	// round down
 	e = (start + length + 15) & ~15;	// round up
 
 	for (off = s; off < e; off += 16) {
+		col = 30;
+		if (flags & DM_RED)
+			col += 1;
+		if (flags & DM_GREEN)
+			col += 2;
+		if (flags & DM_BLUE)
+			col += 4;
+		if (flags & DM_INDENT)
+			printf ("\t");
+		if (flags & DM_BOLD)
+			printf ("[01m");
+		if (flags & (DM_RED | DM_BLUE | DM_GREEN | DM_BOLD))
+			printf ("[%dm", col);
 		if (off == s)
 			printf("%6.6x ", start);
 		else
 			printf("%6.6x ", off);
 
 		for (i = 0; i < 16; i++) {
-			if (i == 8)
+			if ((i == 8) && (!(flags & DM_NO_DIVIDER)))
 				printf (" -");
 			if (((off+i) >= start) && ((off+i) < (start+length)))
-				printf (" %02X", buf[off+i]);
+				printf (" %02X", mem[off+i]);
 			else
 				printf ("   ");
 		}
-		if (ascii) {
+		if (!(flags & DM_NO_ASCII)) {
 			printf ("  ");
 			for (i = 0; i < 16; i++) {
 				if (((off+i) < start) || ((off+i) >= (start+length)))
 					printf (" ");
-				else if (isprint (buf[off + i]))
-					printf ("%c", buf[off + i]);
+				else if (isprint (mem[off + i]))
+					printf ("%c", mem[off + i]);
 				else
 					printf (".");
 			}
 		}
+		if (flags & (DM_RED | DM_BLUE | DM_GREEN | DM_BOLD))
+			printf ("[0m");
 		printf ("\n");
 	}
 }
@@ -952,6 +873,7 @@ void mft_put_search_ctx (struct mft_search_ctx *ctx)
  */
 int mft_next_record (struct mft_search_ctx *ctx)
 {
+	s64 nr_mft_records;
 	ATTR_RECORD *attr10 = NULL;
 	ATTR_RECORD *attr20 = NULL;
 	ATTR_RECORD *attr80 = NULL;
@@ -967,11 +889,16 @@ int mft_next_record (struct mft_search_ctx *ctx)
 		ctx->inode = NULL;
 	}
 
-	for (ctx->mft_num++; (s64)ctx->mft_num < ctx->vol->nr_mft_records; ctx->mft_num++) {
+	nr_mft_records = ctx->vol->mft_na->initialized_size >>
+			ctx->vol->mft_record_size_bits;
+
+	for (ctx->mft_num++; (s64)ctx->mft_num < nr_mft_records; ctx->mft_num++) {
+		int in_use;
+
 		ctx->flags_match = 0;
-		int in_use = utils_mftrec_in_use (ctx->vol, (MFT_REF) ctx->mft_num);
+		in_use = utils_mftrec_in_use (ctx->vol, (MFT_REF) ctx->mft_num);
 		if (in_use == -1) {
-			Eprintf ("Error reading inode %llu.  Aborting.",
+			Eprintf ("Error reading inode %llu.  Aborting.\n",
 					(unsigned long long)ctx->mft_num);
 			return -1;
 		}
@@ -981,9 +908,9 @@ int mft_next_record (struct mft_search_ctx *ctx)
 
 			ctx->inode = ntfs_inode_open (ctx->vol, (MFT_REF) ctx->mft_num);
 			if (ctx->inode == NULL) {
-				Eprintf ("Error reading inode %llu.", (unsigned
+				Eprintf ("Error reading inode %llu.\n", (unsigned
 						long long) ctx->mft_num);
-				return -1;
+				continue;
 			}
 
 			attr10 = find_first_attribute (AT_STANDARD_INFORMATION, ctx->inode->mrec);
@@ -1002,7 +929,7 @@ int mft_next_record (struct mft_search_ctx *ctx)
 				ctx->flags_match |= FEMR_FILE;
 
 			if (ctx->flags_search & FEMR_DIR) {
-				attr_ctx = ntfs_attr_get_search_ctx (NULL, ctx->inode->mrec);
+				attr_ctx = ntfs_attr_get_search_ctx (ctx->inode, NULL);
 				if (attr_ctx) {
 					if (ntfs_attr_lookup (AT_INDEX_ROOT, I30, 4, 0, 0, NULL, 0, attr_ctx) == 0)
 						ctx->flags_match |= FEMR_DIR;
@@ -1019,7 +946,7 @@ int mft_next_record (struct mft_search_ctx *ctx)
 				case 0: ctx->flags_match |= FEMR_NOT_METADATA; break;
 				default:
 					ctx->flags_match |= FEMR_NOT_METADATA; break;
-					//Eprintf ("Error reading inode %lld.", ctx->mft_num);
+					//Eprintf ("Error reading inode %lld.\n", ctx->mft_num);
 					//return -1;
 			}
 
@@ -1030,7 +957,7 @@ int mft_next_record (struct mft_search_ctx *ctx)
 
 			ctx->inode = calloc (1, sizeof (*ctx->inode));
 			if (!ctx->inode) {
-				Eprintf ("Out of memory.  Aborting.");
+				Eprintf ("Out of memory.  Aborting.\n");
 				return -1;
 			}
 
@@ -1039,11 +966,12 @@ int mft_next_record (struct mft_search_ctx *ctx)
 			ctx->inode->mrec   = malloc (ctx->vol->mft_record_size);
 			if (!ctx->inode->mrec) {
 				free (ctx->inode); // == ntfs_inode_close
-				Eprintf ("Out of memory.  Aborting.");
+				Eprintf ("Out of memory.  Aborting.\n");
 				return -1;
 			}
 
-			mft = ntfs_attr_open (ctx->vol->mft_ni, AT_DATA, NULL, 0);
+			mft = ntfs_attr_open (ctx->vol->mft_ni, AT_DATA,
+					AT_UNNAMED, 0);
 			if (!mft) {
 				Eprintf ("Couldn't open $MFT/$DATA: %s\n", strerror (errno));
 				// free / close
@@ -1055,8 +983,11 @@ int mft_next_record (struct mft_search_ctx *ctx)
 						(unsigned long long)
 						ctx->mft_num, strerror (errno));
 				// free / close
+				ntfs_attr_close(mft);
 				return -1;
 			}
+			
+			ntfs_attr_close(mft);
 		}
 
 		if (ctx->flags_match & ctx->flags_search) {
@@ -1064,7 +995,7 @@ int mft_next_record (struct mft_search_ctx *ctx)
 		}
 
 		if (ntfs_inode_close (ctx->inode)) {
-			Eprintf ("Error closing inode %llu.",
+			Eprintf ("Error closing inode %llu.\n",
 					(unsigned long long)ctx->mft_num);
 			return -errno;
 		}
